import bs
import bsUtils
import bsSpaz
import random
import weakref
import bsUI

gRandProfileIndex = 1
gLastWarnTime = 0

class PlayerReadyMessage(object):
    def __init__(self,chooser):
        self.chooser = chooser

class ChangeMessage(object):
    def __init__(self,what,value):
        self.what = what
        self.value = value

class Chooser(object):

    def __init__(self,vPos,playerID,lobby):

        self._deekSound = bs.getSound("deek")
        self._clickSound = bs.getSound("click01")
        self._punchSound = bs.getSound("punch01")
        self._swishSound = bs.getSound('punchSwish')
        self._frameTexture = bs.getTexture("frame")
        self._vPos = vPos
        self._lobby = weakref.ref(lobby)
        self._playerID = playerID
        self._inited = False

        # note: this is just the index in our local list of teams; not the team ID!
        self._selectedTeamIndex = self.getLobby().nextAddTeam

        # use this as an offset so input device 0 (the keyboard on desktops, touch-input on mobile) gets spaz as a default
        spazIndex = self.getLobby().characterNames.index('Spaz')

        # store a persistent random character index; we'll use this for the '_random' profile
        # lets use their inputID to seed it.. this will give a persistent character
        # for them between games and will distribute characters nicely if everyone is random
        self._randomCharacterIndex = (bs.getPlayerInputDevice(self._playerID)+spazIndex)%len(self.getLobby().characterNames)
        self._randomColor,self._randomHighlight = bsUtils.getPlayerProfileColors(None)

        # attempt to pick an initial profile based on what's been stored for this input device
        try:
            inputDevice = bs.getPlayerInputDevice(self._playerID)
            name = bs.getInputDeviceName(inputDevice)
            uniqueID = bs.getInputDeviceUniqueIdentifier(inputDevice)
            self.playerProfileName = bs.config['Default Player Profiles'][name+' '+uniqueID]
            self.playerProfile = self.getLobby().playerProfileNames.index(self.playerProfileName)

        # well, that didn't work.. just start with the first profile (should be '_random')
        except Exception:
            global gRandProfileIndex
            # cycle through our non-random profiles once; after that, all is random
            if gRandProfileIndex < len(self.getLobby().playerProfileNames):
                self.playerProfile = gRandProfileIndex%len(self.getLobby().playerProfileNames)
            else:
                self.playerProfile = 0
            gRandProfileIndex += 1
            self.playerProfileName = self.getLobby().playerProfileNames[self.playerProfile]
        self.character = self._randomCharacterIndex
        self._color = self._randomColor
        self._highlight = self._randomHighlight
        self.ready = False
        self._text = bs.NodeActor(None,bs.Actor.UNRETAINED,
                                  "text",{'translate':(-100,self._vPos),
                                          'text':"hello",
                                          'hAlign':"left",
                                          'vAttach':"top"},
                                  name='Team Chooser Text',
                                  delegate=self)
        bsUtils.animate(self._text.node,'scale',{0:0,100:1.0})
        self.icon = bs.newNode("image",
                               owner=self._text.node,
                               attrs={'translate':(-130,self._vPos+20),
                                      'hasAlphaChannel':False,
                                      'attach':'topCenter'},name='Team Chooser Icon')
        bsUtils.animateArray(self.icon,'scale',2,{0:(0,0),100:(32,32)})
        self.iconBack = bs.newNode("image",
                                   owner=self._text.node,
                                   attrs={'translate':(-130,self._vPos+20),
                                          'texture':self._frameTexture,
                                          'attach':'topCenter'},name='Team Chooser Icon Backing')
        bsUtils.animateArray(self.iconBack,'scale',2,{0:(0,0),100:(53,53)})

        self._setReady(False)

        # make sure we're not reserving real names while we're still picking..
        bs.setPlayerName(self._playerID,'<choosing player>')

        self.updateFromPlayerProfile()
        self.updatePosition()
        self._inited = True

    def getTeam(self):
        """ return the selected team """
        return self._lobby()._teams[self._selectedTeamIndex]()

    def getLobby(self):
        return self._lobby()

    def updateFromPlayerProfile(self):
        # set character based on profile; otherwise default to our pre-picked random
        try:
            # store the name even though we usually use index (in case the profile list changes)
            self.playerProfileName = self.getLobby().playerProfileNames[self.playerProfile]
            pc = self.getLobby().playerProfiles[self.playerProfileName]['character']
            self.character = self.getLobby().characterNames.index(pc)
            self._color,self._highlight = bsUtils.getPlayerProfileColors(self.playerProfileName)
        except Exception:
            self.character = self._randomCharacterIndex
            self._color = self._randomColor
            self._highlight = self._randomHighlight
        self._updateIcon()
        self._updateText()

    def reloadProfiles(self):
        # re-construct our profile index and other stuff since the profile list has changed
        try:
            self.playerProfile = self.getLobby().playerProfileNames.index(self.playerProfileName)
        except Exception:
            self.playerProfile = 0
            self.playerProfileName = self.getLobby().playerProfileNames[self.playerProfile]
        self.updateFromPlayerProfile()

    def updatePosition(self):
        spacing = 350
        offs = spacing*-0.5*len(self.getLobby()._teams) + spacing*self._selectedTeamIndex + 250
        curTranslate = self._text.node.translate
        bsUtils.animateArray(self._text.node,'translate',2,{0:self._text.node.translate,100:(-100+offs,self._vPos+7)})
        bsUtils.animateArray(self.icon,'translate',2,{0:self.icon.translate,100:(-130+offs,self._vPos+20)})
        bsUtils.animateArray(self.iconBack,'translate',2,{0:self.iconBack.translate,100:(-130+offs,self._vPos+20)})
        
    def getCharacterName(self):
        return self.getLobby().characterNames[self.character]
    
    def __doNothing(self):
        pass

    def _getName(self):
        name = self.getLobby().playerProfileNames[self.playerProfile]
        if name == '_random':
            name = bs.getInputDeviceDefaultPlayerName(bs.getPlayerInputDevice(self._playerID))
        elif name == '_zNew':
            name = '<New Profile>'
        return name

    def _setReady(self,ready):

        # when this is called on the '_zNew' option, it triggers the profile editor instead...
        profileName = self.getLobby().playerProfileNames[self.playerProfile]
        
        if profileName == '_zNew':
            import bsUI
            def _editDone(profileName):
                if profileName != None:
                    self.getLobby().reloadProfiles()
                    self.playerProfileName = profileName
                    self.playerProfile = self.getLobby().playerProfileNames.index(self.playerProfileName)
                    self.updateFromPlayerProfile()
                bsUI.resume()

            # grab menu control for this player
            bs.setUIInputDevice(bs.getPlayerInputDevice(self._playerID))
            bsUI.editProfileWindow(existingProfile=None,completionCall=_editDone)
            bsUI.pause()

        else:
            if not ready:
                bs.assignPlayerInput(self._playerID,"leftPress",bs.WeakCall(self.handleMessage,ChangeMessage('team',-1)))
                bs.assignPlayerInput(self._playerID,"rightPress",bs.WeakCall(self.handleMessage,ChangeMessage('team',1)))
                bs.assignPlayerInput(self._playerID,"pickUpPress",bs.WeakCall(self.handleMessage,ChangeMessage('character',1)))
                bs.assignPlayerInput(self._playerID,"upPress",bs.WeakCall(self.handleMessage,ChangeMessage('playerProfile',-1)))
                bs.assignPlayerInput(self._playerID,"downPress",bs.WeakCall(self.handleMessage,ChangeMessage('playerProfile',1)))
                bs.assignPlayerInput(self._playerID,("jumpPress","bombPress","punchPress"),
                                     bs.WeakCall(self.handleMessage,ChangeMessage('ready',1)))
                self.ready = False
                self._updateText()
                bs.setPlayerName(self._playerID,'untitled')
            else:
                bs.assignPlayerInput(self._playerID,("leftPress","rightPress",
                                                     "upPress","downPress",
                                                     "jumpPress","bombPress",
                                                     "pickUpPress"),bs.WeakCall(self.__doNothing))
                bs.assignPlayerInput(self._playerID,("jumpPress","bombPress",
                                                     "pickUpPress","punchPress"),bs.WeakCall(self.handleMessage,ChangeMessage('ready',0)))

                # store the last profile picked by this input for reuse
                inputDevice = bs.getPlayerInputDevice(self._playerID)
                name = bs.getInputDeviceName(inputDevice)
                uniqueID = bs.getInputDeviceUniqueIdentifier(inputDevice)
                try: deviceProfiles = bs.config['Default Player Profiles']
                except Exception: deviceProfiles = bs.config['Default Player Profiles'] = {}

                # exception is if we're set to _random, in which case we clear ourself from this
                if profileName == '_random':
                    try: del(deviceProfiles[name+' '+uniqueID])
                    except Exception: pass
                else:
                    deviceProfiles[name+' '+uniqueID] = profileName
                bs.writeConfig()

                # set this player's name
                bs.setPlayerName(self._playerID,self._getName())
                self.ready = True
                self._updateText()

                # inform the user that this player is ready
                try: delegate = self.getLobby()._delegate()
                except Exception,e: delegate = None
                if delegate: delegate.handleMessage(PlayerReadyMessage(self))
                
    def handleMessage(self,m):

        if isinstance(m,ChangeMessage):

            if m.what == "team":
                if len(self.getLobby()._teams) > 1: bs.playSound(self._swishSound)
                self._selectedTeamIndex = (self._selectedTeamIndex + m.value) % len(self.getLobby()._teams)
                self._updateText()
                self.updatePosition()
                self._updateIcon()

            elif m.what == "playerProfile":
                if len(self.getLobby().playerProfileNames) == 1:
                    global gLastWarnTime
                    if bs.getRealTime() - gLastWarnTime > 5000:
                        gLastWarnTime = bs.getRealTime()
                        bs.screenMessage(bs.getResource('noProfilesErrorText').replace('${NAME}',self._getName()))
                        bs.playSound(bs.getSound('error'))
                else:
                    # pick the next player profile and assign our name and character based on that
                    bs.playSound(self._deekSound)
                    self.playerProfile = (self.playerProfile + m.value) % len(self.getLobby().playerProfileNames)
                    self.updateFromPlayerProfile()
                
            elif m.what == "character":
                bs.playSound(self._clickSound)
                self.character = (self.character + m.value) % len(self.getLobby().characterNames)
                self._updateText()
                self._updateIcon()

            elif m.what == "ready":
                bs.playSound(self._punchSound)
                self._setReady(m.value)

    def _updateText(self):

        if self.ready:
            # once we're ready, we've saved the name, so lets ask the system for it
            # so we get appended numbers and stuff
            text = bs.getPlayerName(self._playerID)
            text = text + ' ('+bs.getResource('readyText')+')'
        else:
            text = self._getName()

        canSwitchTeams = len(self.getLobby()._teams) > 1

        # flash as we're coming in
        finColor = bsUtils.getSafeDisplayColor(self.getColor())+(1,)
        if not self._inited:
            bsUtils.animateArray(self._text.node,'color',4,{150:finColor,250:(2,2,2,1),350:finColor})
        else:
            # blend if we're in teams mode; switch instantly otherwise
            if canSwitchTeams:
                bsUtils.animateArray(self._text.node,'color',4,{0:self._text.node.color,100:finColor})
            else:
                self._text.node.color = finColor

        self._text.node.text = text

    def getColor(self):
        if self.getLobby()._useTeamColors:
            return self.getLobby()._teams[self._selectedTeamIndex]().color
        else:
            return self._color

    def getHighlight(self):
        # if we're using team colors we wanna make sure our highlight color
        # isn't too close to any other team's color
        highlight = list(self._highlight)
        if self.getLobby()._useTeamColors:
            for i,teamRef in enumerate(self.getLobby()._teams):
                team = teamRef()
                if i != self._selectedTeamIndex:
                    # find the dominant component this team's color
                    # and adjust ours so that component is not super-dominant
                    maxVal = 0
                    maxIndex = 0
                    for j in range(3):
                        if team.color[j] > maxVal:
                            maxVal = team.color[j]
                            maxIndex = j
                    thatColorForUs = highlight[maxIndex]
                    ourSecondBiggest = max(highlight[(maxIndex+1)%3],highlight[(maxIndex+2)%3])
                    diff = (thatColorForUs-ourSecondBiggest)
                    if diff > 0:
                        highlight[maxIndex] -= diff * 0.6
                        highlight[(maxIndex+1)%3] += diff * 0.3
                        highlight[(maxIndex+2)%3] += diff * 0.2
        return highlight

    def getPlayerID(self):
        return self._playerID

    def _updateIcon(self):

        if self.getLobby().playerProfileNames[self.playerProfile] == '_zNew':
            self.icon.color = (0,0,0)
            tc = self.getLobby()._teams[self._selectedTeamIndex]().color
            self.iconBack.color = (tc[0]*0.6,tc[1]*0.6,tc[2]*0.6)
        else:
            tex = self.getLobby().characterIconTextures[self.character]
            tintTex = self.getLobby().characterIconTintTextures[self.character]
            self.icon.color = (1,1,1)
            self.icon.texture = tex
            self.icon.tintTexture = tintTex
            c = self.getColor()
            c2 = self.getHighlight()

            canSwitchTeams = len(self.getLobby()._teams) > 1

            # if we're initing, flash
            if not self._inited:
                bsUtils.animateArray(self.icon,'color',3,{150:(1,1,1),250:(2,2,2),350:(1,1,1)})

            # blend in teams mode; switch instantly in ffa-mode
            if canSwitchTeams:
                bsUtils.animateArray(self.icon,'tintColor',3,{0:self.icon.tintColor,100:c})
            else:
                self.icon.tintColor = c
                    

            self.icon.tint2Color = c2

            finColor = (0.2+0.8*c[0],0.2+0.8*c[1],0.2+0.8*c[2])
            if not self._inited:
                bsUtils.animateArray(self.iconBack,'color',3,{150:finColor,250:(3,3,3),350:finColor})
            else:
                if canSwitchTeams:
                    bsUtils.animateArray(self.iconBack,'color',3,{0:self.iconBack.color,100:finColor})
                else:
                    self.iconBack.color = finColor

            # we create a cache of last-used icons for all joining players
            # ..this can be used by score-screens and whatnot since the player himself
            # may no longer be around.
            self.getLobby()._playerIcons[self._playerID] = {'texture':tex,'tintTexture':tintTex,'tintColor':tuple(c),'tint2Color':tuple(c2)}

class Lobby(object):

    def __init__(self,teams=None,delegate=None,useTeamColors=True,vOffset=0):
        self._useTeamColors = useTeamColors
        if teams is not None:
            self._teams = [weakref.ref(team) for team in teams]
        else:
            self._dummyTeam = bs.Team()
            self._teams = [weakref.ref(self._dummyTeam)]

        self.setDelegate(delegate)
        self.choosers = []
        self.baseVOffset = vOffset
        self.updatePositions()
        self.nextAddTeam = 0
        self._playerIcons = {}

        # grab available characters
        #self.characterNames = bsSpaz.appearances.keys()
        self.characterNames = bsSpaz.getAppearances()
        self.characterNames.sort()
        self.reloadProfiles()

        # create a cache of character icons
        self.characterIconTextures = []
        self.characterIconTintTextures = []
        for character in self.characterNames:
            self.characterIconTextures.append(bs.getTexture(bsSpaz.appearances[character].iconTexture,minQuality='High'))
            self.characterIconTintTextures.append(bs.getTexture(bsSpaz.appearances[character].iconMaskTexture))
        self._joinInfoText = None

    def setDelegate(self,delegate):
        self._delegate = weakref.ref(delegate) if delegate != None else None
        
    def getChoosers(self):
        return self.choosers

    class JoinInfo(object):

        def __init__(self,lobby):
            self._text = bs.NodeActor(None,bs.Actor.UNRETAINED,
                                      "text",{'translate':(0,-40),
                                              'hAttach':"center",'vAttach':"top",
                                              'hAlign':'center',
                                              'color':(0.8,0.8,0.8,1.0),
                                              'text':bs.getResource('waitingForLocalPlayersText')},
                                      name='Join Info Text')
            canSwitchTeams = (len(lobby._teams) > 1)
            self._state = 0

            pressToPunch = '\xc2\x84'
            pressToPickup = '\xc2\x85'

            # if we have a keyboard, grab keys for punch and pickup
            # FIXME - this of course is only correct locally; will need to change this for net games
            try: kb = bs.getInputDevice('Keyboard','#1')
            except Exception: kb = None
            if kb is not None:
                punchKey = bs.getKeyName(kb,bsUI.getControllerValue(kb,'buttonPunch'))
                pressToPunch = '\''+punchKey+'\' or '+pressToPunch
                pickupKey = bs.getKeyName(kb,bsUI.getControllerValue(kb,'buttonPickUp'))
                pressToPickup = '\''+pickupKey+'\' or '+pressToPickup
                joinStr = bs.getResource('pressPunchToJoinText')+'  < '+pressToPunch+' >'
            else:
                joinStr = bs.getResource('pressAnyButtonToJoinText')
                
            self._messages = ([joinStr]
                              + ([bs.getResource('pressToSelectTeamText').replace('${BUTTONS}','\xc2\x81 \xc2\x80')] if canSwitchTeams else [])
                              + [bs.getResource('pressToSelectProfileText').replace('${BUTTONS}','\xc2\x82 \xc2\x83')]
                              + [bs.getResource('pressToOverrideCharacterText').replace('${BUTTONS}',bs.getResource('pickUpBoldText'))+' < '+pressToPickup+' >'])
            self._timer = bs.Timer(4000,bs.WeakCall(self._update),repeat=True)

        def _update(self):
            self._text.node.text = self._messages[self._state]
            self._state = (self._state+1)%len(self._messages)

    def createJoinInfo(self):
        return self.JoinInfo(self)

    def getPlayerIcon(self,playerID):
        """ given a player, returns the textures of the last character/team they selected """
        try: return self._playerIcons[playerID]
        except Exception: raise Exception("Player icon not found")

    def reloadProfiles(self):

        # grab available player profiles
        try: self.playerProfiles = dict(bs.config['Player Profiles'])
        except Exception: self.playerProfiles = {}

        # add in a random one so we're ok even if there's no user-created profiles
        self.playerProfiles['_random'] = {}
        self.playerProfileNames = self.playerProfiles.keys()
        self.playerProfileNames.sort(key=lambda x:x[0].lower())
        for chooser in self.choosers:
            try: chooser.reloadProfiles()
            except Exception,e: print 'Error reloading profiles:',e

    def updatePositions(self):
        self._vPos = -100+self.baseVOffset
        for chooser in self.choosers:
            chooser._vPos = self._vPos
            chooser.updatePosition()
            self._vPos -= 48

    def checkAllReady(self):
        for chooser in self.choosers:
            if not chooser.ready:
                return False
        return True
                
    def addChooser(self,playerID):
        self.choosers.append(Chooser(vPos=self._vPos,playerID=playerID,lobby=self))
        self.nextAddTeam = (self.nextAddTeam+1)%len(self._teams)
        self._vPos -= 48

    def removeAllChoosers(self):
        self.choosers = []
        self.updatePositions()
        
    def removeChooser(self,playerID):
        for chooser in self.choosers:
            if chooser.getPlayerID() == playerID:
                self.choosers.remove(chooser)
        self.updatePositions()
