import bs
import os
import bsUtils
import random
import bsSpaz
import copy
import bsMap
import math
import bsCoopGame
import time
import bsAchievement
import weakref
import bsServerData

uiGlobals = {'mainMenuWindow':None}

quitWindowID = None

# we include this extra hash with shared input-mapping names
# so that we don't share mappings between differently-configured systems
# for instance, it looks like OUYA will give different keycodes for some gamepads
# than vanilla android will so we want to store their configs distinctly
gInputMapHash = None

gGameTypeSelection = 'Co-op Games'

gTitleColor = (0.72,0.7,0.75,1.0)
gHeadingColor = (0.72,0.7,0.75,1.0)
gInfoTextColor=(0.7,0.9,0.7,1.0)
gDidMenuIntro = False
gIconSelectColor = (0.4,0.3,1)

gUntestedGamePads = []
gCanAskToConfigGamePads = True
gShouldAskToMakeProfile = False

gSmallUI = bs.getInterfaceType() == 'phone'
gMedUI = bs.getInterfaceType() == 'tabletSmall'

# testing
# gSmallUI = True
# gMedUI = False

# for custom UI settings:
customMenuUI = []
customMenuUI2 = []

if bs.isDebugBuild():
    bsUtils.suppressDebugReports()

def getInputMapHash(inputDevice):
    global gInputMapHash
    # currently we just do a single hash of *all* inputs on android and thats it.. good enough.
    # (grabbing mappings for a specific device looks to be non-trivial)
    try:
        if gInputMapHash is None:
            if 'android' in bs.getUserAgentString():
                import hashlib
                md5 = hashlib.md5()
                for d in ['/system/usr/keylayout','/data/usr/keylayout','/data/system/devices/keylayout']:
                    try:
                        if os.path.isdir(d):
                            for fName in os.listdir(d):
                                # this is usually volume keys and stuff; assume we can skip it?..(since it'll vary a lot across devices)
                                if fName == 'gpio-keys.kl': continue
                                f = open(d+'/'+fName)
                                md5.update(f.read())
                                f.close()
                    except Exception,e:
                        print 'Exception in getInputMapHash inner loop:',e
                        
                gInputMapHash = md5.hexdigest()
            else: gInputMapHash = ''
        return gInputMapHash
    except Exception,e:
        print 'Exception in getInputMapHash:',e
        return ''

def dismissWiiRemotesWindow():
    pass

def _configChanged(val):
    bs.writeConfig()
    bs.applySettings()

def configCheckBox(parent,name,position,size,displayName=None,scale=None,maxWidth=None):
    if displayName is None: displayName = name

    def _valueChanged(val):
        bs.config[name] = val

        # special case: clicking kick-idle-players repeatedly unlocks co-op levels
        if name == 'Kick Idle Players':
            global gKickIdlePlayersKickCount
            try: count = gKickIdlePlayersKickCount
            except Exception: count = 1
            count += 1
            if count == 21:
                DebugWindow()
                count = 1;
            gKickIdlePlayersKickCount = count

        bs.applySettings()
        bs.writeConfig()

    return bs.checkBoxWidget(parent=parent,position=position,size=size,text=displayName,textColor=(0.8,0.8,0.8),
                             value=bs.getSetting(name),valueCall=_valueChanged,scale=scale,maxWidth=maxWidth)
    

def _inc(ctrl,name,minVal,maxVal,increment,callback):
    val = float(bs.textWidget(query=ctrl))
    val += increment
    val = max(minVal,min(val,maxVal))
    bs.textWidget(edit=ctrl,text=str(round(val,2)))
    bs.config[name] = val
    if callback:
        callback(val)
    _configChanged(val)

def configTextBox(parent,name,position,type="string",minVal=0,maxVal=100,increment=1.0,callback=None,changeSound=True,xOffset=0,displayName=None,textScale=1.0):

    if displayName is None: displayName = name
    if type == "int": initStr = "str(int(bs.getSetting(" + repr(name) + ")))"
    elif type == "float": initStr = "str(round(bs.getSetting(" + repr(name) + "),2))"
    else: initStr = "str(bs.getSetting(" + repr(name) + "))"
    t = bs.textWidget(parent=parent,position=position,size=(100,30),text=displayName,maxWidth=160+xOffset,
                      color=(0.8,0.8,0.8,1.0),hAlign="left",vAlign="center",scale=textScale)
    if type == 'string':
        raise Exception("fixme unimplemented");
    else:
        t = bs.textWidget(parent=parent,position=(246+xOffset,position[1]),size=(60,28),editable=False,
                          color=(0.3,1.0,0.3,1.0),
                          hAlign="right",vAlign="center",
                          text=eval(initStr),padding=2)
        b = bs.buttonWidget(parent=parent,position=(330+xOffset,position[1]),size=(28,28),label="- ",
                            call=bs.WeakCall(_inc,t,name,minVal,maxVal,-increment,callback),repeat=True,enableSound=(changeSound is True))
        b = bs.buttonWidget(parent=parent,position=(380+xOffset,position[1]),size=(28,28),label="+ ",
                            call=bs.WeakCall(_inc,t,name,minVal,maxVal,increment,callback),repeat=True,enableSound=(changeSound is True))

def _makeRadioGroup(checkBoxes,valueNames,value,valueChangeCall):
    """ link the provided checkBoxes together into a radio group """
    def _radioPress(checkString,otherCheckBoxes,value):
        if value == 1:
            valueChangeCall(checkString)
            for cb in otherCheckBoxes:
                bs.checkBoxWidget(edit=cb,value=0)
    for i,checkBox in enumerate(checkBoxes):
        bs.checkBoxWidget(edit=checkBox,value=(value==valueNames[i]),valueCall=bs.Call(_radioPress,valueNames[i],[c for c in checkBoxes if c != checkBox]),isRadioButton=True)


class Window(object):
    def getRootWidget(self):
        return self._rootWidget


class ConfirmWindow(Window):

    def __init__(self,text="Are you sure?",action=None,width=360,height=100,cancelButton=True):
        if width < 360: width = 360
        self._action = action
        self._rootWidget = bs.containerWidget(size=(width,height+40),transition='inRight',scale=2.1 if gSmallUI else 1.5 if gMedUI else 1.0)
        padding = 20
        t = bs.textWidget(parent=self._rootWidget,position=(padding,padding+47),size=(width-2*padding,height-2*padding),
                          hAlign="center",vAlign="center",text=text,maxWidth=width*0.9)
        if cancelButton:
            b = bs.buttonWidget(parent=self._rootWidget,position=(20,20),size=(150,50),label=bs.getResource('cancelText'),call=self._cancel)
            bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        b = bs.buttonWidget(parent=self._rootWidget,position=(width-175,20),size=(150,50),label=bs.getResource('okText'),call=self._ok)
        bs.containerWidget(edit=self._rootWidget,selectedChild=b,startButton=b)

    def _cancel(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')

    def _ok(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        if self._action is not None:
            self._action()


class QuitWindow(Window):

    def __init__(self):
        global quitWindowID
        if quitWindowID != None:
            bs.deleteWidget(quitWindowID)
            quitWindowID = None
        self._rootWidget = quitWindowID = ConfirmWindow(bs.getResource('quitGameText') if 'Mac' in bs.getUserAgentString() else bs.getResource('exitGameText'),
                                                        self._doFadeAndQuit).getRootWidget()

    def _doFadeAndQuit(self):
        bs.fadeLocalScreen(False,time=200,endCall=bs.Call(bs.quit,soft=True))
        bs.lockAllInput()
        # unlock and fade back in shortly.. just in case something goes wrong
        # (or on android where quit just exits our activity and we may come back)
        bs.realTimer(300,bs.unlockAllInput)
        bs.realTimer(300,bs.Call(bs.fadeLocalScreen,True))

class DebugWindow(Window):

    def __init__(self):
        width = 460
        height = 460

        self._stressTestGameType = 'Random'
        self._stressTestPlaylist = '__default__'
        self._stressTestPlayerCount = 8
        self._stressTestRoundDuration = 120

        bs.playSound(bs.getSound('gong'))

        R = bs.getResource('debugWindow')
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight',scale=1.5 if gSmallUI else 1.25 if gMedUI else 1.0)
        b = bs.buttonWidget(parent=self._rootWidget,position=(width-130,height-62),size=(120,60),scale=0.8,
                            label=bs.getResource('doneText'), call=self._done)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        t = bs.textWidget(parent=self._rootWidget,position=(0,height-54),size=(width,30),
                          text=R.titleText,hAlign="center",color=gTitleColor,
                          vAlign="center",maxWidth=185)
        v = height - 140
        buttonWidth = 300
        b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)*0.5,v),size=(buttonWidth,60),
                            label=R.unlockCoopText,call=self._unlockPressed)
        v -= 60

        t = bs.textWidget(parent=self._rootWidget,position=(width*0.5,v+22),size=(0,0),text=R.stressTestTitleText,
                          maxWidth=200,color=gHeadingColor,
                          scale=0.85,hAlign="center",vAlign="center")
        v -= 45

        xOffs = 165
        t = bs.textWidget(parent=self._rootWidget,position=(xOffs-10,v+22),size=(0,0),text=R.stressTestPlaylistTypeText,
                          maxWidth=130,color=gHeadingColor,
                          scale=0.65,hAlign="right",vAlign="center")

        PopupMenu(parent=self._rootWidget,position=(xOffs,v),width=150,scale=1.3,
                  choices=['Random','Teams','Free-For-All'],
                  choicesDisplay=[bs.getResource(a) for a in ['randomText','teamsText','freeForAllText']],
                  currentChoice='Auto',changeCall=self._stressTestGameTypeSelected)
        v -= 43
        t = bs.textWidget(parent=self._rootWidget,position=(xOffs-10,v+22),size=(0,0),text=R.stressTestPlaylistNameText,
                          maxWidth=130,color=gHeadingColor,
                          scale=0.65,hAlign="right",vAlign="center")
        self._stressTestPlaylistNameField = bs.textWidget(parent=self._rootWidget,position=(xOffs+5,v),size=(250,36),
                                                          text=self._stressTestPlaylist,hAlign="left",
                                                          vAlign="center",
                                                          color=(0.9,0.9,0.9,1.0),
                                                          description=R.stressTestPlaylistDescriptionText,
                                                          editable=True,padding=4)
        v -= 29

        xSub = 60
        # player count
        t = bs.textWidget(parent=self._rootWidget,position=(xOffs-10,v),size=(0,0),text=R.stressTestPlayerCountText,
                          color=(0.8,0.8,0.8,1.0),hAlign="right",vAlign="center",scale=0.65,
                          maxWidth=130)
        self._stressTestPlayerCountText = bs.textWidget(parent=self._rootWidget,position=(246-xSub,v-14),
                                                      size=(60,28),editable=False,
                                                      color=(0.3,1.0,0.3,1.0),hAlign="right",vAlign="center",
                                                      text=str(self._stressTestPlayerCount),
                          padding=2)
        b = bs.buttonWidget(parent=self._rootWidget,position=(330-xSub,v-11),size=(28,28),label="- ",
                            call=bs.Call(self._stressTestPlayerCountDecrement),
                            repeat=True,enableSound=True)
        b = bs.buttonWidget(parent=self._rootWidget,position=(380-xSub,v-11),size=(28,28),label="+ ",
                            call=bs.Call(self._stressTestPlayerCountIncrement),
                            repeat=True,enableSound=True)


        v -= 42

        # round duration
        t = bs.textWidget(parent=self._rootWidget,position=(xOffs-10,v),size=(0,0),text=R.stressTestRoundDurationText,
                          color=(0.8,0.8,0.8,1.0),hAlign="right",vAlign="center",scale=0.65,
                          maxWidth=130)
        self._stressTestRoundDurationText = bs.textWidget(parent=self._rootWidget,position=(246-xSub,v-14),
                                                      size=(60,28),editable=False,
                                                          color=(0.3,1.0,0.3,1.0),hAlign="right",vAlign="center",
                                                          text=str(self._stressTestRoundDuration),
                          padding=2)
        b = bs.buttonWidget(parent=self._rootWidget,position=(330-xSub,v-11),size=(28,28),label="- ",
                            call=bs.Call(self._stressTestRoundDurationDecrement),
                            repeat=True,enableSound=True)
        b = bs.buttonWidget(parent=self._rootWidget,position=(380-xSub,v-11),size=(28,28),label="+ ",
                            call=bs.Call(self._stressTestRoundDurationIncrement),
                            repeat=True,enableSound=True)


        v -= 82

        b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)*0.5,v),size=(buttonWidth,60),
                            label=R.runStressTestText,call=self._stressTestPressed)

    def _stressTestPlayerCountDecrement(self):
        self._stressTestPlayerCount = max(1,self._stressTestPlayerCount - 1)
        bs.textWidget(edit=self._stressTestPlayerCountText,text=str(self._stressTestPlayerCount))

    def _stressTestPlayerCountIncrement(self):
        self._stressTestPlayerCount = self._stressTestPlayerCount + 1
        bs.textWidget(edit=self._stressTestPlayerCountText,text=str(self._stressTestPlayerCount))

    def _stressTestRoundDurationDecrement(self):
        self._stressTestRoundDuration = max(10,self._stressTestRoundDuration - 10)
        bs.textWidget(edit=self._stressTestRoundDurationText,text=str(self._stressTestRoundDuration))

    def _stressTestRoundDurationIncrement(self):
        self._stressTestRoundDuration = self._stressTestRoundDuration + 10
        bs.textWidget(edit=self._stressTestRoundDurationText,text=str(self._stressTestRoundDuration))


    def _stressTestGameTypeSelected(self,gameType):
        self._stressTestGameType = gameType

    def _unlockPressed(self):
        try:
            import bsCoopGame
            campaign = bsCoopGame.getCampaign('Default')
            levels = campaign.getLevels()
            for level in levels: campaign.completeLevel(level['name'])
            bs.screenMessage("CO-OP LEVELS UNLOCKED!",color=(0,1,0))
            bs.playSound(bs.getSound('gunCocking'))
        except Exception:
            bs.screenMessage("ERROR UNLOCKING CO-OP LEVELS",color=(1,0,0))

    def _stressTestPressed(self):
        bsUtils.runStressTest(playlistType=self._stressTestGameType,
                              playlistName=bs.textWidget(query=self._stressTestPlaylistNameField),
                              playerCount=self._stressTestPlayerCount,
                              roundDuration=self._stressTestRoundDuration)
        self._done()

    def _done(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')


class CreditsWindow(Window):
    def __init__(self):
        width = 670
        height = 490

        R = bs.getResource('creditsWindow')
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight',scale=1.53 if gSmallUI else 1.25 if gMedUI else 1.0,
                                              stackOffset=(0,-16) if gSmallUI else (0,0))

        b = bs.buttonWidget(parent=self._rootWidget,position=(40,height-62),size=(120,60),scale=0.8,
                            label=bs.getResource('backText'),buttonType='back', call=self._back)

        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(0,height-54),size=(width,30),
                          text=R.titleText,hAlign="center",color=gTitleColor,
                          vAlign="center")

        s = bs.scrollWidget(parent=self._rootWidget,position=(40,35),size=(width-80,height-100),captureArrows=True)

        def _formatNames(names,inset):
            s = ''
            spaceWidth = bs.getStringWidth(' '*10)/10.0 # measure a series since theres overlaps and stuff..
            spacing = 330
            c1 = inset
            c2 = c1+spacing
            c3 = c2+spacing
            #c4 = c3+spacing
            lineWidth = 0
            line = ''
            for name in names:
                # move to the next column (or row) and print
                if lineWidth > c3:
                    s += line+'\n'
                    line = ''
                    lineWidth = 0

                #if lineWidth > c3: target = c4
                if lineWidth > c2: target = c3
                elif lineWidth > c1: target = c2
                else: target = c1
                #print 'ADDING',(target-width)/spaceWidth,'TO GET TO',target
                spacing = ' '*int((target-lineWidth)/spaceWidth)
                line += spacing
                line += name
                lineWidth = bs.getStringWidth(line)
            if line != '': s += line+'\n'
            return s

        soundAndMusic = ("     'William Tell (Trumpet Entry)' Performed by The Apollo Symphony Orchestra\n"
                         "     Composed by Gioacchino Rossini, Arranged by Chris Worth, Published by BMI,\n"
                         "     courtesy of www.AudioSparx.com")
        names = ['HubOfTheUniverseProd', 'Jovica', 'LG', 'Leady', 'Percy Duke', 'PhreaKsAccount',
                                       'Pogotron', 'Rock Savage', 'anamorphosis', 'benboncan', 'cdrk', 'chipfork',
                                       'guitarguy1985', 'jascha', 'joedeshon', 'loofa', 'm_O_m', 'mich3d', 'sandyrb',
                                       'shakaharu', 'sirplus', 'stickman', 'thanvannispen', 'virotic', 'zimbot']
        names.sort(key=lambda x:x.lower())

        freesoundNames = _formatNames(names,90)

        creditsText = ('  Coding, Graphics, and Audio by Eric Froemling\n'
                       '  \n'
                       '  Additional Audio, Early Artwork, and Ideas by Raphael Suter\n'
                       '  \n'
                       '  Sound & Music:\n'
                       '\n'
                       '${SOUND_AND_MUSIC}\n'
                       '\n'
                       '     Public-domain music via Musopen.com (a great site for classical music)\n'
                       '        Thanks especially to the US Army, Navy, and Marine Bands.\n'
                       '     Additional music from the YouTube Audio Library\n'
                       '\n'
                       '     Freesound.org Sounds:\n\n'
                       '${FREESOUND_NAMES}\n'
                       '\n'
                       '  Language Translations:\n\n'
                       '${TRANSLATION_NAMES}\n\n'
                       '  Special Thanks:\n'
                       '\n'
                       '     Todd, Laura, and Robert Froemling\n'
                       '     All of my friends and family who helped play test\n'
                       '     Whoever invented coffee\n'
                       '\n'
                       '  Legal:\n'
                       '\n'
                       '${LEGAL_STUFF}\n'
                       '\n'
                       '                                                             www.froemling.net\n')

        translationNames = _formatNames(bsServerData.translationContributors,60)

        txt = creditsText
        txt = txt.replace("${SOUND_AND_MUSIC}",soundAndMusic)
        txt = txt.replace("${FREESOUND_NAMES}",freesoundNames)
        txt = txt.replace("${TRANSLATION_NAMES}",translationNames)
        txt = txt.replace("${LEGAL_STUFF}","     This software is based in part on the work of the Khronos Group.")

        scale = 0.55
        self._subWidth = width-80
        self._subHeight = bs.getStringHeight(txt)*scale + 40
        c = self._subContainer = bs.containerWidget(parent=s,size=(self._subWidth,self._subHeight),singleDepth=True,background=False,claimsLeftRight=False,claimsTab=False)

        t = bs.textWidget(parent=c,
                          padding=4,
                          color=(0.7,0.9,0.7,1.0),
                          scale=scale,
                          size=(0,0),
                          position=(0,self._subHeight-20),
                          hAlign='left',
                          vAlign='top',
                          text=txt
                          )

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = MainMenuWindow(transition='inLeft').getRootWidget()


class HelpWindow(Window):

    def __init__(self,mainMenu=False):

        R = bs.getResource('helpWindow')
        self._mainMenu = mainMenu
        width = 750
        height = 500 if gSmallUI else 550 if gMedUI else 600

        global gCanAskToConfigGamePads
        gCanAskToConfigGamePads = False
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight',scale=1.5 if gSmallUI else 1.25 if gMedUI else 1.0)

        t = bs.textWidget(parent=self._rootWidget,position=(0,height-(58 if gSmallUI else 45)),size=(width,25),text=R.titleText,color=gTitleColor,
                          hAlign="center",vAlign="top")

        self._scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(40,25 if gSmallUI else 65),
                                             size=(width-87,height-120+(35 if gSmallUI else -10)),captureArrows=True)
        bs.containerWidget(edit=self._rootWidget,selectedChild=self._scrollWidget)

        # ugly: create this last so it gets first dibs at touch events (since we have it close to the scroll widget)
        b = bs.buttonWidget(parent=self._rootWidget,position=(40+0 if gSmallUI else 70,
                                                              height-(59 if gSmallUI else 50)),size=(120,60),scale=0.7 if gSmallUI else 0.8,
                            label=bs.getResource('backText') if self._mainMenu else "Close",buttonType='back' if self._mainMenu else None,
                            extraTouchBorderScale=2.0,
                            call=self._close)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        interfaceType = bs.getInterfaceType()

        self._subWidth = 660
        self._subHeight = 1560
        c = self._subContainer = bs.containerWidget(parent=self._scrollWidget,size=(self._subWidth,self._subHeight),singleDepth=True,background=False,claimsLeftRight=False,claimsTab=False)

        spacing = 1.0

        h = self._subWidth*0.5
        v = self._subHeight-55

        logoTex = bs.getTexture('logo')
        iconBuffer = 1.1

        header = (0.7,1.0,0.7,1.0)
        header2 = (0.8,0.8,1.0,1.0)
        paragraph = (0.8,0.8,1.0,1.0)

        txt = R.welcomeText
        txtScale = 1.4
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=txtScale,
                          text=txt,hAlign="center",color=header,vAlign="center")
        iconSize = 70
        h2 = h - (bs.getStringWidth(txt)*txtScale*0.5+iconSize*0.5*iconBuffer)
        i = bs.imageWidget(parent=c,size=(iconSize,iconSize),position=(h2-0.5*iconSize,v-0.5*iconSize),texture=logoTex)

        v -= spacing * 50.0
        txtScale = R.someDaysTextScale
        txt = R.someDaysText
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=1.2,maxWidth=self._subWidth*0.9,
                          text=txt,hAlign="center",color=paragraph,vAlign="center")
        v -= spacing * 20.0
        txtScale = R.orPunchingSomethingTextScale
        txt = R.orPunchingSomethingText
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=txtScale,maxWidth=self._subWidth*0.9,
                          text=txt,hAlign="center",color=paragraph,vAlign="center")
        v -= spacing * 22.0
        txtScale = 1.0
        txt = R.canHelpText
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=txtScale,
                          text=txt,hAlign="center",color=paragraph,vAlign="center")

        v -= spacing * 70.0
        txtScale = R.toGetTheMostTextScale
        txt = R.toGetTheMostText
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=txtScale,maxWidth=self._subWidth*0.9,
                          text=txt,hAlign="center",color=header,vAlign="center")

        v -= spacing * 40.0
        txtScale = R.friendsTextScale
        txt = R.friendsText
        h2 = h - 220
        t = bs.textWidget(parent=c,position=(h2,v),size=(0,0),scale=txtScale,maxWidth=100,
                          text=txt,hAlign="right",color=header,vAlign="center")

        txt = R.friendsGoodText
        txtScale = R.friendsGoodTextScale
        t = bs.textWidget(parent=c,position=(h2+10,v+8),size=(0,0),scale=txtScale,maxWidth=500,
                          text=txt,hAlign="left",color=paragraph)

        v -= spacing * 45.0
        txtScale = R.controllersTextScale
        txt = R.controllersText
        h2 = h - 220
        t = bs.textWidget(parent=c,position=(h2,v),size=(0,0),scale=txtScale,maxWidth=100,
                          text=txt,hAlign="right",color=header,vAlign="center")

        if bs.isOuyaVersion():
            txt = R.controllersInfoTextOuya
            txtScale = R.controllersInfoTextScaleOuya
        else:
            if bs.isRunningOnFantasia():
                txt = R.controllersInfoTextFantasia
                txtScale = R.controllersInfoTextScaleFantasia
            else:
                txt = R.controllersInfoTextMac
                txtScale = R.controllersInfoTextScaleMac

        t = bs.textWidget(parent=c,position=(h2+10,v+8),size=(0,0),scale=txtScale,maxWidth=500,
                          text=txt,hAlign="left",color=paragraph)

        v -= spacing * 130.0

        txt = R.controlsText
        txtScale = R.controlsTextScale
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=txtScale,
                          text=txt,hAlign="center",color=header,vAlign="center")
        iconSize = 70
        h2 = h - (bs.getStringWidth(txt)*txtScale*0.5+iconSize*0.5*iconBuffer)
        i = bs.imageWidget(parent=c,size=(iconSize,iconSize),position=(h2-0.5*iconSize,v-0.5*iconSize),texture=logoTex)

        v -= spacing * 45.0
        txtScale = R.controlsSubtitleTextScale
        txt = R.controlsSubtitleText
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=txtScale,maxWidth=self._subWidth*0.9,
                          text=txt,hAlign="center",color=paragraph,vAlign="center")

        v -= spacing * 160.0

        sep = 70
        iconSize = 100
        iconSize2 = 30
        icon2Offs = 55
        icon2Offs2 = 40
        ouya = bs.isOuyaVersion()
        h2 = h - sep
        v2 = v
        i = bs.imageWidget(parent=c,size=(iconSize,iconSize),position=(h2-0.5*iconSize,v2-0.5*iconSize),
                           texture=bs.getTexture('buttonPunch'),color=(0.2,0.6,1) if ouya else (1,0.7,0.3))
        txtScale = R.punchInfoTextScale
        txt = R.punchInfoText
        t = bs.textWidget(parent=c,position=(h-sep-185+70,v+120),size=(0,0),scale=txtScale,
                          text=txt,hAlign="center",color=(0.3,0.65,1,1) if ouya else (1,0.7,0.3,1.0),vAlign="top")
        if ouya:
            bs.imageWidget(parent=c,size=(iconSize2,iconSize2),position=(h-sep-185+70-66,v+107),
                           texture=bs.getTexture('ouyaUButton'))

        h2 = h + sep
        v2 = v
        i = bs.imageWidget(parent=c,size=(iconSize,iconSize),position=(h2-0.5*iconSize,v2-0.5*iconSize),
                           texture=bs.getTexture('buttonBomb'),color=(1,0.3,0.3))

        txt = R.bombInfoText
        txtScale = R.bombInfoTextScale
        t = bs.textWidget(parent=c,position=(h+sep+50+60,v-35),size=(0,0),scale=txtScale,
                          text=txt,hAlign="center",color=(1,0.3,0.3,1.0),vAlign="top")
        if ouya:
            bs.imageWidget(parent=c,size=(iconSize2,iconSize2),position=(h+sep+50-5,v-48),
                           texture=bs.getTexture('ouyaAButton'))
        h2 = h
        v2 = v + sep
        i = bs.imageWidget(parent=c,size=(iconSize,iconSize),position=(h2-0.5*iconSize,v2-0.5*iconSize),
                           texture=bs.getTexture('buttonPickUp'),color=(1,0.8,0.3) if ouya else (0.8,0.5,1))
        txt = R.pickUpInfoText
        txtScale = R.pickUpInfoTextScale
        t = bs.textWidget(parent=c,position=(h+60+120,v+sep+50),size=(0,0),scale=txtScale,
                          text=txt,hAlign="center",color=(1,0.8,0.3,1) if ouya else (0.8,0.5,1,1.0),vAlign="top")
        if ouya:
            bs.imageWidget(parent=c,size=(iconSize2,iconSize2),position=(h+60+48,v+sep+38),
                           texture=bs.getTexture('ouyaYButton'))
        h2 = h
        v2 = v - sep
        i = bs.imageWidget(parent=c,size=(iconSize,iconSize),position=(h2-0.5*iconSize,v2-0.5*iconSize),
                           texture=bs.getTexture('buttonJump'),color=(0.4,1,0.4))
        txt = R.jumpInfoText
        txtScale = R.jumpInfoTextScale
        t = bs.textWidget(parent=c,position=(h-250+75,v-sep-15+30),size=(0,0),scale=txtScale,
                          text=txt,hAlign="center",color=(0.4,1,0.4,1.0),vAlign="top")
        if ouya:
            bs.imageWidget(parent=c,size=(iconSize2,iconSize2),position=(h-250+13,v-sep+3),
                           texture=bs.getTexture('ouyaOButton'))

        txt = R.runInfoText
        txtScale = R.runInfoTextScale
        t = bs.textWidget(parent=c,position=(h,v-sep-100),size=(0,0),scale=txtScale,maxWidth=self._subWidth*0.9,
                          text=txt,hAlign="center",color=(0.7,0.7,1.0,1.0),vAlign="center")

        v -= spacing * 280.0


        txt = R.powerupsText
        txtScale = R.powerupsTextScale
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=txtScale,
                          text=txt,hAlign="center",color=header,vAlign="center")
        iconSize = 70
        h2 = h - (bs.getStringWidth(txt)*txtScale*0.5+iconSize*0.5*iconBuffer)
        i = bs.imageWidget(parent=c,size=(iconSize,iconSize),position=(h2-0.5*iconSize,v-0.5*iconSize),texture=logoTex)

        v -= spacing * 50.0
        txtScale = R.powerupsSubtitleTextScale
        txt = R.powerupsSubtitleText
        t = bs.textWidget(parent=c,position=(h,v),size=(0,0),scale=txtScale,maxWidth=self._subWidth*0.9,
                          text=txt,hAlign="center",color=paragraph,vAlign="center")

        v -= spacing * 1.0

        m1 = -270
        m2 = -215
        m3 = 0
        iconSize = 50
        shadowSize = 80
        shadowOffsX = 3
        shadowOffsY = -4
        tBig = 1.1
        tSmall = 0.65

        shadowTex = bs.getTexture('shadowSharp')

        for tex in ['powerupPunch',
                    'powerupShield',
                    'powerupBomb',
                    'powerupHealth',
                    'powerupIceBombs',
                    'powerupImpactBombs',
                    'powerupStickyBombs',
                    'powerupLandMines',
                    'powerupCurse']:
            name = R[tex+'NameText']
            desc = R[tex+'DescriptionText']
            
            v -= spacing * 60.0

            i = bs.imageWidget(parent=c,size=(shadowSize,shadowSize),position=(h+m1+shadowOffsX-0.5*shadowSize,v+shadowOffsY-0.5*shadowSize),
                               texture=shadowTex,color=(0,0,0),opacity=0.5)
            i = bs.imageWidget(parent=c,size=(iconSize,iconSize),position=(h+m1-0.5*iconSize,v-0.5*iconSize),
                               texture=bs.getTexture(tex))

            txtScale = tBig
            txt = name
            t = bs.textWidget(parent=c,position=(h+m2,v+3),size=(0,0),scale=txtScale,maxWidth=200,
                              text=txt,hAlign="left",color=header2,vAlign="center")
            txtScale = tSmall
            txt = desc
            t = bs.textWidget(parent=c,position=(h+m3,v),size=(0,0),scale=txtScale,maxWidth=300,
                              text=txt,hAlign="left",color=paragraph,vAlign="center")

    def _close(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        if self._mainMenu:
            uiGlobals['mainMenuWindow'] = MainMenuWindow(transition='inLeft').getRootWidget()
        global gCanAskToConfigGamePads
        gCanAskToConfigGamePads = True


class ConfigControlsWindow(Window):

    def __init__(self,transition='inRight'):

        self._haveSelectedChild = False

        R = bs.getResource('configControllersWindow')

        ua = bs.getUserAgentString();
        interfaceType = bs.getInterfaceType()
        isOuya = bs.isOuyaVersion()
        isFantasia = bs.isRunningOnFantasia()

        spacing = 50
        buttonWidth = 350

        width = 460
        height = 135

        spaceHeight = spacing*0.3


        showGamepads = False
        if 'Mac' in ua or 'linux' in ua or 'windows' in ua or 'android' in ua or 'macos' in ua:
            showGamepads = True
            height += spacing

        showTouch = False
        if bs.haveTouchScreenInput():
            showTouch = True
            height += spacing

        showSpace1 = False
        if showGamepads or showTouch:
            showSpace1 = True
            height += spaceHeight

        showKeyboards = False
        if interfaceType == 'desktop':
            showKeyboards = True
            height += spacing*2

        showSpace2 = False
        if showKeyboards:
            showSpace2 = True
            height += spaceHeight

        showRemote = False
        if True:
            showRemote = True
            height += spacing

        showPS3 = False
        if 'Mac' in ua or isOuya:
            showPS3 = True
            height += spacing

        show360 = False
        if 'Mac' in ua or isOuya or isFantasia:
            show360 = True
            height += spacing

        showMacWiimote = False
        if 'Mac' in ua:
            showMacWiimote = True
            height += spacing

        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,scale=1.75 if gSmallUI else 1.4 if gMedUI else 1.0)
        b = bs.buttonWidget(parent=self._rootWidget,position=(35,height-55),size=(120,60),scale=0.8,textScale=1.2,
                            label=bs.getResource('backText'),buttonType='back',call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        # need these vars to exist even if the buttons dont
        self._gamePadsButton = self._touchButton = self._keyboardButton = self._keyboard2Button = self._iDevicesButton = self._ps3Button = self._xbox360Button = self._wiimotesButton = -1

        t = bs.textWidget(parent=self._rootWidget,position=(0,height-49),size=(width,25),text=R.titleText,color=gTitleColor,
                          hAlign="center",vAlign="top")
        v = height - 75
        v -= spacing

        if showTouch:
            self._touchButton = b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)/2,v),size=(buttonWidth,43),
            label=R.configureTouchText,textScale=0.9,call=self._doTouchscreen)
            if not self._haveSelectedChild:
                bs.containerWidget(edit=self._rootWidget,selectedChild=self._touchButton)
                self._haveSelectedChild = True
            v -= spacing

        if showGamepads:
            self._gamePadsButton = b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)/2-7,v),size=(buttonWidth,43),
                                                 label=R.configureGamepadsText,call=self._doGamepads)
            if not self._haveSelectedChild:
                bs.containerWidget(edit=self._rootWidget,selectedChild=self._gamePadsButton)
                self._haveSelectedChild = True
            v -= spacing

        if showSpace1: v -= spaceHeight

        if showKeyboards:
            self._keyboardButton = b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)/2+5,v),size=(buttonWidth,43),
                                label=R.configureKeyboardText,call=self._configKeyboard)
            if not self._haveSelectedChild:
                bs.containerWidget(edit=self._rootWidget,selectedChild=self._keyboardButton)
                self._haveSelectedChild = True
            v -= spacing
            self._keyboard2Button = b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)/2-3,v),size=(buttonWidth,43),
                                label=R.configureKeyboard2Text,call=self._configKeyboard2)
            v -= spacing
        if showSpace2: v -= spaceHeight
        if showRemote:
            self._iDevicesButton = b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)/2-5,v),size=(buttonWidth,43),
                                label=R.configureMobileText,
                                call=self._doMobileDevices)
            if not self._haveSelectedChild:
                bs.containerWidget(edit=self._rootWidget,selectedChild=self._iDevicesButton)
                self._haveSelectedChild = True
            v -= spacing
        if showPS3:
            self._ps3Button = b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)/2+5,v),size=(buttonWidth,43),
                                label=R.ps3Text,
                                call=self._doPS3Controllers)
            v -= spacing
        if show360:
            self._xbox360Button = b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)/2-1,v),size=(buttonWidth,43),
                                label=R.xbox360Text,
                                call=self._do360Controllers)
            v -= spacing
        if showMacWiimote:
            self._wiimotesButton = b = bs.buttonWidget(parent=self._rootWidget,position=((width-buttonWidth)/2+5,v),size=(buttonWidth,43),
                                label=R.wiimotesText,
                                call=self._doWiimotes)
            v -= spacing

        # re-select previous if applicable
        selected = None
        try:
            global gControlsSelection
            selection = gControlsSelection
            if selection == 'GamePads': selected = self._gamePadsButton
            elif selection == 'Touch': selected = self._touchButton
            elif selection == 'Keyboard': selected = self._keyboardButton
            elif selection == 'Keyboard2': selected = self._keyboard2Button
            elif selection == 'iDevices': selected = self._iDevicesButton
            elif selection == 'PS3': selected = self._ps3Button
            elif selection == 'xbox360': selected = self._xbox360Button
            elif selection == 'Wiimotes': selected = self._wiimotesButton
        except Exception: pass
        if selected != None and selected != -1:
            bs.containerWidget(edit=self._rootWidget,selectedChild=selected)

    def _configKeyboard(self):
        self._storeSelection("Keyboard")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigKeyboardWindow(bs.getInputDevice('Keyboard','#1')).getRootWidget()

    def _configKeyboard2(self):
        self._storeSelection("Keyboard2")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigKeyboardWindow(bs.getInputDevice('Keyboard','#2')).getRootWidget()

    def _doMobileDevices(self):
        self._storeSelection("iDevices")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConnectMobileDevicesWindow().getRootWidget()

    def _doPS3Controllers(self):
        self._storeSelection("PS3")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConnectPS3ControllersWindow().getRootWidget()

    def _do360Controllers(self):
        self._storeSelection("xbox360")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = Connect360ControllersWindow().getRootWidget()

    def _doWiimotes(self):
        self._storeSelection("Wiimotes")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigureWiiRemotesWindow().getRootWidget()

    def _doGamepads(self):
        self._storeSelection('GamePads')
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigGamePadWindow().getRootWidget()

    def _doTouchscreen(self):
        self._storeSelection('Touch')
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigTouchscreenWindow().getRootWidget()

    def _storeSelection(self,value=None):
        # None implies to grab highlighted choice
        if value is None:
            s = bs.getWidgetSelectedChild(self._rootWidget)
            if s == self._gamePadsButton: value = 'GamePads'
            elif s == self._touchButton: value = 'Touch'
            elif s == self._keyboardButton: value = 'Keyboard'
            elif s == self._keyboard2Button: value = 'Keyboard2'
            elif s == self._iDevicesButton: value = 'iDevices'
            elif s == self._ps3Button: value = 'PS3'
            elif s == self._xbox360Button: value = 'xbox360'
            elif s == self._wiimotesButton: value = 'Wiimotes'
            else: value = 'GamePads'
        if value != None:
            global gControlsSelection
            gControlsSelection = value

    def _back(self):
        self._storeSelection()
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow']=SettingsWindow(transition='inLeft').getRootWidget()


class ColorPicker(Window):

    """ pops up a ui to select from a set of colors.
    passes the color to the delegate's colorPickerSelectedColor() method """

    cRaw = bsUtils.getPlayerColors()
    if len(cRaw) != 16: raise Exception("expected 16 player colors")
    colors = [cRaw[0:4],
              cRaw[4:8],
              cRaw[8:12],
              cRaw[12:16]]

    def getRootWidget(self):
        return self._rootWidget
    def __init__(self,parent,position,initialColor=(1,1,1),delegate=None):
        self._delegate = delegate
        self._transitioningOut = False
        self._rootWidget = bs.containerWidget(parent=parent,transition='inScale',
                                              size=(210,210),position=position,
                                              outsideClickCall=self._cancel,
                                              cancelCall=self._cancel,
                                              singleDepth=True)
        rows = []
        closestDist = 9999
        closest = (0,0)
        for y in range(4):
            row = []
            rows.append(row)
            for x in range(4):
                color = self.colors[y][x]
                dist = abs(color[0]-initialColor[0])+abs(color[1]-initialColor[1])+abs(color[2]-initialColor[2])
                if dist < closestDist:
                    closest = (x,y)
                    closestDist = dist
                b = bs.buttonWidget(parent=self._rootWidget,position=(22+45*x,155-45*y),
                                    size=(35,40),label='',buttonType='square',call=bs.WeakCall(self._select,x,y),
                                    color=color,
                                    extraTouchBorderScale=0.0)
                row.append(b)
                if y > 0:
                    bs.buttonWidget(edit=b,upWidget=rows[y-1][x])
                    bs.buttonWidget(edit=rows[y-1][x],downWidget=b)
        bs.containerWidget(edit=parent,selectedChild=self._rootWidget)
        bs.containerWidget(edit=self._rootWidget,selectedChild=rows[closest[1]][closest[0]])

    def _select(self,x,y):
        if self._delegate: self._delegate.colorPickerSelectedColor(self,self.colors[y][x])
        bs.realTimer(50,self._transitionOut)
    def _transitionOut(self):
        if not self._transitioningOut:
            self._transitioningOut = True
            if self._delegate: self._delegate.colorPickerClosing(self)
            bs.containerWidget(edit=self._rootWidget,transition='outScale')
            
    def _cancel(self):
        bs.playSound(bs.getSound('swish'))
        self._transitionOut()

class PopupMenuWindow(Window):

    def getRootWidget(self):
        return self._rootWidget

    def __init__(self,parent,position,choices,currentChoice,delegate=None,width=230,scale=1.0,choicesDisabled=[],choicesDisplay=[]):
        self._choices = list(choices)
        self._choicesDisplay = list(choicesDisplay)
        self._currentChoice=currentChoice
        self._choicesDisabled = list(choicesDisabled)
        self._doneBuilding = False
        if len(choices) < 1: raise Exception("Must pass at least one choice")
        self._width = width
        self._scale = scale
        if len(choices) > 5:
            self._height = min(52+len(choices)*29,230)
            self._useScroll = True
        else:
            self._height = min(20+len(choices)*29,230)
            self._useScroll = False 
        self._delegate = None # dont want this stuff called just yet..

        # extend width to fit our longest string
        for index,choice in enumerate(choices):
            if len(choicesDisplay) == len(choices): choiceDisplayName = choicesDisplay[index]
            else: choiceDisplayName = choice
            if self._useScroll:
                self._width = max(self._width,bs.getStringWidth(choiceDisplayName)+75)
            else:
                self._width = max(self._width,bs.getStringWidth(choiceDisplayName)+60)


        self._rootWidget = bs.containerWidget(parent=parent,transition='inScale',scale=self._scale,
                                              size=(self._width,self._height),position=(position[0]-self._width*self._scale*0.5,position[1]-self._height*self._scale*0.5),
                                              outsideClickCall=self._cancel,
                                              color=(0.35,0.55,0.15,1),
                                              cancelCall=self._cancel,
                                              singleDepth=True)
        bs.containerWidget(edit=parent,selectedChild=self._rootWidget)

        if self._useScroll:
            self._scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(20,20),size=(self._width-40,self._height-40))
            self._columnWidget = bs.columnWidget(parent=self._scrollWidget)
        else:
            self._offsetWidget = bs.containerWidget(parent=self._rootWidget,position=(30,15),size=(self._width-40,self._height),background=False)
            self._columnWidget = bs.columnWidget(parent=self._offsetWidget)
        for index,choice in enumerate(choices):
            if len(choicesDisplay) == len(choices): choiceDisplayName = choicesDisplay[index]
            else: choiceDisplayName = choice
            inactive = (choice in self._choicesDisabled)
            w = bs.textWidget(parent=self._columnWidget,size=(self._width-40,24),
                              selectCall=bs.Call(self._select,index),
                              clickActivate=True,
                              color=(0.5,0.5,0.5,0.5) if inactive else ((0.5,1,0.5,1) if choice == self._currentChoice else (0.8,0.8,0.8,1.0)),
                              padding=0,
                              text=choiceDisplayName,
                              activateAction=self._activate,
                              vAlign='bottom',selectable=False if inactive else True)
            if choice == self._currentChoice:
                bs.containerWidget(edit=self._columnWidget,selectedChild=w,visibleChild=w)

        self._delegate = delegate # ok from now on our delegate can be called
        self._doneBuilding = True

    def _select(self,index):
        if self._doneBuilding:
            self._currentChoice = self._choices[index]

    def _activate(self):
        bs.playSound(bs.getSound('swish'))
        bs.realTimer(50,self._transitionOut)
        if self._delegate:
            # call this in a timer so it doesnt interfere with us killing our widgets and whatnot..
            bs.realTimer(0,bs.Call(self._delegate.popupMenuSelectedChoice,self,self._currentChoice))

    def _transitionOut(self):
        self._delegate.popupMenuClosing(self)
        bs.containerWidget(edit=self._rootWidget,transition='outScale')

    def _cancel(self):
        bs.playSound(bs.getSound('swish'))
        self._transitionOut()

class PopupMenu(object):
    def __init__(self,parent,position,choices,currentChoice=None,changeCall=None,openingCall=None,closingCall=None,width=230,scale=1.0,choicesDisabled=[],choicesDisplay=[],buttonSize=(160,50)):
        if currentChoice not in choices: currentChoice = None
        self._choices = list(choices)
        if len(choices) == 0: raise Exception("no choices given")
        self._choicesDisplay = list(choicesDisplay)
        self._choicesDisabled = list(choicesDisabled)
        self._width = width
        self._scale = scale
        self._currentChoice = currentChoice if currentChoice is not None else self._choices[0]
        self._position = position
        self._parent = parent
        if len(choices) < 1: raise Exception("Must pass at least one choice")
        self._parent = parent
        self._buttonSize = buttonSize
        self._button = bs.buttonWidget(parent=self._parent,position=(self._position[0],self._position[1]),
                                       size=self._buttonSize,scale=1.0,label='',call=bs.Call(bs.realTimer,0,self._makePopup))
        self._changeCall = None # dont wanna call it for initial set
        self._openingCall = openingCall
        self._closingCall = closingCall
        self.setChoice(self._currentChoice)
        self._changeCall = changeCall
        self._changeCall = changeCall
        self._windowWidget = None

    def _makePopup(self):
        if self._openingCall: self._openingCall()
        self._windowWidget = PopupMenuWindow(parent=self._parent,position=(self._position[0]+self._buttonSize[0]*0.5,
                                                                     self._position[1]+self._buttonSize[1]*0.5),delegate=self,
                                       width=self._width,scale=self._scale,choices=self._choices,currentChoice=self._currentChoice,
                                       choicesDisabled=self._choicesDisabled,choicesDisplay=self._choicesDisplay).getRootWidget()
    def getButtonWidget(self):
        return self._button

    def getWindowWidget(self):
        return self._windowWidget

    def popupMenuSelectedChoice(self,popupWindow,choice):
        self.setChoice(choice)
        if self._changeCall: self._changeCall(choice)

    def popupMenuClosing(self,popupWindow):
        if bs.widgetExists(self._button):
            bs.containerWidget(edit=self._parent,selectedChild=self._button)
        self._windowWidget = None
        if self._closingCall: self._closingCall()

    def setChoice(self,choice):
        self._currentChoice = choice
        if len(self._choicesDisplay) == len(self._choices):
            displayName = self._choicesDisplay[self._choices.index(choice)]
        else: displayName = choice
        if bs.widgetExists(self._button):
            bs.buttonWidget(edit=self._button,label=displayName)


class EditProfileWindow(Window):

    def __init__(self,existingProfile,inMainMenu,transition='inRight'):

        import _bs
        self._inMainMenu=inMainMenu
        self._existingProfile = existingProfile
        self._colorPickers = {}

        self._R = R = bs.getResource('editProfileWindow')

        # grab profile colors or pick random ones
        self._color,self._highlight = bsUtils.getPlayerProfileColors(existingProfile)
        width = 480
        height = 400
        spacing = 40
        buttonWidth = 350
        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,scale=1.62 if gSmallUI else 1.3 if gMedUI else 1.0)
        b = bs.buttonWidget(parent=self._rootWidget,position=(21,height-60),size=(155,60),scale=0.8,
                            label=bs.getResource('cancelText'),call=self._cancel)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        saveButton = b = bs.buttonWidget(parent=self._rootWidget,position=(width-147,height-60),size=(155,60),scale=0.8,label=bs.getResource('saveText'))
        bs.containerWidget(edit=self._rootWidget,startButton=b)
        t = bs.textWidget(parent=self._rootWidget,position=(0,height-50),size=(width,25),
                          text=(R.titleNewText if existingProfile is None else R.titleEditText),
                          color=gTitleColor,maxWidth=170,
                          scale=0.8,hAlign="center",vAlign="top")
        v = height - 130

        try: sessionType = bs.getSession().getSessionType()
        except Exception:
            print 'bsUI: error getting session type during profile edit'
            sessionType = 'unknown'
        if sessionType != 'mainMenu' and existingProfile is not None:
            bs.textWidget(parent=self._rootWidget,position=(0,height-90),size=(width,25),
                          text=R.changesNotAffectText,maxWidth=width*0.8,color=(0,1,0,1),
                          scale=0.48,hAlign="center",vAlign="center")

        # make a list of spaz icons
        self._spazzes = bsSpaz.getAppearances()
        self._spazzes.sort()
        self._iconTextures = [bs.getTexture(bsSpaz.appearances[s].iconTexture,minQuality='High') for s in self._spazzes]
        self._iconTintTextures = [bs.getTexture(bsSpaz.appearances[s].iconMaskTexture) for s in self._spazzes]
        try: profile = bs.config['Player Profiles'][self._existingProfile]
        except Exception: profile = {}

        # look for existing character choice or pick random one otherwise
        try: iconIndex = self._spazzes.index(profile['character'])
        except Exception:
            # lets set the default icon to spaz for our first profile; after that we go random
            try: pLen = len(bs.config['Player Profiles'])
            except Exception: pLen = 0
            if pLen == 0: iconIndex = self._spazzes.index('Spaz')
            else:
                random.seed()
                iconIndex = random.randrange(len(self._spazzes))
        self._iconIndex = iconIndex
        bs.buttonWidget(edit=saveButton,call=self._save)
        bs.textWidget(parent=self._rootWidget,text=R.nameText,position=(5,v),size=(100,30),hAlign='right',vAlign='center',color=(1,1,1,0.5),scale=0.75)

        self._textField = bs.textWidget(parent=self._rootWidget,position=(110,v),size=(265,30),
                                        text='' if self._existingProfile is None else self._existingProfile,hAlign="left",
                                        vAlign="center",
                                        maxChars=16,
                                        description=R.nameDescriptionText,
                                        editable=True,padding=4,
                                        color=(0.9,0.9,0.9,1.0),
                                        useReturnAction=True,
                                        returnAction=bs.WeakCall(bs.activateWidget,saveButton))

        bs.buttonWidget(parent=self._rootWidget,label=R.randomText,size=(150,50),scale=0.55,position=(385,v+4),
                        color=(0.45,0.4,0.5),textColor=(0.65,0.6,0.7),call=self._randomName)

        v -= spacing * 2.5

        # auto-select text on desktops since there's no onscreen keyboard to initially obscure stuff
        if bs.getInterfaceType() == 'desktop':
            bs.containerWidget(edit=self._rootWidget,selectedChild=self._textField)

        insetTex = bs.getTexture('frameInset')

        bs.buttonWidget(parent=self._rootWidget,position=(146-21,v+5),size=(50,50),
                        label="",texture=bs.getTexture('leftButton'),
                        call=bs.WeakCall(self._updateCharacter,-1),repeat=True)

        self._iconImage = bs.imageWidget(parent=self._rootWidget,position=(188,v-18),size=(100,100),color=(1,1,1),hasAlphaChannel=False)
        bs.imageWidget(parent=self._rootWidget,position=(182,v-24),size=(113,113),color=(0.4,0.37,0.49),
                       texture=insetTex,modelTransparent=bs.getModel('frameInset'))

        bs.buttonWidget(parent=self._rootWidget,position=(304-5,v+5),size=(50,50),
                        texture=bs.getTexture('rightButton'),
                        label="",call=bs.WeakCall(self._updateCharacter,1),repeat=True)

        self._charNameWidget = bs.textWidget(parent=self._rootWidget,text='',position=(width/2-54,v-50),size=(100,30),scale=0.5,hAlign='center',color=(0.7,0.9,0.7,0.4))

        x = width/2-73
        y = v - 110
        self._colorButton = b = bs.buttonWidget(parent=self._rootWidget,position=(x,y+10),size=(40,40),color=self._color,label='',call=bs.WeakCall(self._makePicker,'color',x,y),buttonType='square')
        bs.textWidget(parent=self._rootWidget,hAlign='center',vAlign='center',position=(x-5,y-25),text=R.colorText,scale=0.7,color=gTitleColor,maxWidth=90)

        x = width/2+30
        y = v - 110
        self._highlightButton = b = bs.buttonWidget(parent=self._rootWidget,position=(x,y+10),size=(40,40),color=self._highlight,label='',call=bs.WeakCall(self._makePicker,'highlight',x,y),buttonType='square')
        bs.textWidget(parent=self._rootWidget,hAlign='center',vAlign='center',position=(x-5,y-25),text=R.highlightText,scale=0.7,color=gTitleColor,maxWidth=90)
        self._updateCharacter()


    def _randomName(self):
        names = bs.getRandomNames()
        bs.textWidget(edit=self._textField,text=names[random.randrange(len(names))])

    def _makePicker(self,pickerType,x,y):
        # need to call this in a timer; can screw things up calling it from within a mouse-up
        def doIt():
            if pickerType == 'color': initialColor = self._color
            elif pickerType == 'highlight': initialColor = self._highlight
            else: raise Exception("invalid pickerType: "+pickerType)
            self._colorPickers[pickerType] = weakref.ref(ColorPicker(parent=self._rootWidget,position=(x-90,y-70),
                                                                     initialColor=initialColor,delegate=self))
        bs.realTimer(0,doIt)

    def _cancel(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigPlayerProfilesWindow('inLeft',selectedProfile=self._existingProfile,inMainMenu=self._inMainMenu).getRootWidget()

    def _setColor(self,color):
        self._color = color
        bs.buttonWidget(edit=self._colorButton,color=color)

    def _setHighlight(self,color):
        self._highlight = color
        bs.buttonWidget(edit=self._highlightButton,color=color)
        
    def colorPickerClosing(self,picker):
        for pType,pRef in self._colorPickers.items():
            p = pRef()
            if picker == p:
                if pType == 'color':
                    bs.containerWidget(edit=self._rootWidget,selectedChild=self._colorButton)
                elif pType == 'highlight':
                    bs.containerWidget(edit=self._rootWidget,selectedChild=self._highlightButton)

    def colorPickerSelectedColor(self,picker,color):
        for pType,pRef in self._colorPickers.items():
            p = pRef()
            if picker == p:
                if pType == 'color':
                    self._setColor(color)
                elif pType == 'highlight':
                    self._setHighlight(color)
                self._updateCharacter()

    def _updateCharacter(self,change=0):
        self._iconIndex = (self._iconIndex + change)%len(self._spazzes)
        bs.imageWidget(edit=self._iconImage,
                       texture=self._iconTextures[self._iconIndex],
                       tintTexture=self._iconTintTextures[self._iconIndex],
                       tintColor=self._color,
                       tint2Color=self._highlight)
        bs.textWidget(edit=self._charNameWidget,text="- "+bs.translate('characterNames',self._spazzes[self._iconIndex])+" -")

    def _save(self):
        newName = bs.textWidget(query=self._textField)
        if len(newName) == 0:
            bs.screenMessage(bs.getResource('nameNotEmptyText'))
            bs.playSound(bs.getSound('error'))
            return
        bs.playSound(bs.getSound('gunCocking'))
        if 'Player Profiles' not in bs.config: bs.config['Player Profiles'] = {}

        # delete old in case we're renaming
        if self._existingProfile:
            try: del(bs.config['Player Profiles'][self._existingProfile])
            except Exception: pass
        bs.config['Player Profiles'][newName] = {'character':self._spazzes[self._iconIndex],
                                                 'color':self._color,
                                                 'highlight':self._highlight}
        bs.writeConfig()
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigPlayerProfilesWindow('inLeft',selectedProfile=newName,inMainMenu=self._inMainMenu).getRootWidget()


class ConfigPlayerProfilesWindow(Window):

    def __init__(self,transition='inRight',inMainMenu=True,selectedProfile=None):
        self._inMainMenu = inMainMenu
        if self._inMainMenu: backLabel = bs.getResource('backText')
        else: backLabel = 'Done'
        self._width = 410
        self._height = 360 if gSmallUI else 380 if gMedUI else 410
        spacing = 40
        buttonWidth = 350

        self._R = R = bs.getResource('playerProfilesWindow')

        self._rootWidget = bs.containerWidget(size=(self._width,self._height),transition=transition,scale=1.8 if gSmallUI else 1.4 if gMedUI else 1.0)

        b = bs.buttonWidget(parent=self._rootWidget,position=(30,self._height-55),size=(120,60),scale=0.8,
                            label=backLabel,buttonType='back' if backLabel==bs.getResource('backText') else None,
                            call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(20,self._height-46),size=(self._width,25),
                          text=R.titleText,maxWidth=180,color=gTitleColor,scale=0.85,hAlign="center",vAlign="center")
        v = self._height - 100

        if gSmallUI: v += 50
        elif gMedUI: v += 25

        scrollHeight = 190 if gSmallUI else 215 if gMedUI else 240
        self._scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(42,v-210),size=(self._width-80,scrollHeight))
        bs.containerWidget(edit=self._rootWidget,selectedChild=self._scrollWidget)

        self._columnWidget = bs.columnWidget(parent=self._scrollWidget,printListExitInstructions=True)
        v -= 255

        self._profiles = None
        self._selectedProfile = selectedProfile
        self._profileWidgets = []

        h = 50
        hspacing = 13
        bColor = (0.6,0.53,0.63)
        newButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(90,30),
                                        call=self._newProfile,
                                        color=bColor,
                                        textColor=(0.75,0.7,0.8),
                                        textScale=0.7,
                                        label=R.newButtonText)
        h += 90+hspacing
        self._editButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(90,30),
                            call=self._editProfile,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=R.editButtonText)
        h += 90+hspacing
        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(110,30),
                            call=self._deleteProfile,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=R.deleteButtonText)

        self._refresh()

        if len(self._profileWidgets) == 0:
            bs.containerWidget(edit=self._rootWidget,selectedChild=newButton)

    def _newProfile(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = EditProfileWindow(existingProfile=None,inMainMenu=self._inMainMenu).getRootWidget()

    def _deleteProfile(self):
        if self._selectedProfile is None: return
        ConfirmWindow(self._R.deleteConfirmText.replace('${PROFILE}',self._selectedProfile),self._doDeleteProfile,350)

    def _doDeleteProfile(self):
        try: del bs.config['Player Profiles'][self._selectedProfile]
        except Exception: pass
        bs.writeConfig()
        bs.playSound(bs.getSound('shieldDown'))
        self._refresh()
        bs.containerWidget(edit=self._rootWidget,selectedChild=self._scrollWidget) # select profile list

    def _editProfile(self):
        if self._selectedProfile is None: return
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = EditProfileWindow(self._selectedProfile,inMainMenu=self._inMainMenu).getRootWidget()

    def _select(self,name,index):
        self._selectedProfile = name

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        if self._inMainMenu:
            uiGlobals['mainMenuWindow'] = SettingsWindow(transition='inLeft').getRootWidget()

    def _refresh(self):
        oldSelection = self._selectedProfile
        # delete old
        while len(self._profileWidgets) > 0: bs.deleteWidget(self._profileWidgets.pop())
        try: self._profiles = bs.config['Player Profiles']
        except Exception: self._profiles = {}
        items = self._profiles.items()
        items.sort(key=lambda x:x[0].lower())
        index = 0
        for pName,p in items:
            color,highlight = bsUtils.getPlayerProfileColors(pName)
            w = bs.textWidget(parent=self._columnWidget,position=(0,32),size=(self._width-40,24),
                              text=pName,hAlign='left',vAlign='bottom',
                              selectCall=bs.WeakCall(self._select,pName,index),
                              color=bsUtils.getSafeDisplayColor(color)+(1,),
                              alwaysHighlight=True,
                              activateAction=bs.WeakCall(bs.activateWidget,self._editButton),
                              selectable=True)
            self._profileWidgets.append(w)

            # select this one if it was previously selected
            if pName == oldSelection:
                bs.columnWidget(edit=self._columnWidget,selectedChild=w,visibleChild=w)

            index += 1

        # if there's a team-chooser in existence, tell it the profile-list has probably changed
        session = bs.getSession()
        if session: session.handleMessage(bs.PlayerProfilesChangedMessage())

class ConfigureWiiRemotesWindow(Window):

    def __init__(self):

        self._R = bs.getResource('wiimoteSetupWindow')
        width = 600
        height = 480
        spacing = 40

        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight')

        b = bs.buttonWidget(parent=self._rootWidget,position=(55,height-50),size=(120,60),scale=0.8,
                            label=bs.getResource('backText'),buttonType='back',call=self._back)

        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(0,height-35),size=(width,25),
                          text=self._R.titleText,
                          maxWidth=270,
                          color=gTitleColor,
                          hAlign="center",vAlign="center")
        v = height - 60
        v -= spacing
        t = bs.textWidget(parent=self._rootWidget,position=(15,v-90),size=(width-30,30),
                          color=(0.7,0.9,0.7,1.0),
                          scale=self._R.macInstructionsTextScale,
                          text=self._R.macInstructionsText,
                          maxWidth=width*0.95,
                          hAlign="center",vAlign="center")
        v -= 230
        buttonWidth = 200
        v -= 30
        b = bs.buttonWidget(parent=self._rootWidget,position=(width/2-buttonWidth/2,v+1),
                            size=(buttonWidth,50),label=self._R.listenText,call=WiimoteListenWindow)
        bs.containerWidget(edit=self._rootWidget,startButton=b)
        v -= spacing * 1.1
        t = bs.textWidget(parent=self._rootWidget,position=(15,v-8),size=(width-30,30),
                          color=(0.7,0.9,0.7,1.0),
                          scale=self._R.thanksTextScale,
                          maxWidth=width*0.95,
                          text=self._R.thanksText.strip(),
                          hAlign="center",vAlign="center")
        v -= 30
        thisButtonWidth = 200
        b = bs.buttonWidget(parent=self._rootWidget,position=(width/2-thisButtonWidth/2,v-14),
                            color=(0.45,0.4,0.5),
                            size=(thisButtonWidth,15),
                            label=self._R.copyrightText,
                            textColor=(0.55,0.5,0.6),
                            textScale=self._R.copyrightTextScale,
                            call=WiimoteLicenseWindow)

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()


class WiimoteListenWindow(Window):

    def __init__(self):
        import _bs

        self._R = bs.getResource('wiimoteListenWindow')
        width = 650
        height = 210
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight')

        b = bs.buttonWidget(parent=self._rootWidget,position=(35,height-60),size=(140,60),label=bs.getResource('cancelText'),scale=0.8,call=self._dismiss)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        _bs._startListeningForWiiRemotes()
        self._wiimoteConnectCounter = 15

        global dismissWiiRemotesWindow
        dismissWiiRemotesWindow = bs.WeakCall(self._dismiss)

        t = bs.textWidget(parent=self._rootWidget,position=(15,height-55),size=(width-30,30),
                          text=self._R.listeningText,color=gTitleColor,
                          maxWidth=320,
                          hAlign="center",vAlign="center")
        t = bs.textWidget(parent=self._rootWidget,position=(15,height-110),size=(width-30,30),
                          scale=self._R.pressTextScale,
                          text=self._R.pressText,
                          maxWidth=width*0.9,
                          color=(0.7,0.9,0.7,1.0),
                          hAlign="center",vAlign="center")
        t = bs.textWidget(parent=self._rootWidget,position=(15,height-140),size=(width-30,30),
                          color=(0.7,0.9,0.7,1.0),
                          scale=self._R.pressText2Scale,
                          text=self._R.pressText2,
                          maxWidth=width*0.95,
                          hAlign="center",vAlign="center")
        self._counterText = t = bs.textWidget(parent=self._rootWidget,position=(15,23),size=(width-30,30),
                                              scale=1.2,
                                              text=("15"),
                                              hAlign="center",vAlign="top")

        for i in range(1,15):
            bs.realTimer(1000*i,bs.WeakCall(self._decrement))

        bs.realTimer(15000,bs.WeakCall(self._dismiss))

    def _decrement(self):
        try:
            self._wiimoteConnectCounter -= 1
            bs.textWidget(edit=self._counterText,text=str(self._wiimoteConnectCounter))
        except Exception: pass

    def _dismiss(self):
        try:
            bs.containerWidget(edit=self._rootWidget,transition='outLeft')
            bs._bs._stopListeningForWiiRemotes()
        except Exception: pass


class WiimoteLicenseWindow(Window):

    def __init__(self):
        self._R = bs.getResource('wiimoteLicenseWindow')
        width = 750
        height = 550
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight')
        b = bs.buttonWidget(parent=self._rootWidget,position=(65,height-50),size=(120,60),scale=0.8,
                            label=bs.getResource('backText'),buttonType='back',call=self._close)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        t = bs.textWidget(parent=self._rootWidget,position=(0,height-48),size=(width,30),
                          text=self._R.titleText,hAlign="center",color=gTitleColor,vAlign="center")
        licenseText = ('Copyright (c) 2007, DarwiinRemote Team\n'
                       'All rights reserved.\n'
                       '\n'
                       '   Redistribution and use in source and binary forms, with or without modification,\n'
                       '   are permitted provided that the following conditions are met:\n'
                       '\n'
                       '1. Redistributions of source code must retain the above copyright notice, this\n'
                       '     list of conditions and the following disclaimer.\n'
                       '2. Redistributions in binary form must reproduce the above copyright notice, this\n'
                       '     list of conditions and the following disclaimer in the documentation and/or other\n'
                       '     materials provided with the distribution.\n'
                       '3. Neither the name of this project nor the names of its contributors may be used to\n'
                       '     endorse or promote products derived from this software without specific prior\n'
                       '     written permission.\n'
                       '\n'
                       'THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"\n'
                       'AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE\n'
                       'IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE\n'
                       'ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE\n'
                       'LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR\n'
                       'CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF\n'
                       ' SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS\n'
                       'INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN\n'
                       'CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)\n'
                       'ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE\n'
                       'POSSIBILITY OF SUCH DAMAGE.\n')
        licenseTextScale = 0.62
        t = bs.textWidget(parent=self._rootWidget,position=(-70,22),size=(width-40,height-80),
                          hAlign="left",vAlign="center",padding=4,
                          color=(0.7,0.9,0.7,1.0),
                          scale=licenseTextScale,
                          text=licenseText)
    def _close(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')


class ConnectPS3ControllersWindow(Window):

    def __init__(self):
        width = 760
        height = 330 if (bs.isOuyaVersion() or bs.isRunningOnFantasia()) else 540
        spacing = 40
        self._R = bs.getResource('ps3ControllersWindow')
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight',scale=1.4 if gSmallUI else 1.3 if gMedUI else 1.0)

        b = bs.buttonWidget(parent=self._rootWidget,position=(40,height-70),size=(120,60),scale=0.85,
                            label=bs.getResource('backText'),buttonType='back',
                            call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(20,height-61),size=(width,25),text=self._R.titleText,color=gTitleColor,
                          hAlign="center",vAlign="top")
        v = height - 90
        v -= spacing

        if bs.isOuyaVersion() or bs.isRunningOnFantasia():
            t = bs.textWidget(parent=self._rootWidget,position=(15,v+15),size=(width-30,30),
                              color=(0.7,0.9,0.7,1.0),
                              maxWidth=width*0.95,
                              scale=self._R.ouyaInstructionsTextScale,
                              text=self._R.ouyaInstructionsText,
                              hAlign="center",vAlign="top")
        else:
            t = bs.textWidget(parent=self._rootWidget,position=(15,v+15),size=(width-30,30),
                              color=(0.7,0.9,0.7,1.0),
                              maxWidth=width*0.95,
                              scale=self._R.macInstructionsTextScale,
                              text=self._R.macInstructionsText,
                              hAlign="center",vAlign="top")
            bs.buttonWidget(parent=self._rootWidget,position=(225,v-176),size=(300,40),
                            label=self._R.pairingTutorialText,
                            call=bs.WeakCall(bs.openURL,'http://www.youtube.com/watch?v=IlR_HxeOQpI&feature=related'))
    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()


class Connect360ControllersWindow(Window):

    def __init__(self):
        self._R = bs.getResource('xbox360ControllersWindow')
        width = 700
        height = 300 if (bs.isOuyaVersion() or bs.isRunningOnFantasia()) else 485
        spacing = 40
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight',scale=1.4 if gSmallUI else 1.4 if gMedUI else 1.0)

        b = bs.buttonWidget(parent=self._rootWidget,position=(35,height-65),size=(120,60),scale=0.84,
                            label=bs.getResource('backText'),buttonType='back',
                            call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(30,height-56),size=(width,25),scale=0.85*3.0,
                          text=self._R.titleText,color=gTitleColor,maxWidth=450,
                          hAlign="center",vAlign="center")
        v = height - 70
        v -= spacing

        if bs.isOuyaVersion() or bs.isRunningOnFantasia():
            t = bs.textWidget(parent=self._rootWidget,position=(15,v+10),size=(width-30,30),
                              color=(0.7,0.9,0.7,1.0),
                              maxWidth=width*0.95,
                              scale=self._R.ouyaInstructionsTextScale,
                              text=self._R.ouyaInstructionsText,
                              hAlign="center",vAlign="top")
        else:
            t = bs.textWidget(parent=self._rootWidget,position=(15,v+10),size=(width-30,30),
                              color=(0.7,0.9,0.7,1.0),
                              maxWidth=width*0.95,
                              scale=self._R.macInstructionsTextScale,
                              text=self._R.macInstructionsText,
                              hAlign="center",vAlign="top")
            v -= 90
            bWidth = 300

            b = bs.buttonWidget(parent=self._rootWidget,position=((width-bWidth)*0.5,v-10),size=(bWidth,50),
                                label=self._R.getDriverText,
                                call=bs.WeakCall(bs.openURL,'http://tattiebogle.net/index.php/ProjectRoot/Xbox360Controller/OsxDriver'))
            bs.containerWidget(edit=self._rootWidget,startButton=b)

            v -= 60
            t = bs.textWidget(parent=self._rootWidget,position=(15,v),size=(width-30,30),
                              color=(0.7,0.9,0.7,1.0),
                              maxWidth=width*0.95,
                              scale=self._R.macInstructions2TextScale,
                              text=self._R.macInstructions2Text,
                              hAlign="center",vAlign="top")

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()


class ConnectMobileDevicesWindow(Window):

    def __init__(self):

        self._R = bs.getResource('connectMobileDevicesWindow')
        width = 700
        height = 390
        spacing = 40
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight',scale=1.6 if gSmallUI else 1.3 if gMedUI else 1.0)
        b = bs.buttonWidget(parent=self._rootWidget,position=(40,height-65),size=(120,60),scale=0.8,
                            label=bs.getResource('backText'),buttonType='back',
                            call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        t = bs.textWidget(parent=self._rootWidget,position=(20,height-56),size=(width,25),text=self._R.titleText,
                          maxWidth=450,
                          color=gTitleColor,scale=0.8,hAlign="center",vAlign="top")
        v = height - 70
        v -= spacing*1.2
        t = bs.textWidget(parent=self._rootWidget,position=(15,v),size=(width-30,30),
                          maxWidth=width*0.95,
                          color=(0.7,0.9,0.7,1.0),
                          scale=self._R.explanationScale,
                          text=self._R.explanationText,
                          hAlign="center",vAlign="top")
        v -= 100
        bWidth = 200

        # hmm the itms:// version doesnt bounce through safari but is kinda apple-specific-ish
        sep = 230

        if bs.isRunningOnFantasia():
            t = bs.textWidget(parent=self._rootWidget,position=(width*0.5,v+10),size=(0,0),
                              color=(0.7,0.9,0.7,1.0), scale=0.8,
                              text=self._R.getItForText,
                              maxWidth=width*0.95,hAlign="center",vAlign="center")
            v -= 40
        else:
            t = bs.textWidget(parent=self._rootWidget,position=(130,v),size=(0,30),
                              color=(0.7,0.9,0.7,1.0), scale=0.8,
                              text=self._R.forIOSText, hAlign="center",vAlign="top")
            t = bs.textWidget(parent=self._rootWidget,position=(470,v),size=(0,30),
                              color=(0.7,0.9,0.7,1.0), scale=0.8,
                              text=self._R.forAndroidText, hAlign="center",vAlign="top")
            v -= 40

            b = bs.buttonWidget(parent=self._rootWidget,position=(width*0.5-sep-bWidth*0.5+20,v-10),size=(bWidth*0.94,50),
                                label=self._R.appStoreText, call=bs.WeakCall(bs.openURL,'http://appstore.com/bombsquadremote'))
            b = bs.buttonWidget(parent=self._rootWidget,position=(width*0.5-bWidth*0.5+27,v-10),size=(bWidth*0.89,50),
                                label=self._R.googlePlayText, call=bs.WeakCall(bs.openURL,'http://play.google.com/store/apps/details?id=net.froemling.bsremote'))
            b = bs.buttonWidget(parent=self._rootWidget,position=(width*0.5+sep-bWidth*0.5-12,v-10),size=(bWidth*1.0,50),
                                label=self._R.amazonText, call=bs.WeakCall(bs.openURL,'http://www.amazon.com/gp/mas/dl/android?p=net.froemling.bsremote'))
            v -= 32

        t = bs.textWidget(parent=self._rootWidget,position=(width*0.5,v),size=(0,0),
                          color=(0.7,0.9,0.7,0.8),
                          scale=self._R.bestResultsScale,
                          text=self._R.bestResultsText,
                          maxWidth=width*0.95,
                          hAlign="center",vAlign="top")

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()


class ConfigGamePadWindow(Window):

    def __init__(self):
        width = 480
        height = 170
        spacing = 40
        R = bs.getResource('configGamepadSelectWindow')
        self._rootWidget = bs.containerWidget(scale=2.0 if gSmallUI else 1.5 if gMedUI else 1.0,
                                              size=(width,height),transition='inRight')
        b = bs.buttonWidget(parent=self._rootWidget,position=(25,height-60),size=(120,60),
                            label=bs.getResource('backText'),buttonType='back',scale=0.8,
                            call=self._back)
        # lets not have anything selected by default; its misleading looking for the controller getting configured..
        bs.containerWidget(edit=self._rootWidget,cancelButton=b,selectedChild=0)
        t = bs.textWidget(parent=self._rootWidget,position=(20,height-50),size=(width,25),text=R.titleText,
                          maxWidth=260,color=gTitleColor,hAlign="center",vAlign="center")
        v = height - 60
        v -= spacing
        t = bs.textWidget(parent=self._rootWidget,position=(15,v),size=(width-30,30),
                          scale=0.8,
                          text=R.pressAnyButtonText,
                          maxWidth=width*0.95,color=gInfoTextColor,hAlign="center",vAlign="top")
        v -= spacing * 1.24
        ua = bs.getUserAgentString()
        if 'android' in ua and not bs.isOuyaVersion() and not bs.isRunningOnFantasia():
            t = bs.textWidget(parent=self._rootWidget,position=(15,v),size=(width-30,30),
                              scale=0.46,
                              text=R.androidNoteText,
                              maxWidth=width*0.95,
                              color=(0.7,0.9,0.7,0.5),
                              hAlign="center",vAlign="top")

        bs.captureGamePadInput(gamePadConfigureCallback)

    def _back(self):
        bs.releaseGamePadInput()
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()


class ConfigTouchscreenWindow(Window):

    def __del__(self):
        import _bs
        # note - this happens in 'back' too just to get it going earlier..
        # (but we do it here too in case somehow the window is closed by another means)
        _bs._setTouchscreenEditing(False)

    def __init__(self):

        import _bs
        self._width = 550
        self._height = 320
        self._spacing = 40
        self._R = bs.getResource('configTouchscreenWindow')

        _bs._setTouchscreenEditing(True)

        self._rootWidget = bs.containerWidget(size=(self._width,self._height),transition='inRight',scale=1.8 if gSmallUI else 1.5 if gMedUI else 1.0)
        self._buildGui()

    def _buildGui(self):

        # clear anything already there
        children = bs.getWidgetChildren(self._rootWidget)
        for c in children: bs.deleteWidget(c)

        b = bs.buttonWidget(parent=self._rootWidget,position=(55,self._height-60),size=(120,60),
                            label=bs.getResource('backText'),buttonType='back',scale=0.8,
                            call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(25,self._height-50),size=(self._width,25),text=self._R.titleText,
                          color=gTitleColor,maxWidth=280,
                          hAlign="center",vAlign="center")

        h = 50
        v = self._height - 140

        cl = (0.8,0.8,0.8,1.0)
        cl2 = (0.8,0.8,0.8)

        bs.textWidget(parent=self._rootWidget,position=(-10,v+43),size=(self._width,25),
                      text=self._R.swipeInfoText,
                      color=(0,0.9,0.1,0.7),maxWidth=self._width*0.9,
                      scale=0.55,hAlign="center",vAlign="center")

        try: curVal = bs.config['Touch Movement Control Type']
        except Exception: curVal = 'swipe'
        bs.textWidget(parent=self._rootWidget,position=(h,v-2),size=(0,30),
                      text=self._R.movementText,maxWidth=190,color=cl,vAlign='center')
        c1 = bs.checkBoxWidget(parent=self._rootWidget,position=(h+200,v),size=(170,30),
                               text=self._R.joystickText,maxWidth=100,textColor=cl2,scale=0.9)
        c2 = bs.checkBoxWidget(parent=self._rootWidget,position=(h+337,v),size=(170,30),
                               text=self._R.swipeText,maxWidth=100,textColor=cl2,value=0,scale=0.9)
        _makeRadioGroup((c1,c2),('joystick','swipe'),curVal,self._movementChanged)

        v -= 50

        try: curVal = bs.config['Touch Action Control Type']
        except Exception: curVal = 'buttons'
        bs.textWidget(parent=self._rootWidget,position=(h,v-2),size=(0,30),
                      text=self._R.actionsText,maxWidth=190,color=cl,vAlign='center')
        c1 = bs.checkBoxWidget(parent=self._rootWidget,position=(h+200,v),size=(170,30),
                               text=self._R.buttonsText,maxWidth=100,textColor=cl2,scale=0.9)
        c2 = bs.checkBoxWidget(parent=self._rootWidget,position=(h+337,v),size=(170,30),
                               text=self._R.swipeText,maxWidth=100,textColor=cl2,scale=0.9)
        _makeRadioGroup((c1,c2),('buttons','swipe'),curVal,self._actionsChanged)


        v -= 50
        configTextBox(parent=self._rootWidget,position=(h,v),xOffset=65,name="Touch Controls Scale",displayName=self._R.touchControlsScaleText,
                      type="float",changeSound=False,minVal=0.1,maxVal=3.0,increment=0.1)

        v -= 45

        b = bs.buttonWidget(parent=self._rootWidget,position=(self._width*0.5-56,v),size=(150,50),
                            label=self._R.resetText,scale=0.65,call=self._reset)

        t = bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,v-6),size=(0,0),hAlign='center',
                          text=self._R.dragControlsText,maxWidth=self._width*0.8,
                          scale=0.55,color=(1,1,1,0.4))
        

    def _actionsChanged(self,v):
        bs.config['Touch Action Control Type'] = v
        bs.writeConfig()
        bs.applySettings()

    def _movementChanged(self,v):
        bs.config['Touch Movement Control Type'] = v
        bs.writeConfig()
        bs.applySettings()

    def _reset(self):
        prefs = ['Touch Movement Control Type',
                 'Touch Action Control Type',
                 'Touch Controls Scale',
                 'Touch DPad X',
                 'Touch DPad Y',
                 'Touch Buttons X',
                 'Touch Buttons Y']
        for p in prefs:
            if p in bs.config:
                del(bs.config[p])

        bs.writeConfig()
        bs.applySettings()

        bs.realTimer(0,self._buildGui)


    def _back(self):
        import _bs
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()
        _bs._setTouchscreenEditing(False)


def gamePadConfigureCallback(event):

    # ignore all but button-presses
    if event['type'] not in ['BUTTONDOWN','HATMOTION']: return

    bs.releaseGamePadInput()

    try: bs.containerWidget(edit=uiGlobals['mainMenuWindow'],transition='outLeft')
    except Exception,e:
        try:
            import traceback
            print 'Exception transitioning out mainMenuWindow on gamePadConfigureCallback:',e,'mainMenuWindow is',uiGlobals['mainMenuWindow']
            print 'STACK:'
            traceback.print_stack()
        except Exception,e: print 'EXC-OMG-WTF',e
    bs.playSound(bs.getSound('activateBeep'))
    bs.playSound(bs.getSound('swish'))
    if bs.getInputDeviceAllowsConfiguring(event['inputID']):
        uiGlobals['mainMenuWindow'] = GamePadConfigWindow(event["inputID"]).getRootWidget()
    else:
        width = 700
        height = 200
        buttonWidth = 100
        uiGlobals['mainMenuWindow'] = d = bs.containerWidget(scale=1.7 if gSmallUI else 1.4 if gMedUI else 1.0,
                                                             size=(width,height),transition='inRight')
        deviceName = bs.getInputDeviceName(event['inputID'])
        if deviceName == 'iDevice':
            msg = bs.getResource('bsRemoteConfigureInAppText')
        else:
            msg = bs.getResource('cantConfigureDeviceText').replace('${DEVICE}',deviceName)
        t = bs.textWidget(parent=d,position=(0,height-80),size=(width,25),
                          text=msg,
                          scale=0.8,
                          hAlign="center",vAlign="top")
        def _ok():
            bs.containerWidget(edit=d,transition='outRight')
            uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()
        b = bs.buttonWidget(parent=d,position=((width-buttonWidth)/2,20),size=(buttonWidth,60),label=bs.getResource('okText'),call=_ok)


class AwaitGamePadInputWindow(Window):

    def __init__(self,gamePad,button,callback,message=None,message2=None):

        if message is None:
            print 'AwaitGamePadInputWindow message is None!'
            message = 'Press any button...' # shouldnt get here
            
        self._callback = callback
        self._input = gamePad
        self._captureButton = button

        width = 400
        height = 150
        self._rootWidget = bs.containerWidget(scale=2.0 if gSmallUI else 1.9 if gMedUI else 1.0,
                                              size=(width,height),transition='inScale')
        t = bs.textWidget(parent=self._rootWidget,position=(0,(height-60) if message2 is None else (height-50)),
                          size=(width,25),text=message,maxWidth=width*0.9,
                          hAlign="center",vAlign="center")
        if message2 is not None:
            t = bs.textWidget(parent=self._rootWidget,position=(width*0.5,height-60),size=(0,0),text=message2,maxWidth=width*0.9,
                              scale=0.47,color=(0.7,1.0,0.7,0.6),hAlign="center",vAlign="center")
        self._counter = 5
        self._countDownText = bs.textWidget(parent=self._rootWidget,hAlign='center',position=(0,height-110),size=(width,25),color=(1,1,1,0.3),text=str(self._counter));
        self._decrementTimer = bs.Timer(1000,bs.Call(self._decrement),repeat=True,timeType='real')
        bs.captureGamePadInput(bs.WeakCall(self._eventCallback))

    def __del__(self):
        pass

    def die(self):
        self._decrementTimer = None # this strong-refs us; killing it allow us to die now
        bs.releaseGamePadInput()
        bs.containerWidget(edit=self._rootWidget,transition='outScale')
        
    def _eventCallback(self,event):
        match = (event["inputID"] == self._input)

        # update - we now allow *any* input device of this type
        try:
            if bs.getInputDeviceName(event['inputID']) == bs.getInputDeviceName(self._input): match = True
        except Exception,e:
            print 'AwaitGamePadInputWindow: error comparing input devices:',e

        if match: self._callback(self._captureButton,event,self)

    def _decrement(self):
        self._counter -= 1
        if (self._counter >= 1):
            bs.textWidget(edit=self._countDownText,text=str(self._counter))
        else:
            bs.playSound(bs.getSound('error'))
            self.die()

        
class AwaitKeyboardInputWindow(Window):

    def __init__(self,button,ui,settings):

        self._captureButton = button
        self._captureKeyUI=ui
        self._settings = settings

        width = 400
        height = 150
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight')
        t = bs.textWidget(parent=self._rootWidget,position=(0,height-60),size=(width,25),text=bs.getResource('pressAnyKeyText'),
                          hAlign="center",vAlign="top")

        self._counter = 5
        self._countDownText = bs.textWidget(parent=self._rootWidget,hAlign='center',position=(0,height-110),size=(width,25),color=(1,1,1,0.3),text=str(self._counter));
        self._decrementTimer = bs.Timer(1000,bs.Call(self._decrement),repeat=True,timeType='real')
        bs.captureKeyboardInput(bs.WeakCall(self._buttonCallback))

    def __del__(self):
        bs.releaseKeyboardInput()

    def _die(self):
        self._decrementTimer = None # this strong-refs us; killing it allow us to die now
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        
    def _buttonCallback(self,event):
        self._settings[self._captureButton] = event["keysym"]
        t = bs.getKeyName(None,event["keysym"])
        bs.textWidget(edit=self._captureKeyUI,text=t)
        bs.playSound(bs.getSound('gunCocking'))
        self._die()

    def _decrement(self):
        self._counter -= 1
        if (self._counter >= 1):
            bs.textWidget(edit=self._countDownText,text=str(self._counter))
        else:
            self._die()


# given a gamepad name, id, and button, returns the gamePad's currently configured button id
def getControllerValue(c,button,exceptionOnUnknown=False):

    name = bs.getInputDeviceName(c)
    uniqueID = bs.getInputDeviceUniqueIdentifier(c)
    ua = bs.getUserAgentString()

    if (bs.config.has_key("Controllers")):
        jsconfig = bs.config["Controllers"]
        if (jsconfig.has_key(name)):
            src = None
            if (jsconfig[name].has_key(uniqueID)):
                src = jsconfig[name][uniqueID]
            elif (jsconfig[name].has_key("default")):
                src = jsconfig[name]["default"]
            if (src != None):
                if (src.has_key(button)):
                    return src[button]
                else: return -1

    # print 'TEMP RETURNING NO DEFAULTS'
    # if exceptionOnUnknown: raise Exception()
    # else: return -1
        
    # lets try to set some defaults for well-known joystick types
    # to anyone reading this, if your gamepad isn't covered by this list,
    # send me what your default buttons should be.
    # standard arrangement is jump=bottom, punch=left, bomb=right, throw=top
    if 'OUYA' in ua:
        if name == 'Generic X-Box pad':
            try: return {
                'analogStickDeadZone':1.2,'buttonBomb':98,'buttonIgnored':111,'buttonJump':97,'buttonPickUp':101,
                'buttonPunch':100,'buttonStart':109,'triggerRun1':12,'triggerRun2':15}[button]
            except Exception: pass
        # elif name == 'Logitech Dual Action':
        #     try: return {'buttonBomb':98,'buttonJump':97,'buttonPickUp':101,'buttonPunch':100,'buttonStart':109}[button]
        #     except Exception: pass
        elif name == 'Microsoft X-Box 360 pad':
            try: return {'analogStickDeadZone':1.2,'buttonBomb':98,'buttonIgnored':83,'buttonJump':97,'buttonPickUp':101,
                         'buttonPunch':100,'buttonStart':109,'triggerRun1':18,'triggerRun2':19}[button]
            except Exception: pass
        elif name == 'OUYA Game Controller':
            try: return {'analogStickDeadZone':1.2,'triggerRun2':19,'triggerRun1':18,'startButtonActivatesDefaultWidget':False, 'autoRecalibrateAnalogStick': True,'buttonBomb':98,'buttonDown':21,'buttonJump':97,'buttonLeft':22,'buttonPickUp':101,'buttonPunch':100,
                         'buttonRight':23,'buttonStart':83,'buttonUp':20}[button]
            except Exception: pass
        elif name == 'PLAYSTATION(R)3 Controller':
            try: return {'buttonBomb':98,'buttonDown':21,'buttonIgnored':83,'buttonJump':97,'buttonLeft':22,
                         'buttonPickUp':101,'buttonPunch':100,'buttonRight':23,'buttonStart':109,'buttonUp':20}[button]
            except Exception: pass
        elif name == 'Sony PLAYSTATION(R)3 Controller':
            try: return {'buttonPickUp': 101, 'buttonBomb': 98, 'buttonJump': 97, 'buttonUp': 20, 'buttonLeft': 22,
                        'buttonRight': 23, 'buttonStart': 109, 'buttonPunch': 100, 'buttonDown': 21, 'buttonIgnored': 83}[button]
            except Exception: pass
        elif name == 'Xbox 360 Wireless Receiver':
            try: return {'analogStickDeadZone':1.2,'buttonBomb':98,'buttonDown':21,'buttonIgnored':83,
                         'buttonJump':97,'buttonLeft':22,'buttonPickUp':101,'buttonPunch':100,
                         'buttonRight':23,'buttonStart':109,'buttonUp':20,'triggerRun1':18,'triggerRun2':19}[button]
            except Exception: pass

    elif 'Mac' in ua:
        if name == 'PLAYSTATION(R)3 Controller': # ps3 gamepad
            try: return  {'buttonLeft':8,'buttonUp':5,'buttonRight':6,'buttonDown':7,'buttonJump':15,'buttonPunch':16,'buttonBomb':14,'buttonPickUp':13,'buttonStart':4,'buttonIgnored':17}[button]
            except Exception: pass
        if name == 'Wireless 360 Controller' or name == 'Controller': # xbox360 gamepads
            try: return  {'analogStickDeadZone':1.2,'buttonBomb':13,'buttonDown':2,'buttonJump':12,'buttonLeft':3,
                          'buttonPickUp':15,'buttonPunch':14,'buttonRight':4,'buttonStart':5,
                          'buttonUp':1,'triggerRun1':5,'triggerRun2':6,'buttonIgnored':11}[button]
            except Exception: pass
        if name in ['Logitech Dual Action','Logitech Cordless RumblePad 2']:
            try: return  {'buttonJump':2,'buttonPunch':1,'buttonBomb':3,'buttonPickUp':4,'buttonStart':10}[button]
            except Exception: pass
        if name == 'GamePad Pro USB ' : # old gravis gamepad
            try: return  {'buttonJump':2,'buttonPunch':1,'buttonBomb':3,'buttonPickUp':4,'buttonStart':10}[button]
            except Exception: pass
        if name == 'Microsoft SideWinder Plug & Play Game Pad':
            try: return  {'buttonJump':1,'buttonPunch':3,'buttonBomb':2,'buttonPickUp':4,'buttonStart':6}[button]
            except Exception: pass
        if name == 'Saitek P2500 Rumble Force Pad': # Saitek P2500 Rumble Force Pad.. (hopefully works for others too?..)
            try: return  {'buttonJump':3,'buttonPunch':1,'buttonBomb':4,'buttonPickUp':2,'buttonStart':11}[button]
            except Exception: pass
        if name == 'Twin USB Joystick': # some crazy 'Senze' dual gamepad (the second half is handled under the hood)
            try: return  {'analogStickLR':3,'analogStickLR_B':7,'analogStickUD':4,'analogStickUD_B':8,'buttonBomb':2,'buttonBomb_B':14,
                          'buttonJump':3,'buttonJump_B':15,'buttonPickUp':1,'buttonPickUp_B':13,'buttonPunch':4,'buttonPunch_B':16,'buttonRun1':7,
                          'buttonRun1_B':19,'buttonRun2':8,'buttonRun2_B':20,'buttonStart':10,'buttonStart_B':22,'enableSecondary':1,'unassignedButtonsRun':False}[button]
            except Exception: pass
        if name == 'USB Gamepad ': # some weird 'JITE' gamepad
            try: return  {'analogStickLR':4,'analogStickUD':5,'buttonJump':3,'buttonPunch':4,'buttonBomb':2,'buttonPickUp':1,'buttonStart':10}[button]
            except Exception: pass

    # default keyboard vals across platforms..
    if name == 'Keyboard' and uniqueID == '#2':
        try: return  {'buttonJump':258,'buttonPunch':257,'buttonBomb':262,'buttonPickUp':261,'buttonUp':273,'buttonDown':274,'buttonLeft':276,'buttonRight':275,'buttonStart':263}[button]
        except Exception: pass
    if name == 'Keyboard' and uniqueID == '#1':
        try: return {'buttonJump':107,'buttonPunch':106,'buttonBomb':111,'buttonPickUp':105,'buttonUp':119,'buttonDown':115,'buttonLeft':97,'buttonRight':100}[button]
        except Exception: pass


    if bs.isRunningOnFantasia():
        if name == 'Thunder':
            try: return {'triggerRun2': 23, 'triggerRun1': 24, 'buttonPickUp': 101, 'buttonBomb': 98, 'buttonJump': 97,
                         'buttonUp': 20, 'buttonLeft': 22, 'buttonRight': 23, 'buttonStart': 109, 'buttonPunch': 100, 'buttonDown': 21}[button]
            except Exception: pass

    # ok, this gamepad's not in our specific preset list; lets error if requested to
    # otherwise lets try some good guesses based on known types...
    if exceptionOnUnknown: raise Exception("Unknown joystick type")

    # leaving these in here for now but not gonna add any more now that we have
    # fancy-pants config sharing across the internet...
    if 'Mac' in ua:
        if 'PLAYSTATION' in name: # ps3 gamepad?..
            try: return  {'buttonLeft':8,'buttonUp':5,'buttonRight':6,'buttonDown':7,'buttonJump':15,'buttonPunch':16,'buttonBomb':14,'buttonPickUp':13,'buttonStart':4}[button]
            except Exception: pass
        if 'Logitech' in name: # Dual Action Config - hopefully applies to more...
            try: return  {'buttonJump':2,'buttonPunch':1,'buttonBomb':3,'buttonPickUp':4,'buttonStart':10}[button]
            except Exception: pass
        if 'Saitek' in name: # Saitek P2500 Rumble Force Pad.. (hopefully works for others too?..)
            try: return  {'buttonJump':3,'buttonPunch':1,'buttonBomb':4,'buttonPickUp':2,'buttonStart':11}[button]
            except Exception: pass
        if 'GamePad' in name: # gravis stuff?...
            try: return  {'buttonJump':2,'buttonPunch':1,'buttonBomb':3,'buttonPickUp':4,'buttonStart':10}[button]
            except Exception: pass

    # reasonable defaults..
    if 'android' in ua:
        try: return {'triggerRun2': 23, 'triggerRun1': 24, 'buttonPickUp': 101, 'buttonBomb': 98, 'buttonJump': 97,
                     'buttonUp': 20, 'buttonLeft': 22, 'buttonRight': 23, 'buttonStart': 109, 'buttonPunch': 100, 'buttonDown': 21}[button]
        except Exception: pass
    else:
        try: return {'buttonJump':1,'buttonPunch':2,'buttonBomb':3,'buttonPickUp':4,'buttonStart':5}[button]
        except Exception: pass

    # epic fail.
    return -1



class GamePadConfigWindow(Window):

    def __init__(self,gamePad,isMainMenu=True,transition='inRight',transitionOut='outRight',settings=None):

        self._input = gamePad

        # if this fails, our input device went away or something.. just return an empty zombie then
        try: self._name = bs.getInputDeviceName(self._input)
        except Exception: return

        self._R = bs.getResource('configGamepadWindow')

        self._settings = settings
        self._transitionOut = transitionOut

        # we're a secondary gamepad if supplied with settings
        self._isSecondary = (settings is not None)
        self._ext = '_B' if self._isSecondary else ''
        self._isMainMenu = isMainMenu
        self._displayName = self._name
        buttonWidth = 240
        self._width = 700 if self._isSecondary else 730
        self._height = 440 if self._isSecondary else 450
        self._spacing = 40
        self._rootWidget = bs.containerWidget(size=(self._width,self._height),
                                              scale=1.55 if gSmallUI else 1.35 if gMedUI else 1.0,
                                              singleDepth=True,
                                              transition=transition)
        # dont ask to config joysticks while we're in here..
        global gCanAskToConfigGamePads
        gCanAskToConfigGamePads = False

        self._rebuildUI()


    def _rebuildUI(self):

        # clear existing UI
        for w in bs.getWidgetChildren(self._rootWidget): bs.deleteWidget(w)

        self._textWidgets = {}

        # if we were supplied with settings, we're a secondary joystick and just operate on that.
        # in the other (normal) case we make our own..
        if not self._isSecondary:

            # fill our temp config with present values (for our primary and secondary controls)
            self._settings = {}
            for s in ['buttonJump','buttonJump_B',
                      'buttonPunch','buttonPunch_B',
                      'buttonBomb','buttonBomb_B',
                      'buttonPickUp','buttonPickUp_B',
                      'buttonStart','buttonStart_B',
                      'buttonUp','buttonUp_B',
                      'buttonDown','buttonDown_B',
                      'buttonLeft','buttonLeft_B',
                      'buttonRight','buttonRight_B',
                      'buttonRun1','buttonRun1_B',
                      'buttonRun2','buttonRun2_B',
                      'triggerRun1','triggerRun1_B',
                      'triggerRun2','triggerRun2_B',
                      'buttonIgnored','buttonIgnored_B',
                      'analogStickDeadZone','analogStickDeadZone_B',
                      'dpad','dpad_B',
                      'unassignedButtonsRun','unassignedButtonsRun_B',
                      'startButtonActivatesDefaultWidget','startButtonActivatesDefaultWidget_B',
                      'autoRecalibrateAnalogStick','autoRecalibrateAnalogStick_B',
                      'analogStickLR','analogStickLR_B',
                      'analogStickUD','analogStickUD_B',
                      'enableSecondary',]:
                val = getControllerValue(self._input,s)
                if val != -1: self._settings[s] = val

        if not self._isSecondary:
            cancelButton = b = bs.buttonWidget(parent=self._rootWidget,position=(51,self._height-65),size=(160,60),label=bs.getResource('cancelText'),scale=0.9,call=self._cancel);
            bs.containerWidget(edit=self._rootWidget,cancelButton=cancelButton)
        saveButton = b = bs.buttonWidget(parent=self._rootWidget,position=(self._width-(165 if self._isSecondary else 195),self._height-65),
                                         size=((160 if self._isSecondary else 180),60),
                                         label=bs.getResource('doneText') if self._isSecondary else bs.getResource('makeItSoText'),
                                         scale=0.9,
                                         call=self._save)
        bs.containerWidget(edit=self._rootWidget,startButton=saveButton)

        if not self._isSecondary:

            v = self._height - 59
            t = bs.textWidget(parent=self._rootWidget,position=(0,v+5),size=(self._width,25),
                              text=self._R.titleText,
                              color=gTitleColor,maxWidth=310,
                              hAlign="center",vAlign="center")
            v -= 48

            t = bs.textWidget(parent=self._rootWidget,position=(0,v+3),size=(self._width,25),text=self._name,
                              color=gInfoTextColor,maxWidth=self._width*0.9,
                              hAlign="center",vAlign="center")

            v -= self._spacing * 1

            bs.textWidget(parent=self._rootWidget,position=(50,v+10),size=(self._width-100,30),
                          text=self._R.appliesToAllText,maxWidth=330,
                          scale=0.65, color=(0.5,0.6,0.5,1.0),hAlign='center',vAlign='center')
            v -= 70
        else:
            v = self._height - 59
            t = bs.textWidget(parent=self._rootWidget,position=(0,v+5),size=(self._width,25),text=self._R.secondaryText,
                              color=gTitleColor,maxWidth=300,hAlign="center",vAlign="center")
            v -= self._spacing * 1

            bs.textWidget(parent=self._rootWidget,position=(50,v+10),size=(self._width-100,30),
                          text=self._R.secondHalfText,
                          maxWidth=300,scale=0.65, color=(0.6,0.8,0.6,1.0),hAlign='center')
            self._enableCheckBox = bs.checkBoxWidget(parent=self._rootWidget,position=(self._width*0.5-80,v-73),
                                                     value=self._getEnableSecondaryValue(),
                                                     valueCall=self._enableCheckBoxChanged,
                                                     size=(200,30),text=self._R.secondaryEnableText,scale=1.2)
            v = self._height - 220

        hOffs = 160
        dist = 70

        dColor = (0.4,0.4,0.8)

        sx = 1.2
        sy = 0.85

        dpm = self._R.pressAnyButtonOrDpadText
        dpm2 = self._R.ifNothingHappensTryAnalogText
        self._captureButton(pos=(hOffs,v+sy*dist),color=dColor,button='buttonUp'+self._ext,texture=bs.getTexture('upButton'),scale=1.0,message=dpm,message2=dpm2)
        self._captureButton(pos=(hOffs-sx*dist,v),color=dColor,button='buttonLeft'+self._ext,texture=bs.getTexture('leftButton'),scale=1.0,message=dpm,message2=dpm2)
        self._captureButton(pos=(hOffs+sx*dist,v),color=dColor,button='buttonRight'+self._ext,texture=bs.getTexture('rightButton'),scale=1.0,message=dpm,message2=dpm2)
        self._captureButton(pos=(hOffs,v-sy*dist),color=dColor,button='buttonDown'+self._ext,texture=bs.getTexture('downButton'),scale=1.0,message=dpm,message2=dpm2)

        dpm3 = self._R.ifNothingHappensTryDpadText
        self._captureButton(pos=(hOffs+130,v-125),color=(0.4,0.4,0.6),button='analogStickLR'+self._ext,
                            texture=bs.getTexture('analogStick'),scale=1.2,message=self._R.pressLeftRightText,message2=dpm3)

        self._captureButton(pos=(self._width*0.5,v),color=(0.4,0.4,0.6),button='buttonStart'+self._ext,texture=bs.getTexture('startButton'),scale=0.7)

        hOffs = self._width-160

        self._captureButton(pos=(hOffs,v+sy*dist),color=(0.6,0.4,0.8),button='buttonPickUp'+self._ext,texture=bs.getTexture('buttonPickUp'),scale=1.0)
        self._captureButton(pos=(hOffs-sx*dist,v),color=(0.7,0.5,0.1),button='buttonPunch'+self._ext,texture=bs.getTexture('buttonPunch'),scale=1.0)
        self._captureButton(pos=(hOffs+sx*dist,v),color=(0.5,0.2,0.1),button='buttonBomb'+self._ext,texture=bs.getTexture('buttonBomb'),scale=1.0)
        self._captureButton(pos=(hOffs,v-sy*dist),color=(0.2,0.5,0.2),button='buttonJump'+self._ext,texture=bs.getTexture('buttonJump'),scale=1.0)

        bs.buttonWidget(parent=self._rootWidget,label=self._R.advancedText,textScale=0.9,color=(0.45,0.4,0.5),
                        textColor=(0.65,0.6,0.7),position=(self._width-350,30),size=(130,40),call=self._doAdvanced)

    def _doAdvanced(self):

        class _AdvancedWindow(Window):

            def __init__(self,parentWindow,transition='inRight'):
                self._parentWindow = parentWindow

                self._R = parentWindow._R
                self._width = 400
                self._height = 380 if self._parentWindow._isSecondary else 470
                self._textWidgets = {}
                self._rootWidget = bs.containerWidget(transition='inScale',size=(self._width,self._height),
                                                      scale=1.06*(1.5 if gSmallUI else 1.35 if gMedUI else 1.0),singleDepth=True)
                b = bs.buttonWidget(parent=self._rootWidget,position=(self._width-180,self._height-55),size=(146,48),textScale=0.8,
                            label=bs.getResource('doneText'),call=self._done)
                bs.containerWidget(edit=self._rootWidget,startButton=b)

                h = 30
                v = self._height - 46

                # dont allow secondary joysticks to handle unassigned buttons
                if not self._parentWindow._isSecondary:
                    v -= 40
                    c1 = bs.checkBoxWidget(parent=self._rootWidget,position=(h+25,v),size=(350,30),
                                           text=self._R.unassignedButtonsRunText,textColor=(0.6,0.6,0.6),maxWidth=330,
                                           scale=self._R.unassignedButtonsRunTextScale,valueCall=self._parentWindow._setUnassignedButtonsRunValue,
                                           value=self._parentWindow._getUnassignedButtonsRunValue())

                v -= 45
                h2 = h + 12
                self._captureButton(pos=(h2,v),name=self._R.runButton1Text,control='buttonRun1'+self._parentWindow._ext)
                v -= 32
                self._captureButton(pos=(h2,v),name=self._R.runButton2Text,control='buttonRun2'+self._parentWindow._ext)
                v -= 40
                self._captureButton(pos=(h2,v),name=self._R.runTrigger1Text,
                                    control='triggerRun1'+self._parentWindow._ext,
                                    message=self._R.pressAnyAnalogTriggerText)
                v -= 32
                self._captureButton(pos=(h2,v),name=self._R.runTrigger2Text,
                                    control='triggerRun2'+self._parentWindow._ext,
                                    message=self._R.pressAnyAnalogTriggerText)
                v -= 42
                self._captureButton(pos=(h2,v),name=self._R.ignoredButtonText,control='buttonIgnored'+self._parentWindow._ext)
                bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,v-8),size=(0,0),
                              text=self._R.ignoredButtonDescriptionText,
                              color=(0.7,1,0.7,0.6),
                              scale=0.4,maxWidth=350.0,
                              hAlign='center',vAlign='center')

                v -= 43
                c1 = bs.checkBoxWidget(parent=self._rootWidget,position=(h+16,v),size=(350,30),
                                       text=self._R.startButtonActivatesDefaultText,
                                       textColor=(0.6,0.6,0.6),maxWidth=450,
                                       scale=0.65,valueCall=self._parentWindow._setStartButtonActivatesDefaultWidgetValue,
                                       value=self._parentWindow._getStartButtonActivatesDefaultWidgetValue())

                v -= 35
                self._configValueEditor(self._R.analogStickDeadZoneText,control='analogStickDeadZone'+self._parentWindow._ext,
                                        position=(h-13,v),minVal=0,maxVal=10.0,increment=0.1,xOffset=-38)
                v -= 25

                c1 = bs.checkBoxWidget(parent=self._rootWidget,position=(h+25,v),size=(350,30),
                                       text=self._R.autoRecalibrateText,textColor=(0.7,0.7,0.7),maxWidth=470,
                                       scale=0.65,valueCall=self._parentWindow._setAutoRecalibrateAnalogStickValue,
                                       value=self._parentWindow._getAutoRecalibrateAnalogStickValue())
                v -= 0
                bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,v-8),size=(0,0),
                              text=self._R.autoRecalibrateDescriptionText,
                              color=(0.7,1,0.7,0.6),maxWidth=380,
                              scale=0.4,hAlign='center',vAlign='center')
                v -= 55

                # child joysticks cant have child joysticks.. that's just crazy talk
                if not self._parentWindow._isSecondary:
                    bs.buttonWidget(parent=self._rootWidget,label=self._R.twoInOneSetupText,textScale=0.7,
                                    position=(40,v),size=(self._width-80,30),call=self._parentWindow._showSecondaryEditor)


            def _captureButton(self,pos,name,control,message=self._R.pressAnyButtonText):
                b = bs.buttonWidget(parent=self._rootWidget,position=(pos[0],pos[1]),
                                    label=name,size=(200,45),scale=0.7)
                b2 = bs.buttonWidget(parent=self._rootWidget,position=(pos[0]+260,pos[1]),
                                     color=(0.45,0.4,0.5),textColor=(0.65,0.6,0.7),
                                     label=self._R.clearText,size=(110,50),scale=0.6,call=bs.Call(self._clearControl,control))
                # make this in a timer so that it shows up on top of all other buttons
                def foo():
                    t = bs.textWidget(parent=self._rootWidget,position=(pos[0]+199,pos[1]+15),
                                      color=(1,1,1,0.3),size=(0,0),hAlign='center',vAlign='center',scale=0.62,
                                      text=self._parentWindow._getControlValueName(control),maxWidth=100)
                    self._textWidgets[control] = t
                    bs.buttonWidget(edit=b,call=bs.WeakCall(AwaitGamePadInputWindow,self._parentWindow._input,control,self._gamePadEvent,message))
                bs.realTimer(0,foo)


            def _inc(self,control,minVal,maxVal,inc):
                try: val = self._parentWindow._settings[control]
                except Exception: val = 1.0
                val = min(maxVal,max(minVal,val+inc))
                if abs(1.0-val) < 0.001:
                    if control in self._parentWindow._settings: del(self._parentWindow._settings[control])
                else: self._parentWindow._settings[control] = round(val,1)
                bs.textWidget(edit=self._textWidgets[control],text=self._parentWindow._getControlValueName(control))

            def _configValueEditor(self,name,control,position,type='float',minVal=0,maxVal=100,increment=1.0,callback=None,changeSound=True,xOffset=0,displayName=None,textScale=1.0):

                if displayName is None: displayName = name
                t = bs.textWidget(parent=self._rootWidget,position=position,size=(100,30),text=displayName,
                                  color=(0.8,0.8,0.8,1.0),hAlign="left",vAlign="center",scale=0.65,maxWidth=180)
                if type == 'string':
                    raise Exception("fixme unimplemented");
                else:
                    self._textWidgets[control] = t = bs.textWidget(parent=self._rootWidget,position=(246+xOffset,position[1]),
                                                                   size=(60,28),editable=False,
                                                                   color=(0.3,1.0,0.3,1.0),hAlign="right",vAlign="center",
                                                                   text=self._parentWindow._getControlValueName(control),padding=2)
                    b = bs.buttonWidget(parent=self._rootWidget,position=(330+xOffset,position[1]),size=(28,28),label="- ",
                                        call=bs.Call(self._inc,control,minVal,maxVal,-increment),
                                        repeat=True,enableSound=(changeSound is True))
                    b = bs.buttonWidget(parent=self._rootWidget,position=(380+xOffset,position[1]),size=(28,28),label="+ ",
                                        call=bs.Call(self._inc,control,minVal,maxVal,increment),
                                        repeat=True,enableSound=(changeSound is True))

            def _clearControl(self,control):
                e = self._parentWindow._ext
                if control in self._parentWindow._settings: del(self._parentWindow._settings[control])
                bs.textWidget(edit=self._textWidgets[control],text=self._parentWindow._getControlValueName(control))

            def _gamePadEvent(self,control,event,dialog):
                e = self._parentWindow._ext
                if control in ['triggerRun1'+e,'triggerRun2'+e]:
                    if event['type'] == 'AXISMOTION':
                        # ignore small values or else we might get triggered by noise
                        if abs(event['value']) > 0.5:
                            self._parentWindow._settings[control] = event['axis']
                            # update the button's text widget
                            bs.textWidget(edit=self._textWidgets[control],text=self._parentWindow._getControlValueName(control))
                            bs.playSound(bs.getSound('gunCocking'))
                            dialog.die()
                else:
                    if event['type'] == 'BUTTONDOWN':
                        value = event['button']
                        self._parentWindow._settings[control] = value
                        # update the button's text widget
                        bs.textWidget(edit=self._textWidgets[control],text=self._parentWindow._getControlValueName(control))
                        bs.playSound(bs.getSound('gunCocking'))
                        dialog.die()

            def _done(self):
                bs.containerWidget(edit=self._rootWidget,transition='outScale')

        w = _AdvancedWindow(self)

    def _enableCheckBoxChanged(self,value):
        if value: self._settings['enableSecondary'] = 1
        else:
            if 'enableSecondary' in self._settings: del self._settings['enableSecondary']

    def _getUnassignedButtonsRunValue(self):
        if 'unassignedButtonsRun' in self._settings: return self._settings['unassignedButtonsRun']
        else: return True
    def _setUnassignedButtonsRunValue(self,value):
        if value:
            if 'unassignedButtonsRun' in self._settings:
                del(self._settings['unassignedButtonsRun']) # clear since this is default
        else: self._settings['unassignedButtonsRun'] = False

    def _getStartButtonActivatesDefaultWidgetValue(self):
        if 'startButtonActivatesDefaultWidget' in self._settings:
            return self._settings['startButtonActivatesDefaultWidget']
        else:
            return True

    def _setStartButtonActivatesDefaultWidgetValue(self,value):
        if value:
            if 'startButtonActivatesDefaultWidget' in self._settings:
                del(self._settings['startButtonActivatesDefaultWidget']) # clear since this is default
        else: self._settings['startButtonActivatesDefaultWidget'] = False

    def _getAutoRecalibrateAnalogStickValue(self):
        if 'autoRecalibrateAnalogStick' in self._settings: return self._settings['autoRecalibrateAnalogStick']
        else: return False

    def _setAutoRecalibrateAnalogStickValue(self,value):
        if not value:
            if 'autoRecalibrateAnalogStick' in self._settings:
                del(self._settings['autoRecalibrateAnalogStick']) # clear since this is default
        else: self._settings['autoRecalibrateAnalogStick'] = True

    def _getEnableSecondaryValue(self):
        if not self._isSecondary: raise Exception("enable value only applies to secondary editor")
        if 'enableSecondary' in self._settings: return self._settings['enableSecondary']
        else: return False

    def _showSecondaryEditor(self):
        GamePadConfigWindow(self._input,isMainMenu=False,settings=self._settings,transition='inScale',transitionOut='outScale')

    def _getControlValueName(self,control):

        if control == 'analogStickLR'+self._ext:
            # this actually shows both LR and UD
            s1 = str(self._settings['analogStickLR'+self._ext] if 'analogStickLR'+self._ext in self._settings else 5 if self._isSecondary else 1)
            s2 = str(self._settings['analogStickUD'+self._ext] if 'analogStickUD'+self._ext in self._settings else 6 if self._isSecondary else 2)
            return self._R.axisText+' '+s1+' & '+s2

        # if they're looking for triggers
        if control in ['triggerRun1'+self._ext,'triggerRun2'+self._ext]:
            return (self._R.axisText+' '+str(self._settings[control])) if control in self._settings else self._R.unsetText
            
        # dead-zone
        if control == 'analogStickDeadZone'+self._ext:
            if control in self._settings: return str(self._settings[control])
            else: return str(1.0)

        # for dpad buttons: show individual buttons if any are set..
        # otherwise show whichever dpad is set (defaulting to 1)
        dPadButtons = ['buttonLeft'+self._ext,'buttonRight'+self._ext,'buttonUp'+self._ext,'buttonDown'+self._ext]
        if control in dPadButtons:
            # if *any* dpad buttons are assigned, show only button assignments
            if any(b in self._settings for b in dPadButtons):
                if control in self._settings:
                    return bs.getKeyName(self._input,self._settings[control])
                else: return self._R.unsetText
            # no dpad buttons - show the dpad number for all 4
            else:
                return self._R.dpadText+' '+str(self._settings['dpad'+self._ext] if 'dpad'+self._ext in self._settings else 2 if self._isSecondary else 1)

        # other buttons..
        if control in self._settings:
            return bs.getKeyName(self._input,self._settings[control])
        else: return self._R.unsetText

    def _gamePadEvent(self,control,event,dialog):
        e = self._ext
        # for our dpad-buttons we're looking for either a button-press or a hat-switch press
        if control in ['buttonUp'+e,'buttonLeft'+e,'buttonDown'+e,'buttonRight'+e]:
            if event['type'] in ['BUTTONDOWN','HATMOTION']:
                # if its a button-down..
                if event['type'] == 'BUTTONDOWN':
                    value = event['button']
                    self._settings[control] = value
                # if its a dpad
                elif event['type'] == 'HATMOTION':
                    # clear out any set dir-buttons
                    for b in ['buttonUp'+e,'buttonLeft'+e,'buttonRight'+e,'buttonDown'+e]:
                        if b in self._settings: del(self._settings[b])
                    if event['hat'] == (2 if self._isSecondary else 1): # exclude value in default case
                        if 'dpad'+e in self._settings: del(self._settings['dpad'+e])
                    else: self._settings['dpad'+e] = event['hat']
                # update the 4 dpad button txt widgets
                bs.textWidget(edit=self._textWidgets['buttonUp'+e],text=self._getControlValueName('buttonUp'+e))
                bs.textWidget(edit=self._textWidgets['buttonLeft'+e],text=self._getControlValueName('buttonLeft'+e))
                bs.textWidget(edit=self._textWidgets['buttonRight'+e],text=self._getControlValueName('buttonRight'+e))
                bs.textWidget(edit=self._textWidgets['buttonDown'+e],text=self._getControlValueName('buttonDown'+e))
                bs.playSound(bs.getSound('gunCocking'))
                dialog.die()

        elif control == 'analogStickLR'+e:
            if event['type'] == 'AXISMOTION':
                # ignore small values or else we might get triggered by noise
                if abs(event['value']) > 0.5:
                    axis = event['axis']
                    if axis == (5 if self._isSecondary else 1): # exclude value in default case
                        if 'analogStickLR'+e in self._settings: del(self._settings['analogStickLR'+e])
                    else: self._settings['analogStickLR'+e] = axis
                    bs.textWidget(edit=self._textWidgets['analogStickLR'+e],text=self._getControlValueName('analogStickLR'+e))
                    bs.playSound(bs.getSound('gunCocking'))
                    dialog.die()
                    # now launch the up/down waiter
                    AwaitGamePadInputWindow(self._input,'analogStickUD'+e,self._gamePadEvent,self._R.pressUpDownText)

        elif control == 'analogStickUD'+e:
            if event['type'] == 'AXISMOTION':
                # ignore small values or else we might get triggered by noise
                if abs(event['value']) > 0.5:
                    axis = event['axis']
                    # ignore our LR axis
                    if 'analogStickLR'+e in self._settings: lrAxis = self._settings['analogStickLR'+e]
                    else: lrAxis = (5 if self._isSecondary else 1)
                    if axis != lrAxis:
                        if axis == (6 if self._isSecondary else 2): # exclude value in default case
                            if 'analogStickUD'+e in self._settings: del(self._settings['analogStickUD'+e])
                        else: self._settings['analogStickUD'+e] = axis
                        bs.textWidget(edit=self._textWidgets['analogStickLR'+e],text=self._getControlValueName('analogStickLR'+e))
                        #print 'WOULD CHANGE TXT'
                        bs.playSound(bs.getSound('gunCocking'))
                        dialog.die()
        else:
            # for other buttons we just want a button-press
            if event['type'] == 'BUTTONDOWN':
                value = event['button']
                self._settings[control] = value
                # update the button's text widget
                bs.textWidget(edit=self._textWidgets[control],text=self._getControlValueName(control))
                bs.playSound(bs.getSound('gunCocking'))
                dialog.die()

    def _captureButton(self,pos,color,texture,button,scale=1.0,message=None,message2=None):
        if message is None: message = self._R.pressAnyButtonText

        baseSize = 79
        b = bs.buttonWidget(parent=self._rootWidget,position=(pos[0]-baseSize*0.5*scale,pos[1]-baseSize*0.5*scale),
                            size=(baseSize*scale,baseSize*scale),
                            texture=texture,label='',color=color)
        # make this in a timer so that it shows up on top of all other buttons
        def foo():
            uiScale = 0.66*scale
            t = bs.textWidget(parent=self._rootWidget,position=(pos[0]+0.0*scale,pos[1]-54.0*scale),
                              color=(1,1,1,0.3),size=(0,0),hAlign='center',vAlign='center',
                              scale=uiScale, text=self._getControlValueName(button),maxWidth=80)
            self._textWidgets[button] = t
            bs.buttonWidget(edit=b,call=bs.WeakCall(AwaitGamePadInputWindow,self._input,button,self._gamePadEvent,message,message2))

        bs.realTimer(0,foo)

    def _cancel(self):
        bs.containerWidget(edit=self._rootWidget,transition=self._transitionOut)
        if self._isMainMenu:
            uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()
        global gCanAskToConfigGamePads
        gCanAskToConfigGamePads = True

    def _save(self):

        bs.containerWidget(edit=self._rootWidget,transition=self._transitionOut)

        # if we're a secondary editor we just go away (we were editing our parent's settings dict)
        if self._isSecondary: return

        dst = getConfigLocationForInputDevice(self._input,default=True)
        dst = dst[0][dst[1]]
        dst.clear()

        # store any values that aren't -1
        for key,val in self._settings.items():
            if val != -1: dst[key] = val

        # if we're allowed to phone home, send this config so we can generate more defaults in the future
        if bs._shouldSubmitDebugInfo():
            bsUtils.serverPut('controllerConfig',{'ua':bs.getUserAgentString(),'name':self._name,'inputMapHash':getInputMapHash(self._input),'config':dst,'v':2})

        bs.applySettings()
        bs.writeConfig()

        bs.playSound(bs.getSound('gunCocking'))
        if self._isMainMenu:
            uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()
        global gCanAskToConfigGamePads
        gCanAskToConfigGamePads = True


class ConfigKeyboardWindow(Window):


    def __init__(self,c,transition='inRight'):

        self._R = bs.getResource('configKeyboardWindow')
        self._input = c
        self._name = bs.getInputDeviceName(self._input)
        self._uniqueID = bs.getInputDeviceUniqueIdentifier(self._input)
        self._displayName = self._name
        if self._uniqueID != '#1': self._displayName += ' '+self._uniqueID.replace('#','P')
        buttonWidth = 240
        self._width = 700
        if self._uniqueID != "#1": self._height = 450
        else: self._height = 345
        self._spacing = 40
        self._rootWidget = bs.containerWidget(size=(self._width,self._height),transition=transition)

        # dont ask to config joysticks while we're in here..
        global gCanAskToConfigGamePads
        gCanAskToConfigGamePads = False

        self._rebuildUI()


    def _rebuildUI(self):

        for w in bs.getWidgetChildren(self._rootWidget):
            bs.deleteWidget(w)

        # fill our temp config with present values
        self._settings = {}
        for button in ['buttonJump','buttonPunch','buttonBomb','buttonPickUp',
                       'buttonStart','buttonUp','buttonDown','buttonLeft','buttonRight','buttonStart']:
            self._settings[button] = getControllerValue(self._input,button)

        cancelButton = b = bs.buttonWidget(parent=self._rootWidget,position=(38,self._height-65),size=(170,60),label=bs.getResource('cancelText'),scale=0.9,call=self._cancel);
        saveButton = b = bs.buttonWidget(parent=self._rootWidget,position=(self._width-210,self._height-65),size=(180,60),label=bs.getResource('makeItSoText'),scale=0.9,textScale=0.9,call=self._save)
        bs.containerWidget(edit=self._rootWidget,cancelButton=cancelButton,startButton=saveButton)

        v = self._height - 54
        t = bs.textWidget(parent=self._rootWidget,position=(-10,v),size=(self._width,25),text=self._R.configuringText.replace('${DEVICE}',self._displayName),color=gTitleColor,
                          scale=0.83,hAlign="center",vAlign="top")
        # t = bs.textWidget(parent=self._rootWidget,position=(-10,v),size=(self._width,25),text="Configuring " + self._displayName,color=gTitleColor,
        #                   scale=0.83,hAlign="center",vAlign="top")
        v -= 20

        if self._uniqueID != "#1":
            v -= 20
            v -= self._spacing
            t = bs.textWidget(parent=self._rootWidget,position=(0,v),size=(self._width,50),
                              text=self._R.keyboard2NoteText,
                              scale=self._R.keyboard2NoteScale,color=gInfoTextColor,
                              hAlign="center",vAlign="top")
            v -= 45
        v -= 10
        v -= self._spacing * 2.2

        v += 25
        v -= 42

        hOffs = 175
        dist = 70

        dColor = (0.4,0.4,0.8)
        self._captureButton(pos=(hOffs,v+0.8*dist),color=dColor,button='buttonUp',texture=bs.getTexture('upButton'),scale=1.0)
        self._captureButton(pos=(hOffs-1.2*dist,v),color=dColor,button='buttonLeft',texture=bs.getTexture('leftButton'),scale=1.0)
        self._captureButton(pos=(hOffs+1.2*dist,v),color=dColor,button='buttonRight',texture=bs.getTexture('rightButton'),scale=1.0)
        self._captureButton(pos=(hOffs,v-0.8*dist),color=dColor,button='buttonDown',texture=bs.getTexture('downButton'),scale=1.0)

        if self._uniqueID == "#2":
            self._captureButton(pos=(self._width*0.5,v+0.73*dist),color=(0.4,0.4,0.6),button='buttonStart',texture=bs.getTexture('startButton'),scale=0.8)

        hOffs = self._width-175

        self._captureButton(pos=(hOffs,v+0.8*dist),color=(0.6,0.4,0.8),button='buttonPickUp',texture=bs.getTexture('buttonPickUp'),scale=1.0)
        self._captureButton(pos=(hOffs-1.2*dist,v),color=(0.7,0.5,0.1),button='buttonPunch',texture=bs.getTexture('buttonPunch'),scale=1.0)
        self._captureButton(pos=(hOffs+1.2*dist,v),color=(0.5,0.2,0.1),button='buttonBomb',texture=bs.getTexture('buttonBomb'),scale=1.0)
        self._captureButton(pos=(hOffs,v-0.8*dist),color=(0.2,0.5,0.2),button='buttonJump',texture=bs.getTexture('buttonJump'),scale=1.0)

        # bs.buttonWidget(parent=self._rootWidget,label='Reset',textScale=0.7,color=(0.45,0.4,0.5),
        #                 textColor=(0.65,0.6,0.7),position=(self._width*0.5-40,20),size=(80,30),call=self._reset)

    # def _reset(self):
    #     # simply clear our config dict and rebuild the UI
    #     getConfigDictForInputDevice(self._input,default=False,clear=True)
    #     bs.realTimer(0,self._rebuildUI)

    def _captureButton(self,pos,color,texture,button,scale=1.0):
        baseSize = 79
        b = bs.buttonWidget(parent=self._rootWidget,position=(pos[0]-baseSize*0.5*scale,pos[1]-baseSize*0.5*scale),
                            size=(baseSize*scale,baseSize*scale),
                            texture=texture,label='',color=color)
        # make this in a timer so that it shows up on top of all other buttons
        def foo():
            uiScale = 0.66*scale*2.0
            maxWidth = 76.0*scale
            t = bs.textWidget(parent=self._rootWidget,position=(pos[0]+0.0*scale,pos[1]-(57.0-18.0)*scale),
                              color=(1,1,1,0.3),size=(0,0),hAlign='center',vAlign='top',
                              scale=uiScale, maxWidth=maxWidth,text=bs.getKeyName(self._input,self._settings[button]));
            bs.buttonWidget(edit=b,call=bs.WeakCall(AwaitKeyboardInputWindow,button,t,self._settings))

        bs.realTimer(0,foo)

    def _cancel(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()
        global gCanAskToConfigGamePads
        gCanAskToConfigGamePads = True

    def _save(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        bs.playSound(bs.getSound('gunCocking'))

        dst = getConfigLocationForInputDevice(self._input,default=False)
        dst = dst[0][dst[1]]
        dst.clear()
        # store any values that aren't -1
        for key,val in self._settings.items():
            if val != -1: dst[key] = val
        # if we're allowed to phone home, send this config so we can generate more defaults in the future
        if bs._shouldSubmitDebugInfo(): bsUtils.serverPut('controllerConfig',{'ua':bs.getUserAgentString(),'name':self._name,'config':dst,'v':2})
        bs.applySettings()
        bs.writeConfig()

        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inLeft').getRootWidget()
        global gCanAskToConfigGamePads
        gCanAskToConfigGamePads = True


def getControllerName(controller,returnDefault=True):
    """ given a controller id, returns a sensible player name for it """
    try:
        val = bs.config['Default Player Profiles'][bs.getInputDeviceName(controller)+" "+bs.getInputDeviceUniqueIdentifier(controller)]
        if val == '': raise Exception() # blank name is same as no name
        return val
    except Exception:
        if returnDefault: return bs.getInputDeviceDefaultPlayerName(controller)
        else: return ''


def getConfigLocationForInputDevice(device,default):
    name = bs.getInputDeviceName(device)
    
    if not bs.config.has_key("Controllers"):
        bs.config["Controllers"] = {}
    jsconfig = bs.config["Controllers"]
    
    if not jsconfig.has_key(name):
        jsconfig[name] = {}

    uniqueID = bs.getInputDeviceUniqueIdentifier(device)
    if default:
        if jsconfig[name].has_key(uniqueID):
            del(jsconfig[name][uniqueID])
        if not 'default' in jsconfig[name]: jsconfig[name]['default'] = {}
        return jsconfig[name],'default'
        #dst = jsconfig[name]["default"]
    else:
        if not uniqueID in jsconfig[name]: jsconfig[name][uniqueID] = {}
        return jsconfig[name],uniqueID
        #dst = jsconfig[name][uniqueID]
    #return dst
    

class EditSoundtrackWindow(Window):

    def __init__(self,existingSoundtrack,completionCall,transition='inRight'):

        self._R = R = bs.getResource('editSoundtrackWindow')
        self._existingSoundtrack = existingSoundtrack
        self._completionCall = completionCall

        self._width = 648
        self._height = 560
        spacing = 40
        buttonWidth = 350
        self._rootWidget = bs.containerWidget(size=(self._width,self._height),transition=transition,singleDepth=True)

        cancelButton = b = bs.buttonWidget(parent=self._rootWidget,position=(45,self._height-60),size=(155,60),label=bs.getResource('cancelText'),scale=0.8)
        saveButton = b = bs.buttonWidget(parent=self._rootWidget,position=(self._width-170,self._height-60),size=(175,60),label=bs.getResource('saveText'),scale=0.8)

        t = bs.textWidget(parent=self._rootWidget,position=(0,self._height-50),
                          size=(self._width,25),
                          text=(R.editSoundtrackText if self._existingSoundtrack != None else R.newSoundtrackText),color=gTitleColor,
                          hAlign="center",vAlign="center",maxWidth=280)
        v = self._height - 130

        # make sure config exists
        try: bs.config['Soundtracks']
        except Exception: bs.config['Soundtracks'] = {}

        if self._existingSoundtrack:
            self._soundtrack = copy.deepcopy(bs.config['Soundtracks'][self._existingSoundtrack])

        else: self._soundtrack = {}

        bs.textWidget(parent=self._rootWidget,text=R.nameText,maxWidth=80,
                      position=(10,v+3),color=(0.8,0.8,0.8,1.0),size=(90,30),hAlign='right',vAlign='center')

        # if there's no initial value, find a good initial unused name
        if self._existingSoundtrack != None: initialStr = self._existingSoundtrack
        else:
            i = 1
            while True:
                initialStr = 'Soundtrack '+str(i)
                if initialStr not in bs.config['Soundtracks']: break
                i += 1

        self._textField = bs.textWidget(parent=self._rootWidget,position=(110,v),size=(self._width - 150,30),
                                  text=initialStr,hAlign="left",
                                  vAlign="center",
                                  maxChars=32,
                                  editable=True,padding=4,
                                  useReturnAction=True,
                                  returnAction=self._doItWithSound)

        scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(40,v-365),size=(self._width-80,340))
        self._col = bs.columnWidget(parent=scrollWidget,singleDepth=True)

        bs.containerWidget(edit=scrollWidget,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)
        bs.containerWidget(edit=self._col,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)

        v -= 410

        self._songTypeButtons = {}
        self._refresh()
        bs.buttonWidget(edit=cancelButton,call=self._cancel)
        bs.containerWidget(edit=self._rootWidget,cancelButton=cancelButton)
        bs.buttonWidget(edit=saveButton,call=self._doIt)
        bs.containerWidget(edit=self._rootWidget,startButton=saveButton)

    def _refresh(self):
        for w in bs.getWidgetChildren(self._col):
            bs.deleteWidget(w)
        types = ['Menu',
                 'CharSelect',
                 'ToTheDeath',
                 'Onslaught',
                 'Keep Away',
                 'Race',
                 'Epic Race',
                 'ForwardMarch',
                 'FlagCatcher',
                 'Survival',
                 'Epic',
                 'Hockey',
                 'Football',
                 'Flying',
                 'Scary',
                 'Marching',
                 'GrandRomp',
                 'Chosen One',
                 'Scores',
                 'Victory',
                 ]
        typeNamesTranslated = bs.getResource('soundtrackTypeNames')
        for songType in types:
            r = bs.rowWidget(parent=self._col,size=(self._width-40,40))
            bs.containerWidget(edit=r,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)
            try: typeName = typeNamesTranslated[songType]
            except Exception: typeName = songType
            t = bs.textWidget(parent=r,size=(230,25),alwaysHighlight=True,
                              text=typeName,scale=0.7,hAlign='left',vAlign='center',maxWidth=190)
            try:
                name = self._soundtrack[songType]
                if type(name) is str: nameEncoded = name
                else: nameEncoded = name.encode('utf-8')
                #nameEncoded = name.encode('utf-8')
            except Exception:
                name = None
                nameEncoded = self._R.defaultGameMusicText
            self._songTypeButtons[songType] = b = bs.buttonWidget(parent=r,size=(230,32),label=nameEncoded,textScale=0.6,call=bs.WeakCall(self._getPlaylist,songType,name))
            t = bs.textWidget(parent=r,size=(10,32),text='') # spacing..
            b = bs.buttonWidget(parent=r,size=(50,32),label=self._R.testText+'  ',textScale=0.6,call=bs.WeakCall(self._test,songType))

    def _getPlaylist(self,songType,name):
        GetITunesPlaylistWindow(self._completionCB,songType,name)

    def _test(self,songType):
        # warn if volume is zero
        if bs.getSetting("Music Volume") < 0.01:
            bs.playSound(bs.getSound('error'))
            bs.screenMessage(self._R.musicVolumeZeroWarning,color=(1,0.5,0))
        bsUtils.setMusicPlayMode('test')
        bsUtils.playMusic(songType,'test',self._soundtrack)

    def _completionCB(self,playlist,musicType):
        if playlist is None:
            bs.buttonWidget(edit=self._songTypeButtons[musicType],label=self._R.defaultGameMusicText,call=bs.WeakCall(GetITunesPlaylistWindow,self._completionCB,musicType,playlist))
            if musicType in self._soundtrack: del self._soundtrack[musicType]
        else:
            #bs.buttonWidget(edit=self._songTypeButtons[musicType],label=playlist.encode('utf-8'),call=bs.WeakCall(GetITunesPlaylistWindow,self._completionCB,musicType,playlist))
            # print playlist,type(playlist)
            # print playlist.encode('utf-8'),type(playlist.encode('utf-8'))
            bs.buttonWidget(edit=self._songTypeButtons[musicType],label=playlist,call=bs.WeakCall(GetITunesPlaylistWindow,self._completionCB,musicType,playlist))
            self._soundtrack[musicType] = playlist
        bs.playSound(bs.getSound('swish'))

    def _cancel(self):
        bsUtils.setMusicPlayMode('regular')
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        self._completionCall(None)

    def _doIt(self):

        newName = bs.textWidget(query=self._textField)
        if len(newName) == 0:
            bs.playSound(bs.getSound('error'))
            return
        if newName == self._R.defaultSoundtrackNameText:
            bs.screenMessage(self._R.cantOverwriteDefaultText)
            bs.playSound(bs.getSound('error'))
            return

        # make sure config exists
        try: bs.config['Soundtracks']
        except Exception: bs.config['Soundtracks'] = {}

        # if we had an old one, delete it
        if self._existingSoundtrack != None: del bs.config['Soundtracks'][self._existingSoundtrack]
        bs.config['Soundtracks'][newName] = self._soundtrack
        bs.writeConfig()
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        bsUtils.setMusicPlayMode('regular')
        self._completionCall(newName)

    def _doItWithSound(self):
        bs.playSound(bs.getSound('swish'))
        self._doIt()
    

class GetITunesPlaylistWindow(Window):

    def __init__(self,callback,callbackData,existing):
        self._R = R = bs.getResource('editSoundtrackWindow')
        self._callback = callback
        self._callbackData = callbackData
        self._existing = existing
        self._width = 520
        self._height = 520
        self._spacing = 45
        v = self._height - 90
        v -= self._spacing*1.0
        self._rootWidget = bs.containerWidget(size=(self._width,self._height),transition='inRight',singleDepth=True)
        b = bs.buttonWidget(parent=self._rootWidget,position=(35,self._height-65),size=(130,50),
                            label=bs.getResource('cancelText'),call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        t = bs.textWidget(parent=self._rootWidget,position=(20,self._height-54),size=(self._width,25),
                          text=R.selectAPlaylistText,color=gTitleColor,
                          hAlign="center",vAlign="center",maxWidth=200)
        scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(40,v-340),size=(self._width-80,400))
        self._column = bs.columnWidget(parent=scrollWidget,singleDepth=True)
        bs.textWidget(parent=self._column,size=(self._width-80,22),text=self._R.fetchingITunesText,color=(0.6,0.9,0.6,1.0),scale=0.8)
        bsUtils.getITunesThread().getPlaylists(self._playlistsCB)

    def _playlistsCB(self,playlists):
        if bs.widgetExists(self._column):
            for w in bs.getWidgetChildren(self._column):
                bs.deleteWidget(w)
            bs.textWidget(parent=self._column,size=(self._width-80,22),text=self._R.defaultGameMusicText,color=(0.6,0.8,0.6,1.0),
                          selectable=True,activateAction=bs.Call(self._sel,self._R.defaultGameMusicText),clickActivate=True)
            for p in playlists:
                #pStr = p.encode('utf-8') if type(
                pStr = p.encode('utf-8')
                # if type(p) is unicode: pStr = p.encode('utf-8')
                # else: pStr = p
                # print 'DAMMIT PSTR IS',pStr,type(pStr)
                t = bs.textWidget(parent=self._column,size=(self._width-80,22),text=pStr,
                              selectable=True,activateAction=bs.Call(self._sel,pStr),clickActivate=True)
                if self._existing == pStr: bs.columnWidget(edit=self._column,selectedChild=t,visibleChild=t)

    def _sel(self,item):
        if bs.widgetExists(self._rootWidget):
            bs.containerWidget(edit=self._rootWidget,transition='outLeft')
            self._callback(None if item == self._R.defaultGameMusicText else item,self._callbackData)

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')


class CustomizeSoundtrackWindow(Window):

    def __init__(self,transition='inRight'):

        self._R = R = bs.getResource('editSoundtrackWindow')
        self._width = 500
        self._height = 340
        spacing = 40
        v = self._height - 40
        v -= spacing*1.0
        self._rootWidget = bs.containerWidget(size=(self._width,self._height),transition=transition)
        backButton = b = bs.buttonWidget(parent=self._rootWidget,position=(35,self._height-50),size=(120,60),scale=0.8,label=bs.getResource('backText'),buttonType='back')
        t = bs.textWidget(parent=self._rootWidget,position=(0,self._height-44),size=(self._width,25),text=self._R.soundtrackText,color=gTitleColor,
                          hAlign="center",vAlign="top")
        hspacing = 15
        scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(40,v-170),size=(self._width-80,180))
        self._c = bs.columnWidget(parent=scrollWidget)
        v -= 205

        h = 40
        hspacing = 15
        bColor = (0.6,0.53,0.63)
        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(84,30),
                            call=self._newSoundtrack,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=self._R.newText)
        h += 84+hspacing

        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(90,30),
                            call=self._editSoundtrack,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=self._R.editText)
        h += 90+hspacing

        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(100,30),
                            call=self._duplicateSoundtrack,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=self._R.duplicateText)
        h += 97+hspacing

        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(100,30),
                            call=self._deleteSoundtrack,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=self._R.deleteText)
        h += 100+hspacing
        v -= 47

        self._soundtracks = None
        self._selectedSoundtrack = None
        self._selectedSoundtrackIndex = None
        self._soundtrackWidgets = []
        self._allowChangingSoundtracks = False
        self._refresh()
        bs.buttonWidget(edit=backButton,call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=backButton)

    def _doDeleteSoundtrack(self):
        try: del bs.config['Soundtracks'][self._selectedSoundtrack]
        except Exception: pass
        bs.writeConfig()
        bs.playSound(bs.getSound('shieldDown'))
        self._refresh()

    def _deleteSoundtrack(self):
        if self._selectedSoundtrack is None: return
        #if self._selectedSoundtrack == self._R.defaultSoundtrackNameText:
        if self._selectedSoundtrack == '__default__':
            bs.playSound(bs.getSound('error'))
            bs.screenMessage(self._R.cantDeleteDefaultText)
        else:
            ConfirmWindow(self._R.deleteConfirmText.replace("${NAME}",self._selectedSoundtrack),self._doDeleteSoundtrack,450,150)

    def _duplicateSoundtrack(self):
        try: soundtracks = [bs.config['Soundtracks']]
        except Exception: bs.config['Soundtracks'] = {}

        if self._selectedSoundtrack is None: return
        #if self._selectedSoundtrack == self._R.defaultSoundtrackNameText:
        if self._selectedSoundtrack == '__default__':
            pl = {}
        else:
            pl = bs.config['Soundtracks'][self._selectedSoundtrack]

        # find a valid dup name that doesn't exist
        testIndex = 1
        copyWordc = bs.getResource('copyText')
        copyStr = '' if copyWord in self._selectedSoundtrack else ' '+copyWord
        baseName = self._getSoundtrackDisplayName(self._selectedSoundtrack)
        # if its a copy, strip digits and spaces off the end
        if 'Copy' in baseName:
            while baseName[-1].isdigit() or baseName[-1] == ' ': baseName = baseName[:-1]
        while True:
            testName = baseName + (copyStr if testIndex < 2 else copyStr+' '+str(testIndex))
            if not testName in bs.config['Soundtracks']:
                break
            testIndex += 1

        bs.config['Soundtracks'][testName] = copy.deepcopy(pl)
        bs.writeConfig()
        bs.playSound(bs.getSound('gunCocking'))
        self._refresh(selectSoundtrack=testName)


    def _select(self,name,index):
        self._selectedSoundtrackIndex = index
        self._selectedSoundtrack = name

        try: currentSoundtrack = bs.config['Soundtrack']
        except Exception: currentSoundtrack = '__default__'
        #except Exception: currentSoundtrack = self._R.defaultSoundtrackNameText

        # if it varies from current, write to prefs and apply
        if currentSoundtrack != name and self._allowChangingSoundtracks:
            bs.config['Soundtrack'] = self._selectedSoundtrack
            bs.writeConfig()
            # just play whats already playing.. this'll grab it from the new soundtrack
            bsUtils.playMusic(bsUtils.gMusicTypes['regular'])

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = ConfigAudioWindow(transition='inLeft').getRootWidget()

    def _editSoundtrackWithSound(self):
        bs.playSound(bs.getSound('swish'))
        self._editSoundtrack()

    def _editSoundtrack(self):
        if self._selectedSoundtrack is None: return
        #if self._selectedSoundtrack == self._R.defaultSoundtrackNameText:
        if self._selectedSoundtrack == '__default__':
            bs.playSound(bs.getSound('error'))
            bs.screenMessage(self._R.cantEditDefaultText)
            return
        EditSoundtrackWindow(existingSoundtrack=self._selectedSoundtrack,completionCall=self._editDone)

    def _editDone(self,newSoundtrack):
        if newSoundtrack != None:
            bs.playSound(bs.getSound('gunCocking'))
            self._refresh(selectSoundtrack=newSoundtrack)
            # just play whats already playing.. this'll grab any changes we made..
            bsUtils.playMusic(bsUtils.gMusicTypes['regular'])

    def _getSoundtrackDisplayName(self,soundtrack):
        if soundtrack == '__default__': return self._R.defaultSoundtrackNameText
        else: return soundtrack
        
    def _refresh(self,selectSoundtrack=None):
        self._allowChangingSoundtracks = False
        oldSelection = self._selectedSoundtrack
        # if there was no prev selection, look in prefs
        if oldSelection is None:
            try: oldSelection = bs.config['Soundtrack']
            except Exception: pass
        oldSelectionIndex = self._selectedSoundtrackIndex

        # delete old
        while len(self._soundtrackWidgets) > 0: bs.deleteWidget(self._soundtrackWidgets.pop())
        try: self._soundtracks = bs.config['Soundtracks']
        except Exception: self._soundtracks = {}
        items = self._soundtracks.items()
        items.sort(key=lambda x:x[0].lower())
        #items = [[self._R.defaultSoundtrackNameText,None]] + items # default is always first
        items = [['__default__',None]] + items # default is always first
        index = 0
        for pName,p in items:
            w = bs.textWidget(parent=self._c,size=(self._width-40,24),
                              text=self._getSoundtrackDisplayName(pName),hAlign='left',vAlign='center',
                              selectCall=bs.WeakCall(self._select,pName,index),
                              activateAction=self._editSoundtrackWithSound,
                              selectable=True)
            self._soundtrackWidgets.append(w)
            # select this one if the user requested it
            if selectSoundtrack != None:
                if pName == selectSoundtrack:
                    bs.columnWidget(edit=self._c,selectedChild=w,visibleChild=w)
            else:
                # select this one if it was previously selected
                if oldSelectionIndex != None: # go by index if there's one
                    if index == oldSelectionIndex:
                        bs.columnWidget(edit=self._c,selectedChild=w,visibleChild=w)
                else: # otherwise look by name
                    if pName == oldSelection:
                        bs.columnWidget(edit=self._c,selectedChild=w,visibleChild=w)
            index += 1

        # explicitly run select callback on current one and re-enable callbacks
        self._allowChangingSoundtracks = True
        self._select(self._selectedSoundtrack,self._selectedSoundtrackIndex)

    def _newSoundtrack(self):
        EditSoundtrackWindow(existingSoundtrack=None,completionCall=self._createDone)

    def _createDone(self,newSoundtrack):
        if newSoundtrack is not None:
            bs.playSound(bs.getSound('gunCocking'))
            self._refresh(selectSoundtrack=newSoundtrack)


class ConfigAudioWindow(Window):

    def __init__(self,transition='inRight'):

        self._R = R = bs.getResource('audioSettingsWindow')

        spacing = 50
        width = 460
        height = 210

        ua = bs.getUserAgentString();

        showSoundtracks = False
        if 'Mac' in ua:
            showSoundtracks = True
            height += spacing*2.0

        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,scale=2.0 if gSmallUI else 1.5 if gMedUI else 1.0)
        b = bs.buttonWidget(parent=self._rootWidget,position=(35,height-55),size=(120,60),scale=0.8,textScale=1.2,
                            label=bs.getResource('backText'),buttonType='back',call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        v = height - 60
        v -= spacing*1.0
        t = bs.textWidget(parent=self._rootWidget,position=(0,height-50),size=(width,25),text=R.titleText,color=gTitleColor,
                          hAlign="center",vAlign="top")

        configTextBox(parent=self._rootWidget,position=(40,v),xOffset=10,name="Sound Volume",displayName=R.soundVolumeText,type="float",minVal=0,maxVal=1.0,increment=0.1)
        v -= spacing
        configTextBox(parent=self._rootWidget,position=(40,v),xOffset=10,name="Music Volume",displayName=R.musicVolumeText,type="float",minVal=0,maxVal=1.0,increment=0.1,callback=bsUtils._musicVolumeChanged,changeSound=None)

        if showSoundtracks:
            v -= 1.7*spacing
            b = bs.buttonWidget(parent=self._rootWidget,position=((width-310)/2,v),size=(310,50),
                                label=R.soundtrackButtonText,call=self._doSoundtracks)
            v -= spacing*0.5
            bs.textWidget(parent=self._rootWidget,position=(0,v),size=(width,20),text=R.soundtrackDescriptionText,
                          hAlign='center',scale=0.5,color=(0.5,0.6,0.5,1.0),maxWidth=400)

    def _doSoundtracks(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = CustomizeSoundtrackWindow().getRootWidget()

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = SettingsWindow(transition='inLeft').getRootWidget()


class ConfigGraphicsWindow(Window):

    def __init__(self,transition='inRight'):

        self._R = R = bs.getResource('graphicsSettingsWindow')

        spacing = 32
        self._haveSelectedChild = False
        ua = bs.getUserAgentString()
        interfaceType = bs.getInterfaceType()
        width = 450
        height = 302

        self._showFullscreen = False
        fullscreenSpacingTop = spacing*0.2
        fullscreenSpacing = spacing*1.2
        if interfaceType == 'desktop':
            self._showFullscreen = True
            height += fullscreenSpacing+fullscreenSpacingTop

        showGamma = False
        gammaSpacing = spacing*1.3
        if interfaceType == 'desktop':
            showGamma = True
            height += gammaSpacing

        showVSync = False
        if 'Mac' in ua or 'macos' in ua:
            showVSync = True

        v = height - 50
        v -= spacing*1.15
        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,scale=2.0 if gSmallUI else 1.5 if gMedUI else 1.0)

        b = bs.buttonWidget(parent=self._rootWidget,position=(35,height-50),size=(120,60),scale=0.8,textScale=1.2,
                            label=bs.getResource('backText'),buttonType='back',call=self._back)

        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(0,height-44),size=(width,25),text=R.titleText,color=gTitleColor,
                          hAlign="center",vAlign="top")

        if self._showFullscreen:
            v -= fullscreenSpacingTop
            self._fullscreenCheckBox = configCheckBox(parent=self._rootWidget,position=(100,v),maxWidth=200,
                                                      size=(300,30),name="Fullscreen",displayName=R.fullScreenCmdText if 'Mac' in ua else R.fullScreenCtrlText)
            if not self._haveSelectedChild:
                bs.containerWidget(edit=self._rootWidget,selectedChild=self._fullscreenCheckBox)
                self._haveSelectedChild = True
            v -= fullscreenSpacing

        if showGamma:
            tb = configTextBox(parent=self._rootWidget,position=(90,v),
                               name="Screen Gamma",displayName=R.gammaText,
                               type="float",minVal=0.1,maxVal=2.0,
                               increment=0.1,xOffset=-70,textScale=0.85)
            if not self._haveSelectedChild:
                bs.containerWidget(edit=self._rootWidget,selectedChild=tb)
                self._haveSelectedChild = True
            v -= gammaSpacing

        self._selectedColor=(0.5,1,0.5,1)
        self._unselectedColor=(0.7,0.7,0.7,1)

        # quality
        hOffs = 15
        t = bs.textWidget(parent=self._rootWidget,position=(60,v),size=(160,25),
                          text=R.visualsText,
                          color=gHeadingColor,scale=0.65,maxWidth=150,
                          hAlign="center",vAlign="center")
        PopupMenu(parent=self._rootWidget,position=(60,v-50),width=150,scale=1.3,
                  choices=['Auto','Higher','High','Medium','Low'],
                  choicesDisabled=['Higher','High'] if bs.getMaxGraphicsQuality() == 'Medium' else [],
                  choicesDisplay=[bs.getResource('autoText'),R.higherText,R.highText,R.mediumText,R.lowText],
                  currentChoice=bs.getSetting('Graphics Quality'),changeCall=self._setQuality)

        # texture controls
        hOffs = 210
        t = bs.textWidget(parent=self._rootWidget,position=(230,v),size=(160,25),
                          text=R.texturesText,
                          color=gHeadingColor,scale=0.65,maxWidth=150,
                          hAlign="center",vAlign="center")
        PopupMenu(parent=self._rootWidget,position=(230,v-50),width=150,scale=1.3,
                  choices=['Auto','High','Medium','Low'],
                  choicesDisplay=[bs.getResource('autoText'),R.highText,R.mediumText,R.lowText],
                  currentChoice=bs.getSetting('Texture Quality'),changeCall=self._setTextures)

        v -= 80

        #hOffs = 0 if showVSync else 80
        hOffs = 0

        # resolution settings
        t = bs.textWidget(parent=self._rootWidget,position=(hOffs+60,v),size=(160,25),
                          text=R.resolutionText,
                          color=gHeadingColor,scale=0.65,maxWidth=150,
                          hAlign="center",vAlign="center")

        currentRes = bs.getDisplayResolution()

        # if we're on a system that doesn't allow setting resolution, set pixel-scale instead
        if currentRes is None:
            currentRes = str(min(100,max(10,int(round(bs.getSetting('Screen Pixel Scale')*100.0)))))+'%'
            PopupMenu(parent=self._rootWidget,position=(hOffs+60,v-50),width=120,scale=1.3,
                      choices=['100%','88%','75%','63%','50%'],
                      currentChoice=currentRes,changeCall=self._setPixelScale)
        else:
            resList = ['Desktop Res']+bs.getDisplayResolutions()
            self._resPopup = weakref.ref(PopupMenu(parent=self._rootWidget,position=(hOffs+60,v-50),scale=1.3,
                                                   choices=resList,
                                                   currentChoice=currentRes,changeCall=self._setRes))
        hOffs = 210

        # vsync
        if showVSync:
            t = bs.textWidget(parent=self._rootWidget,position=(230,v),
                              size=(160,25),text=R.verticalSyncText,
                              color=gHeadingColor,scale=0.65,maxWidth=150,
                              hAlign="center",vAlign="center")

            PopupMenu(parent=self._rootWidget,position=(230,v-50),width=150,scale=1.3,
                      choices=['Auto','Always','Never'],
                      choicesDisplay=[bs.getResource('autoText'),R.alwaysText,R.neverText],
                      currentChoice=bs.getSetting('Vertical Sync'),changeCall=self._setVSync)

        v -= 90
        fpsc = configCheckBox(parent=self._rootWidget,position=(69,v-6),
                              size=(210,30),scale=0.86,name="Show FPS",displayName=R.showFPSText,maxWidth=130)
        tvc = configCheckBox(parent=self._rootWidget,position=(240,v-6),
                              size=(210,30),scale=0.86,name="TV Border",displayName=R.tvBorderText,maxWidth=130)
        # fpsc = configCheckBox(parent=self._rootWidget,position=(150,v-7) if showVSync else (243,v+49),
        #                       size=(210,30),scale=1.0 if showVSync else 0.85,name="Show FPS")
        v -= spacing

        # make a timer to update our controls in case the config changes under us
        self._updateTimer = bs.Timer(250,bs.WeakCall(self._updateControls),repeat=True,timeType='real')

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = SettingsWindow(transition='inLeft').getRootWidget()

    def _setQuality(self,quality):
        bs.config['Graphics Quality'] = quality
        bs.applySettings()
        bs.writeConfig()

    def _setTextures(self,val):
        bs.config['Texture Quality'] = val
        bs.applySettings()
        bs.writeConfig()

    def _setPixelScale(self,res):
        bs.config['Screen Pixel Scale'] = float(res[:-1])/100.0
        bs.writeConfig()
        bs.applySettings()
        
    def _setRes(self,res,confirm=True,playSound=False):
        self._oldRes = bs.getDisplayResolution()

        if playSound: bs.playSound(bs.getSound('swish'))

        bs.config['Screen Resolution'] = res

        bs.writeConfig()
        bs.applySettings()

        rp = self._resPopup()
        if rp is not None: rp.setChoice(res)

        # if we're changing resolution in fullscreen mode, bring up a confirmation window to make sure the res is visible
        if self._oldRes != res and confirm and bs.getSetting('Fullscreen'):
            ResChangeConfirmWindow(self)

    def _setVSync(self,val):
        bs.config['Vertical Sync'] = val
        bs.applySettings()
        bs.writeConfig()

    def _updateControls(self):
        if self._showFullscreen:
            bs.checkBoxWidget(edit=self._fullscreenCheckBox,value=bs.getSetting('Fullscreen'))

class ResChangeConfirmWindow(object):

    def __init__(self,configGraphicsWindow):
        self._configGraphicsWindow = configGraphicsWindow
        width = 550
        height = 200
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight')
        self._confirmCounter = 10
        t = bs.textWidget(parent=self._rootWidget,position=(15,height-80),size=(width-30,30),
                          text=bs.getResource('keepTheseSettingsText'),
                          maxWidth=width*0.95,
                          color=(0,1,0,1),
                          scale=1.3,
                          hAlign="center",vAlign="center")
        self._timeText = bs.textWidget(parent=self._rootWidget,position=(15,40),size=(width-30,30),
                                       text=("10"),
                                       hAlign="center",vAlign="top")
        self._revertButton = b = bs.buttonWidget(parent=self._rootWidget,position=(25,40),size=(160,50),label=bs.getResource('revertText'),call=self._revert)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        b = bs.buttonWidget(parent=self._rootWidget,position=(width-190,40),size=(160,50),label=bs.getResource('keepText'),call=self._keep)
        self._timer = bs.Timer(1000,bs.WeakCall(self._tick),repeat=True,timeType='real')

    def _keep(self):
        self._timer = None
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')

    def _revert(self):
        self._timer = None
        self._configGraphicsWindow._setRes(self._configGraphicsWindow._oldRes,confirm=False,playSound=True)
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')

    def _tick(self):
        self._confirmCounter -= 1
        bs.textWidget(edit=self._timeText,text=str(self._confirmCounter))
        if self._confirmCounter == 0:
            bs.activateWidget(self._revertButton)

def _promoCodeCallback(data):
    if data is None:
        bs.screenMessage(bs.getResource('promoSubmitErrorText'),color=(1,0,0))
        bs.playSound(bs.getSound('error'))
    else:
        if data['success']:
            bs.screenMessage(data['message'],color=(0,1,0))
            bs.playSound(bs.getSound('ding'))
        else:
            bs.screenMessage(data['message'],color=(1,0,0))
            bs.playSound(bs.getSound('error'))

class PromoCodeWindow(Window):
    def __init__(self,transition='inRight'):
        
        width = 450
        height = 230
        spacing = 32

        R = bs.getResource('promoCodeWindow')

        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,scale=1.7 if gSmallUI else 1.42 if gMedUI else 1.0)
        b = bs.buttonWidget(parent=self._rootWidget,position=(35,height-55),size=(140,60),scale=0.8,
                            label=bs.getResource('cancelText'),call=self._doBack)

        bs.textWidget(parent=self._rootWidget,text=R.codeText,position=(22,height-113),color=(0.8,0.8,0.8,1.0),size=(90,30),hAlign='right')
        self._textField = bs.textWidget(parent=self._rootWidget,position=(125,height-115),size=(280,36),
                                        text='',hAlign="left",
                                        vAlign="center",
                                        maxChars=64,
                                        color=(0.9,0.9,0.9,1.0),
                                        description=R.codeTextDescription,
                                        editable=True,padding=4,
                                        useReturnAction=True,
                                        returnAction=self._activateEnterButton)

        bWidth = 200
        self._enterButton = b2 = bs.buttonWidget(parent=self._rootWidget,position=(width*0.5-bWidth*0.5,height-200),size=(bWidth,60),scale=1.0,
                            label=R.enterText,call=self._doEnter)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b,startButton=b2,selectedChild=self._textField)

    def _doBack(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')

    def _activateEnterButton(self):
        bs.activateWidget(self._enterButton)

    def _doEnter(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        bs.enterPromoCode(bs.textWidget(query=self._textField),callback=_promoCodeCallback)

class AdvancedSettingsWindow(Window):

    # def __del__(self):
    #     print '~AdvancedSettingsWindow()'

    def _updateLangStatus(self):
        #bs.textWidget(edit=self._langStatusText,text='LANG STATUS UPDATED')
        if self._completeLangsList is not None:
            upToDate = (bs.getLanguage() in self._completeLangsList)
            t = bs.textWidget(edit=self._langStatusText,text='' if bs.getLanguage() == 'Test' else self._R.translationNoUpdateNeededText if upToDate else self._R.translationUpdateNeededText,
                              color=(0.2,1.0,0.2,0.8) if upToDate else (1.0,0.2,0.2,0.8))

    def _rebuild(self):

        # dont rebuild if the menu is open or if our language and language-list hasn't changed
        #if self._menuOpen or (self._prevLang == bs.getLanguage(returnNoneForDefault=True) and self._prevLangList == bs.getLanguages() and self._prevCompleteLangsList == self._completeLangsList): return
        if self._menuOpen or (self._prevLang == bs.getLanguage(returnNoneForDefault=True) and self._prevLangList == bs.getLanguages()): return
        self._prevLang = bs.getLanguage(returnNoneForDefault=True)
        self._prevLangList = bs.getLanguages()
        #self._prevCompleteLangsList = self._completeLangsList

        # if our popup menu was selected we wanna restore that..
        langWasSelected = (self._languagePopup is not None and bs.getWidgetSelectedChild(self._rootWidget) == self._languagePopup.getButtonWidget())

        # clear anything already there
        children = bs.getWidgetChildren(self._rootWidget)
        for c in children: bs.deleteWidget(c)

        self._R = R = bs.getResource('settingsWindowAdvanced')
        b = bs.buttonWidget(parent=self._rootWidget,position=(35,self._height-55),size=(120,60),scale=0.8,
                            label=bs.getResource('backText'),buttonType='back',call=self._doBack)

        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(0,self._height-49),size=(self._width,25),text=R.titleText,color=gTitleColor,
                          hAlign="center",vAlign="top")
        v = self._height - 70


        v -= self._spacing * 1.1

        configCheckBox(parent=self._rootWidget,position=(105,v),size=(self._width-100,30),name="Show Player Names",
                       displayName=R.showPlayerNamesText,scale=1.0,maxWidth=250)
        v -= self._spacing * 1.1
        configCheckBox(parent=self._rootWidget,position=(105,v),size=(self._width-100,30),name="Kick Idle Players",
                       displayName=R.kickIdlePlayersText,scale=1.0,maxWidth=250)
        v -= self._spacing * 1.7


        t = bs.textWidget(parent=self._rootWidget,position=(165,v+10),size=(0,0),text=R.languageText,maxWidth=150,
                          scale=0.95*R.languageTextScale,color=gTitleColor,hAlign='right',vAlign='center')

        languages = bs.getLanguages()
        curLang = bs.getLanguage(returnNoneForDefault=True)
        if curLang is None: curLang = 'Auto'
        
        langsTranslated = {}
        for lang in languages:
            langsTranslated[lang] = lang
            try:
                val = __import__('bsLanguage'+lang).values['translations']['languages'][lang]
                if val is not None: langsTranslated[lang] = val
            except Exception: pass

        langsFull = {}
        for lang in languages:
            lt = bs.translate('languages',lang,printErrors=False)
            if langsTranslated[lang] == lt:
                langsFull[lang] = lt
            else: langsFull[lang] = langsTranslated[lang]+' ('+lt+')'

        self._languagePopup = PopupMenu(parent=self._rootWidget,position=(180,v-13),width=150,scale=1.3,
                                        openingCall=bs.WeakCall(self._onMenuOpen),closingCall=bs.WeakCall(self._onMenuClose),
                                        changeCall=bs.WeakCall(self._onMenuChoice),
                                        choices=['Auto']+languages,buttonSize=(250,50),
                                        choicesDisplay=([bs.getResource('autoText')+' ('+bs.translate('languages',bs.getDefaultLanguage())+')']
                                                        +[langsFull[l] for l in languages]),
                                        currentChoice=curLang)

        if langWasSelected:
            bs.containerWidget(edit=self._rootWidget,selectedChild=self._languagePopup.getButtonWidget())
            
        v -= self._spacing * 1.75

        t = bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,v+10),size=(0,0),text=R.helpTranslateText,maxWidth=self._width*0.9,
                          scale=0.65*R.helpTranslateTextScale,color=(0.4,0.9,0.4,0.8),hAlign='center',vAlign='center')
        v -= self._spacing * 1.51
        thisButtonWidth = 390
        b = bs.buttonWidget(parent=self._rootWidget,
                            position=(self._width/2-thisButtonWidth/2,v-14),
                            size=(thisButtonWidth,40),
                            label=R.translationEditorButtonText,
                            textScale=0.75,
                            call=bs.Call(bs.openURL,'https://acrobattleserver.appspot.com/bsLangList'))

        self._langStatusText = bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,v-30),size=(0,0),
                                             text='',
                                             scale=0.63,
                                             hAlign='center',vAlign='center',maxWidth=400.0)
        self._updateLangStatus()
        # if self._completeLangsList is not None:
        #     upToDate = (bs.getLanguage() in self._completeLangsList)
        #     t = bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,v-30),size=(0,0),
        #                       text=R.translationNoUpdateNeededText if upToDate else R.translationUpdateNeededText,
        #                       scale=0.63,color=(0.2,1.0,0.2,0.8) if upToDate else (1.0,0.2,0.2,0.8),
        #     hAlign='center',vAlign='center',maxWidth=400.0)
        v -= self._spacing * 3.0



        thisButtonWidth = 290
        b = bs.buttonWidget(parent=self._rootWidget,
                            position=(self._width/2-thisButtonWidth*0.7,v-10),
                            size=(thisButtonWidth,70),
                            label=R.showUserModsText,
                            scale=0.7,
                            textScale=1.0,
                            call=bsUtils.showUserScripts)

        b = bs.buttonWidget(parent=self._rootWidget,
                            position=(self._width/2,v-10),
                            size=(thisButtonWidth,70),
                            label=R.moddingGuideText,
                            scale=0.7,
                            textScale=1.0,
                            #call=bs.Call(bs.openURL,'http://www.froemling.net/archives/111'))
                            call=bs.Call(bs.openURL,'http://www.froemling.net/docs/bombsquad-modding-guide'))

        v -= self._spacing * 1.4
        b = bs.buttonWidget(parent=self._rootWidget,
                            position=(self._width/2-thisButtonWidth/2,v-14),
                            size=(thisButtonWidth,40),
                            label=R.enterPromoCodeText,
                            textScale=1.0,
                            call=PromoCodeWindow)

    def _onMenuOpen(self):
        self._menuOpen = True

    def _onMenuClose(self):
        self._menuOpen = False

    def _onMenuChoice(self,choice):
        if choice == 'Auto': bs.setLanguage(None)
        else:
            try:
                # attempt to reload the language module first.. (this makes iterating on a language easier)
                langModule = __import__('bsLanguage'+choice)
                reload(langModule)
            except Exception,e:
                print 'Exception on language switch reload for lang \''+choice+'\':',e
            bs.setLanguage(choice)
        bs.realTimer(100,bs.WeakCall(self._rebuild))

    def _completedLangsCB(self,results):
        if results is not None:
            self._completeLangsList = results['langs']
            bs.realTimer(1,bs.WeakCall(self._updateLangStatus))

    def __init__(self,transition='inRight'):
        # print 'AdvancedSettingsWindow()'
        self._width = 550
        self._height = 480
        self._spacing = 32
        self._menuOpen = False
        self._rootWidget = bs.containerWidget(size=(self._width,self._height),transition=transition,scale=1.48 if gSmallUI else 1.41 if gMedUI else 1.0)
        self._prevLang = ""
        self._prevLangList = []
        #self._prevCompleteLangsList = None
        self._completeLangsList = None
        self._languagePopup = None
        self._rebuild()

        # rebuild periodically to pick up language changes/additions/etc
        self._rebuildTimer = bs.Timer(1000,bs.WeakCall(self._rebuild),repeat=True,timeType='real')

        # fetch the list of completed languages
        bsUtils.serverGet('bsLangGetCompleted',{},callback=bs.WeakCall(self._completedLangsCB))

    def _doBack(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = SettingsWindow(transition='inLeft').getRootWidget()

class SettingsWindow(Window):

    def __init__(self,transition='inRight'):

        width = 420
        height = 465
        spacing = 34
        buttonHeight = 42

        R = bs.getResource('settingsWindow')

        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,scale=1.48 if gSmallUI else 1.25 if gMedUI else 1.0)
        b = bs.buttonWidget(parent=self._rootWidget,position=(35,height-51),size=(120,60),scale=0.8,textScale=1.2,
                            label=bs.getResource('backText'),buttonType='back',call=self._doBack)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(0,height-44),size=(width,25),text=R.titleText,color=gTitleColor,
                          hAlign="center",vAlign="center",maxWidth=140)
        v = height - 80
        v -= 150

        bw = 180
        bh = 175
        xOffs = 36
        xOffs2 = xOffs+bw-7

        def _bTitle(x,y,button,text):
            bs.textWidget(parent=self._rootWidget,text=text,position=(x+bw*0.47,y+bh*0.22),
                          maxWidth=bw*0.7,size=(0,0),hAlign='center',vAlign='center',drawController=button,
                          color=(0.7,0.9,0.7,1.0))
            
        pb = self._profilesButton = b = bs.buttonWidget(parent=self._rootWidget,position=(xOffs,v),size=(bw,bh),buttonType='square',
                                                   label='',call=self._doProfiles)
        _bTitle(xOffs,v,pb,R.playerProfilesText)
        iw = ih = 100
        bs.imageWidget(parent=self._rootWidget,position=(xOffs+bw*0.49-iw*0.5,v+43),size=(iw,ih),
                                   texture=bs.getTexture('cuteSpaz'),drawController=pb)
        #v -= spacing * 1.6
        cb = self._controllersButton = b = bs.buttonWidget(parent=self._rootWidget,position=(xOffs2,v),size=(bw,bh),buttonType='square',
                                                      label='',call=self._doControllers)
        _bTitle(xOffs2,v,cb,R.controllersText)
        iw = ih = 130
        bs.imageWidget(parent=self._rootWidget,position=(xOffs2+bw*0.49-iw*0.5,v+35),size=(iw,ih),
                                   texture=bs.getTexture('controllerIcon'),drawController=cb)

        v -= (bh-10)

        gb = self._graphicsButton = b = bs.buttonWidget(parent=self._rootWidget,position=(xOffs,v),size=(bw,bh),buttonType='square',
                                                   label='',call=self._doGraphics)
        _bTitle(xOffs,v,gb,R.graphicsText)
        iw = ih = 110
        bs.imageWidget(parent=self._rootWidget,position=(xOffs+bw*0.49-iw*0.5,v+42),size=(iw,ih),
                                   texture=bs.getTexture('graphicsIcon'),drawController=gb)
        #v -= spacing * 1.6
        ab = self._audioButton = b = bs.buttonWidget(parent=self._rootWidget,position=(xOffs2,v),size=(bw,bh),buttonType='square',
                                                label='',call=self._doAudio)
        _bTitle(xOffs2,v,ab,R.audioText)
        iw = ih = 120
        bs.imageWidget(parent=self._rootWidget,position=(xOffs2+bw*0.49-iw*0.5+5,v+35),size=(iw,ih),
                       color=(1,1,0),
                       texture=bs.getTexture('audioIcon'),drawController=ab)

        # self._profilesButton = b = bs.buttonWidget(parent=self._rootWidget,position=(53,v),size=(width-110,buttonHeight),
        #                                            label=R.playerProfilesText,call=self._doProfiles)
        # v -= spacing * 1.6
        # self._controllersButton = b = bs.buttonWidget(parent=self._rootWidget,position=(61,v),size=(width-110,buttonHeight),
        #                                               label=R.controllersText,call=self._doControllers)
        # v -= spacing * 1.6

        # self._graphicsButton = b = bs.buttonWidget(parent=self._rootWidget,position=(51,v),size=(width-110,buttonHeight),
        #                                            label=R.graphicsText,call=self._doGraphics)
        # v -= spacing * 1.6
        # self._audioButton = b = bs.buttonWidget(parent=self._rootWidget,position=(57,v),size=(width-110,buttonHeight),
        #                                         label=R.audioText,call=self._doAudio)
        v -= spacing * 1.4
        avb = self._advancedButton = b = bs.buttonWidget(parent=self._rootWidget,position=(59,v),size=(width-110,buttonHeight),
                                                         label=R.advancedText,call=self._doAdvanced,
                                                         color=(0.55,0.5,0.6),
                                                         textColor=(0.65,0.6,0.7),
        )

        bs.buttonWidget(edit=pb,downWidget=gb)
        bs.buttonWidget(edit=cb,downWidget=ab)
        bs.buttonWidget(edit=gb,downWidget=avb,upWidget=pb)
        bs.buttonWidget(edit=ab,downWidget=avb,upWidget=cb)

        # if 0:
        #     v -= spacing * 1.57
        #     configCheckBox(parent=self._rootWidget,position=(60,v),size=(width-100,30),name="Show Player Names",displayName=R.showPlayerNamesText,scale=0.9)
        #     v -= spacing * 1.27
        #     configCheckBox(parent=self._rootWidget,position=(60,v),size=(width-100,30),name="Kick Idle Players",displayName=R.kickIdlePlayersText,scale=0.9)
        #     v -= spacing * 0.63

        #     thisButtonWidth = 140
        #     b = bs.buttonWidget(parent=self._rootWidget,position=(width/2-thisButtonWidth/2,v-14),
        #                         color=(0.45,0.4,0.5),
        #                         size=(thisButtonWidth,22),
        #                         label=R.enterPromoCodeText,
        #                         textColor=(0.55,0.5,0.6),
        #                         textScale=0.7,
        #                         call=PromoCodeWindow)


        # re-select previous if applicable
        selected = None
        try:
            global gSettingsSelection
            selName = gSettingsSelection
            if selName == 'Profiles': selected = self._profilesButton
            elif selName == 'Controllers': selected = self._controllersButton
            elif selName == 'Graphics': selected = self._graphicsButton
            elif selName == 'Audio': selected = self._audioButton
            elif selName == 'Advanced': selected = self._advancedButton

        except Exception: pass
        if selected != None: bs.containerWidget(edit=self._rootWidget,selectedChild=selected)
        else: bs.containerWidget(edit=self._rootWidget,selectedChild=self._profilesButton)

    def _doBack(self):
        self._storeSelection()
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = MainMenuWindow(transition='inLeft').getRootWidget()

    def _doProfiles(self):
        self._storeSelection("Profiles")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigPlayerProfilesWindow().getRootWidget()

    def _doControllers(self):
        self._storeSelection("Controllers")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigControlsWindow(transition='inRight').getRootWidget()

    def _doGraphics(self):
        self._storeSelection("Graphics")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigGraphicsWindow().getRootWidget()

    def _doAudio(self):
        self._storeSelection("Audio")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = ConfigAudioWindow().getRootWidget()

    def _doAdvanced(self):
        self._storeSelection("Advanced")
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = AdvancedSettingsWindow().getRootWidget()

    def _storeSelection(self,value=None):
        # None implies to grab highlighted choice
        if value is None:
            s = bs.getWidgetSelectedChild(self._rootWidget)
            if s == self._profilesButton: value = 'Profiles'
            elif s == self._controllersButton: value = 'Controllers'
            elif s == self._graphicsButton: value = 'Graphics'
            elif s == self._audioButton: value = 'Audio'
            elif s == self._advancedButton: value = 'Advanced'
            else: value = 'Profiles'
        if value != None:
            global gSettingsSelection
            gSettingsSelection = value


class SelectMapWindow(Window):

    def __init__(self,gameClass,sessionType,config,editInfo,completionCall,transition='inRight'):

        self._gameClass = gameClass
        self._sessionType = sessionType
        self._config = config
        self._completionCall = completionCall
        self._editInfo = editInfo

        mapList = gameClass.getSupportedMaps(sessionType)

        try: previousMap = bsMap.getFilteredMapName(config['map'])
        except Exception: previousMap = ''

        width = 615
        height = 480 if gSmallUI else 600
        spacing = 40

        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,
                                              scale=1.66 if gSmallUI else 1.1 if gMedUI else 1.0,
                                              stackOffset=(0,-50) if gSmallUI else (0,0))

        b = bs.buttonWidget(parent=self._rootWidget,position=(45,height-60),size=(130,60),scale=0.9,textScale=1.2,
                            label=bs.getResource('backText'),buttonType='back',call=self._cancel)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)
        t = bs.textWidget(parent=self._rootWidget,position=(0,height-49),size=(width,25),maxWidth=260,
                          text=bs.getResource('mapSelectTitleText').replace('${GAME}',gameClass.getNameTranslated()),color=gTitleColor,
        # t = bs.textWidget(parent=self._rootWidget,position=(0,height-49),size=(width,25),text=gameClass.getName()+" Maps",color=gTitleColor,
                          scale=0.9,hAlign="center",vAlign="top")
        v = height - 70
        scrollHeight = height - 140
        scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(40,v-scrollHeight),size=(width-80,scrollHeight))
        bs.containerWidget(edit=self._rootWidget,selectedChild=scrollWidget)
        c = bs.columnWidget(parent=scrollWidget)
        v -= scrollHeight
        modelOpaque = bs.getModel('levelSelectButtonOpaque')
        modelTransparent = bs.getModel('levelSelectButtonTransparent')
        mapListSorted = list(mapList)
        mapListSorted.sort()
        for m in mapListSorted:
            b = bs.buttonWidget(parent=c,size=(500,125),label=bsMap.getTranslatedMapName(m),call=bs.WeakCall(self._selectWithDelay,m),
                                texture=bsMap.getMapClass(m).getPreviewTexture(),
                                color=(0.5,0.5,0.5),
                                modelOpaque=modelOpaque,
                                modelTransparent=modelTransparent)
            if m == previousMap:
                bs.columnWidget(edit=c,selectedChild=b,visibleChild=b)

    def _select(self,mapName):
        self._config['map'] = mapName
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = GameSettingsSelectWindow(self._gameClass,self._sessionType,self._config,
                                                               self._editInfo,self._completionCall,
                                                               defaultSelection='map',
                                                               transition='inLeft').getRootWidget()
    def _selectWithDelay(self,mapName):
        bs.lockAllInput()
        bs.realTimer(100,bs.unlockAllInput)
        bs.realTimer(100,bs.WeakCall(self._select,mapName))

    def _cancel(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = GameSettingsSelectWindow(self._gameClass,self._sessionType,self._config,
                                                               self._editInfo,self._completionCall,
                                                               defaultSelection='map',
                                                               transition='inLeft').getRootWidget()


class GameSettingsSelectWindow(Window):

    def __init__(self,gameClass,sessionType,config,editInfo,completionCall,defaultSelection=None,transition='inRight'):
        self._gameClass = gameClass
        self._sessionType = sessionType
        self._editInfo = editInfo
        self._R = bs.getResource('gameSettingsWindow')

        validMaps = gameClass.getSupportedMaps(sessionType)
        if len(validMaps) == 0:
            bs.screenMessage(bs.getResource('noValidMapsErrorText'))
            raise Exception("No valid maps")

        self._settingsDefs = gameClass.getSettings(sessionType)
        self._completionCall = completionCall

        self._map = validMaps[random.randrange(len(validMaps))]

        isAdd = (editInfo['editType'] == 'add')

        # if there's a valid level name in the existing config, use that.
        try:
            filteredMapName = bsMap.getFilteredMapName(config['map'])
            if filteredMapName in validMaps: self._map = filteredMapName
        except Exception: pass

        try: self._settings = config['settings']
        except Exception: self._settings = {}

        self._choiceSelections = {}

        width = 620
        height = 460
        spacing = 52
        # yExtra = 0 if gSmallUI else 15
        # yExtra2 = 0 if gSmallUI else 21
        yExtra =15
        yExtra2 = 21

        levelTex = bsMap.getMapClass(self._map).getPreviewTexture()
        if levelTex is None: raise Exception("no map preview tex found for"+self._map)

        self._rootWidget =  bs.containerWidget(size=(width,height),transition=transition,
                                               scale=1.68 if gSmallUI else 1.35 if gMedUI else 1.0,
                                               stackOffset=(0,-31) if gSmallUI else (0,0))

        # if isAdd:
        b = bs.buttonWidget(parent=self._rootWidget,
                            position=(30,height-82+yExtra2),
                            size=(180,70) if isAdd else (180,65),
                            # position=(40,height-82+yExtra2) if isAdd else (40,height-79+yExtra2),
                            # size=(140,65) if isAdd else (140,60),
                            label=bs.getResource('backText') if isAdd else bs.getResource('cancelText'),
                            buttonType='back' if isAdd else None,
                            scale=0.75,textScale=1.3,
                            call=bs.Call(self._cancel))
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        addButton = b = bs.buttonWidget(parent=self._rootWidget,
                                        position=(width-183,height-82+yExtra2),
                                        size=(200,65),
                                        # position=(width-178,height-79+yExtra2) if isAdd else (width-150,height-79+yExtra2),
                                        # size=(180,60) if isAdd else (140,60),
                                        scale=0.75,
                                        textScale=1.3,
                                        label=self._R.addGameText if isAdd else bs.getResource('doneText'))
        t = bs.textWidget(parent=self._rootWidget,position=(-8,height-71+yExtra2),size=(width,25),
                          text=gameClass.getNameTranslated(),
                          color=gTitleColor,maxWidth=235,
                          scale=1.0,hAlign="center",vAlign="center")

        mapHeight = 80

        scrollHeight = mapHeight + 10 # map select and margin
        # calc our total height we'll need
        for settingName,setting in self._settingsDefs:
            scrollHeight += spacing

        scrollWidth = width-86
        self._scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(44,35+yExtra),size=(scrollWidth,height-116))
        self._subContainer = c = bs.containerWidget(parent=self._scrollWidget,size=(scrollWidth,scrollHeight),singleDepth=True,background=False)

        # so selection loops through everything and doesn't get stuck in sub-containers
        bs.containerWidget(edit=self._scrollWidget,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)
        bs.containerWidget(edit=c,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)

        v = scrollHeight - 5
        h = -40

        # map select button
        t = bs.textWidget(parent=self._subContainer,position=(h+49,v-63),size=(100,30),maxWidth=110,
                          text=bs.getResource('mapText'),hAlign="left",color=(0.8,0.8,0.8,1.0),vAlign="center")

        mapName = bsMap.getTranslatedMapName(self._map)

        b = bs.imageWidget(parent=self._subContainer,size=(500*0.56,125*0.56),
                           position=(h+261-113,v-84),
                           texture=levelTex,
                           modelOpaque=bs.getModel('levelSelectButtonOpaque'),
                           modelTransparent=bs.getModel('levelSelectButtonTransparent'))
        mapButton = b = bs.buttonWidget(parent=self._subContainer,size=(140,60),
                                        position=(h+448,v-72),
                                        call=bs.Call(self._selectMap),
                                        scale=0.7,
                                        label=bs.getResource('mapSelectText'))
        levelName = bs.textWidget(parent=self._subContainer,position=(h+363-123,v-78),size=(100,30),
                                  scale=0.6,text=bsMap.getTranslatedMapName(self._map),hAlign="center",color=(0.6,1.0,0.6,1.0),vAlign="center")
        v -= mapHeight

        for settingName,setting in self._settingsDefs:
            value = setting['default']
            valueType = type(value)

            # now if there's an existing value for it in the config, override with that
            try: value = valueType(config['settings'][settingName])
            except Exception: pass

            # shove the starting value in there to start..
            self._settings[settingName] = value


            nameTranslated = self._getTranslatedSettingName(settingName)

            mw1 = 280
            mw2 = 70
            # handle types with choices specially:
            if 'choices' in setting:

                for choice in setting['choices']:
                    if len(choice) != 2: raise Exception("Expected 2-member tuples for 'choices'; got: "+repr(choice))
                    if type(choice[0]) not in (str,unicode): raise Exception("First value for choice tuple must be a str; got: "+repr(choice))
                    if type(choice[1]) is not valueType: raise Exception("Choice type does not match default value; choice:"+repr(choice)+"; setting:"+repr(setting))
                if valueType not in (int,float): raise Exception("Choice type setting must have int or float default; got: "+repr(setting))

                # start at the choice corresponding to the default if possible
                self._choiceSelections[settingName] = 0
                found = False
                for index,choice in enumerate(setting['choices']):
                    if choice[1] == value:
                        self._choiceSelections[settingName] = index
                        break;

                v -= spacing
                t = bs.textWidget(parent=self._subContainer,position=(h+50,v),size=(100,30),maxWidth=mw1,
                                  text=nameTranslated,hAlign="left",color=(0.8,0.8,0.8,1.0),vAlign="center")
                t = bs.textWidget(parent=self._subContainer,position=(h+509-95,v),size=(0,28),
                                  text=self._getTranslatedSettingName(setting['choices'][self._choiceSelections[settingName]][0]),editable=False,
                                  color=(0.6,1.0,0.6,1.0),maxWidth=mw2,
                                  hAlign="right",vAlign="center",padding=2)
                b = bs.buttonWidget(parent=self._subContainer,position=(h+509-50-1,v),size=(28,28),label="< ",
                                    call=bs.Call(self._choiceInc,settingName,t,setting,-1),repeat=True)
                b = bs.buttonWidget(parent=self._subContainer,position=(h+509+5,v),size=(28,28),label="> ",
                                    call=bs.Call(self._choiceInc,settingName,t,setting,1),repeat=True)

            elif valueType in [int,float]:
                v -= spacing
                try: minValue = setting['minValue']
                except Exception: minValue = 0
                try: maxValue = setting['maxValue']
                except Exception: maxValue = 9999
                try: increment = setting['increment']
                except Exception: increment = 1
                t = bs.textWidget(parent=self._subContainer,position=(h+50,v),size=(100,30),
                                  text=nameTranslated,hAlign="left",color=(0.8,0.8,0.8,1.0),vAlign="center",maxWidth=mw1)
                t = bs.textWidget(parent=self._subContainer,position=(h+509-95,v),size=(0,28),
                                  text=str(value),editable=False,
                                  color=(0.6,1.0,0.6,1.0),maxWidth=mw2,
                                  hAlign="right",vAlign="center",padding=2)
                b = bs.buttonWidget(parent=self._subContainer,position=(h+509-50-1,v),size=(28,28),label="- ",
                                    call=bs.Call(self._inc,t,minValue,maxValue,-increment,valueType,settingName),repeat=True)
                b = bs.buttonWidget(parent=self._subContainer,position=(h+509+5,v),size=(28,28),label="+ ",
                                    call=bs.Call(self._inc,t,minValue,maxValue,increment,valueType,settingName),repeat=True)
            elif valueType == bool:
                v -= spacing
                t = bs.textWidget(parent=self._subContainer,position=(h+50,v),size=(100,30),
                                  text=nameTranslated,hAlign="left",color=(0.8,0.8,0.8,1.0),vAlign="center",maxWidth=mw1)
                t = bs.textWidget(parent=self._subContainer,position=(h+509-95,v),size=(0,28),
                                  text=bs.getResource('onText') if value else bs.getResource('offText'),editable=False,
                                  color=(0.6,1.0,0.6,1.0),maxWidth=mw2,
                                  hAlign="right",vAlign="center",padding=2)
                c = bs.checkBoxWidget(parent=self._subContainer,text='',position=(h+505-50-5,v-2),size=(200,30),
                                      textColor=(0.8,0.8,0.8),
                                      value=value,valueCall=bs.Call(self._checkValueChange,settingName,t))
            else: raise Exception()

        bs.buttonWidget(edit=addButton,call=bs.Call(self._add))
        bs.containerWidget(edit=self._rootWidget,selectedChild=addButton,startButton=addButton)

        if defaultSelection == 'map':
            bs.containerWidget(edit=self._rootWidget,selectedChild=self._scrollWidget)
            bs.containerWidget(edit=self._subContainer,selectedChild=mapButton)

    def _getTranslatedSettingName(self,name):
        try: nameTranslated = bs.getResource('translations.settingNames')[name]
        except Exception: nameTranslated = None
        if nameTranslated is None: nameTranslated = name
        return nameTranslated
        
    def _selectMap(self):
        # replace ourself with the map-select UI
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = SelectMapWindow(self._gameClass,self._sessionType,copy.deepcopy(self._getConfig()),
                                                  self._editInfo,self._completionCall).getRootWidget()
    def _choiceInc(self,settingName,widget,setting,increment):
        choices = setting['choices']
        if increment > 0: self._choiceSelections[settingName] = min(len(choices)-1,self._choiceSelections[settingName]+1)
        else: self._choiceSelections[settingName] = max(0,self._choiceSelections[settingName]-1)
        bs.textWidget(edit=widget,text=self._getTranslatedSettingName(choices[self._choiceSelections[settingName]][0]))
        self._settings[settingName] = choices[self._choiceSelections[settingName]][1]

    def _cancel(self):
        self._completionCall(None)

    def _checkValueChange(self,settingName,widget,value):
        bs.textWidget(edit=widget,text=bs.getResource('onText') if value else bs.getResource('offText'))
        self._settings[settingName] = value

    def _getConfig(self):
        return {'map':self._map,'settings':self._settings}

    def _add(self):
        self._completionCall(copy.deepcopy(self._getConfig()))

    def _inc(self,ctrl,minVal,maxVal,increment,settingType,settingName):
        if settingType == float: val = float(bs.textWidget(query=ctrl))
        else: val = int(bs.textWidget(query=ctrl))
        val += increment
        val = max(minVal,min(val,maxVal))
        if settingType == float: bs.textWidget(edit=ctrl,text=str(round(val,2)))
        elif settingType == int: bs.textWidget(edit=ctrl,text=str(int(val)))
        else: raise Exception('invalid vartype: '+str(vartype))
        self._settings[settingName] = val


class AddGameWindow(Window):

    def __init__(self,editSession,transition='inRight'):
        self._editSession = editSession
        self._R = R = bs.getResource('addGameWindow')
        width = 500
        height = 360 if gSmallUI else 410
        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,
                                              scale=2.15 if gSmallUI else 1.5 if gMedUI else 1.0,
                                              stackOffset=(0,-52) if gSmallUI else (0,0))
        
        backButton = b = bs.buttonWidget(parent=self._rootWidget,position=(39,height-53),
                                         size=(165,70),scale=0.75,textScale=1.2,label=bs.getResource('backText'),
                                         buttonType='back',call=self._back)
        selectButton = b = bs.buttonWidget(parent=self._rootWidget,position=(width-140,height-50),
                                           size=(160,60),scale=0.75,textScale=1.2,label=bs.getResource('selectText'),call=self._add)
        v = height - 40
        bs.textWidget(parent=self._rootWidget,position=(10,v-4),size=(width,30),scale=1.0,
                      text=R.titleText,hAlign="center",color=gTitleColor,maxWidth=180,
                      vAlign="center")
        v -= 40

        v += 15
        self._selectedTitleText = bs.textWidget(parent=self._rootWidget,position=(40,v),size=(0,0),
                                                scale=1.0,color=(0.7,1.0,0.7,1.0),maxWidth=width-70,
                                                hAlign="left")
        v -= 30

        self._selectedDescriptionText = bs.textWidget(parent=self._rootWidget,position=(40,v),size=(0,0),
                                                      scale=0.7,color=(0.5,0.8,0.5,1.0),maxWidth=width-70,
                                                      hAlign="left")

        v -= 95
        v -= 94
        v -= 75
        if gSmallUI: v += 50


        scrollWidget = s = bs.scrollWidget(parent=self._rootWidget,position=(30,v),size=(width-60,150 if gSmallUI else 200))
        col = bs.columnWidget(parent=s)

        gameTypes = [t for t in bs.getGameTypes() if t.supportsSessionType(self._editSession._sessionType)]

        # sort in this language
        gameTypes.sort(key=lambda g:g.getNameTranslated())

        v -= 35
        bs.containerWidget(edit=self._rootWidget,cancelButton=backButton,startButton=selectButton)
        self._selectedGameType = None

        for gameType in gameTypes:
            bs.textWidget(parent=col,position=(0,0),size=(width-88,24),text=gameType.getNameTranslated(),
                          hAlign="left",vAlign="bottom",
                          color=(0.8,0.8,0.8,1.0),
                          selectCall=bs.Call(self._setSelectedGameType,gameType),
                          alwaysHighlight=True,
                          selectable=True,activateAction=self._addWithSoundAndDelay)
        bs.containerWidget(edit=self._rootWidget,selectedChild=scrollWidget)
        

    def _add(self):
        self._editSession._addGameTypeSelected(self._selectedGameType)

    def _addWithSoundAndDelay(self):
        bs.playSound(bs.getSound('swish'))
        bs.lockAllInput() # make sure no more commands happen
        bs.realTimer(100,bs.unlockAllInput)
        bs.realTimer(100,self._add)

    def _setSelectedGameType(self,gameType):
        self._selectedGameType = gameType
        bs.textWidget(edit=self._selectedTitleText,text=gameType.getNameTranslated())
        bs.textWidget(edit=self._selectedDescriptionText,text=gameType.getDescriptionTranslated(self._editSession.getSessionType()))

    def _back(self):
        self._editSession._addGameCancelled()


class EditPlaylistWindow(Window):

    def __init__(self,editSession,transition='inRight'):

        self._editSession = editSession

        self._R = R = bs.getResource('editGameListWindow')

        try: prevSelection = self._editSession._editUISelection
        except Exception: prevSelection = None

        self._width = 615
        self._height = 420 if gSmallUI else 470
        spacing = 40
        buttonWidth = 350
        self._rootWidget = bs.containerWidget(size=(self._width,self._height),
                                              transition=transition,
                                              scale=1.8 if gSmallUI else 1.3 if gMedUI else 1.0,
                                              stackOffset=(0,-20) if gSmallUI else (0,0))

        cancelButton = b = bs.buttonWidget(parent=self._rootWidget,position=(35,self._height-60),scale=0.8,size=(175,60),
                                           label=bs.getResource('cancelText'),textScale=1.2)
        saveButton = b = bs.buttonWidget(parent=self._rootWidget,position=(self._width-195,self._height-60),scale=0.8,size=(200,60),
                                         label=R.saveText,textScale=1.2)

        t = bs.textWidget(parent=self._rootWidget,position=(-10,self._height-50),size=(self._width,25),
                          text=R.gameListText,color=gTitleColor,
                          hAlign="center",vAlign="center",maxWidth=230)
        v = self._height - 115

        bs.textWidget(parent=self._rootWidget,text=R.nameText,position=(15,v),maxWidth=70,
                      color=(0.8,0.8,0.8,0.5),size=(90,38),scale=0.9,hAlign='right',vAlign='center')

        self._textField = bs.textWidget(parent=self._rootWidget,position=(110,v),size=(self._width - 150,36),
                                        text=self._editSession._name,hAlign="left",
                                        vAlign="center",
                                        maxChars=32,
                                        color=(0.9,0.9,0.9,1.0),
                                        description=R.listNameText,
                                        editable=True,padding=4,
                                        useReturnAction=True,
                                        returnAction=self._savePressWithSound)


        scrollHeight = 215 if gSmallUI else 265
        scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(40,v-scrollHeight),
                                       selectCall=bs.Call(self._setUISelection,'gameList'),
                                       size=(self._width-80,(scrollHeight-15)))
        self._columnWidget = bs.columnWidget(parent=scrollWidget,printListExitInstructions=True)
        v -= (scrollHeight+45)

        self._listWidgets = []


        h = 40
        hspacing = 17

        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(22,30),
                            call=self._moveUp,
                            label="\xc2\x82 ",
                            repeat=True)
        h += 33+hspacing
        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(22,30),
                            call=self._moveDown,
                            label="\xc2\x83 ",
                            repeat=True)
        h += 38+hspacing

        addGameButton = bs.buttonWidget(parent=self._rootWidget,position=(h,v-7),size=(135,45),
                                        call=self._add,
                                        selectCall=bs.Call(self._setUISelection,'addButton'),
                                        textScale=0.8,
                                        label=R.addGameText)
        h += 130+hspacing

        editGameButton = bs.buttonWidget(parent=self._rootWidget,position=(h,v-7),size=(135,45),
                                         call=self._edit,
                                         selectCall=bs.Call(self._setUISelection,'editButton'),
                                         textScale=0.8,
                                         label=R.editGameText)
        h += 130+hspacing

        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v-7),size=(135,45),
                            textScale=0.8,
                            call=self._remove,
                            label=R.removeGameText)
        #h += 100+hspacing

        self._refresh()

        bs.buttonWidget(edit=cancelButton,call=self._cancel)
        bs.containerWidget(edit=self._rootWidget,cancelButton=cancelButton,selectedChild=scrollWidget)

        bs.buttonWidget(edit=saveButton,call=self._savePress)
        bs.containerWidget(edit=self._rootWidget,startButton=saveButton)

        if prevSelection == 'addButton': bs.containerWidget(edit=self._rootWidget,selectedChild=addGameButton)
        elif prevSelection == 'editButton': bs.containerWidget(edit=self._rootWidget,selectedChild=editGameButton)
        elif prevSelection == 'gameList': bs.containerWidget(edit=self._rootWidget,selectedChild=scrollWidget)

    def _setUISelection(self,selection):
        self._editSession._editUISelection = selection

    def _cancel(self):
        bs.playSound(bs.getSound('powerdown01'))
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = PlaylistWindow(transition='inLeft',playlistType=self._editSession._playlistType,
                                                     selectPlaylist=self._editSession._existingPlaylist).getRootWidget()
    def _add(self):
        # store list name then tell the session to perform an add
        self._editSession._name = bs.textWidget(query=self._textField)
        self._editSession._addGamePressed()

    def _edit(self):
        # store list name then tell the session to perform an add
        self._editSession._name = bs.textWidget(query=self._textField)
        self._editSession._editGamePressed()

    def _editWithSound(self):
        bs.playSound(bs.getSound('swish'))
        self._edit()
        
    def _savePress(self):
        newName = bs.textWidget(query=self._textField)
        if len(newName) == 0:
            bs.playSound(bs.getSound('error'))
            return
        if len(self._editSession._playlist) == 0:
            bs.screenMessage(self._R.cantSaveEmptyListText)
            bs.playSound(bs.getSound('error'))
            return
        if newName == self._editSession._defaultListName:
            bs.screenMessage(self._R.cantOverwriteDefaultText)
            bs.playSound(bs.getSound('error'))
            return

        # if we had an old one, delete it
        if self._editSession._existingPlaylist != None: del bs.config[self._editSession._configName+' Playlists'][self._editSession._existingPlaylist]
        bs.config[self._editSession._configName+' Playlists'][newName] = self._editSession._playlist
        bs.writeConfig()
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        bs.playSound(bs.getSound('gunCocking'))
        uiGlobals['mainMenuWindow'] = PlaylistWindow(transition='inLeft',playlistType=self._editSession._playlistType,
                                                     selectPlaylist=newName).getRootWidget()
    def _savePressWithSound(self):
        bs.playSound(bs.getSound('swish'))
        self._savePress()

    def _select(self,index):
        self._editSession._selectedIndex = index

    def _refresh(self):

        # need to grab this here as rebuilding the list will change it otherwise
        oldSelectionIndex = self._editSession._selectedIndex

        while len(self._listWidgets) > 0: bs.deleteWidget(self._listWidgets.pop())
        for index,p in enumerate(self._editSession._playlist):

            try: desc = bsUtils.resolveTypeName(p['type']).getConfigDescription(p)
            except Exception: desc = "(invalid: '"+p['type']+"')"

            w = bs.textWidget(parent=self._columnWidget,size=(self._width-80,24),
                              selectCall=bs.Call(self._select,index),
                              alwaysHighlight=True,
                              color=(0.8,0.8,0.8,1.0),
                              padding=0,
                              text=desc,
                              activateAction=self._editWithSound,
                              vAlign='bottom',selectable=True)
            self._listWidgets.append(w)
            if oldSelectionIndex == index:
                bs.columnWidget(edit=self._columnWidget,selectedChild=w,visibleChild=w)

    def _moveDown(self):
        if self._editSession._selectedIndex >= len(self._editSession._playlist)-1: return
        i = self._editSession._selectedIndex
        tmp = self._editSession._playlist[i]
        self._editSession._playlist[i] = self._editSession._playlist[i+1]
        self._editSession._playlist[i+1] = tmp
        self._editSession._selectedIndex += 1
        self._refresh()

    def _moveUp(self):
        if self._editSession._selectedIndex < 1: return
        i = self._editSession._selectedIndex
        tmp = self._editSession._playlist[i]
        self._editSession._playlist[i] = self._editSession._playlist[i-1]
        self._editSession._playlist[i-1] = tmp
        self._editSession._selectedIndex -= 1
        self._refresh()

    def _remove(self):
        if len(self._editSession._playlist) == 0: return
        del self._editSession._playlist[self._editSession._selectedIndex]
        if self._editSession._selectedIndex >= len(self._editSession._playlist): self._editSession._selectedIndex = len(self._editSession._playlist)-1
        bs.playSound(bs.getSound('shieldDown'))
        self._refresh()

def _setupPlaylistTypeVars(obj,playlistType):
    if playlistType == 'teams':
        obj._defaultListName = bs.getResource('defaultTeamGameListNameText')
        obj._defaultNewListName = bs.getResource('defaultNewTeamGameListNameText')
        obj._getDefaultListCall = bs.getDefaultTeamsPlaylist
        obj._sessionTypeName = 'bsTeamGame.TeamsSession'
        obj._configName = 'Team Tournament'
        obj._windowTitleName = bs.getResource('teamsText')
        obj._sessionType = 'teams'
    elif playlistType == 'freeForAll':
        obj._defaultListName = bs.getResource('defaultFreeForAllGameListNameText')
        obj._defaultNewListName = bs.getResource('defaultNewFreeForAllGameListNameText')
        obj._getDefaultListCall = bs.getDefaultFreeForAllPlaylist
        obj._sessionTypeName = 'bsTeamGame.FreeForAllSession'
        obj._configName = 'Free-for-All'
        obj._windowTitleName = bs.getResource('freeForAllText')
        obj._sessionType = 'freeForAll'
    else: raise Exception('invalid playlist type: '+playlistType)


class GameListEditSession(object):

    def __init__(self,existingPlaylist=None,transition='inRight',playlist=None,playlistName=None,playlistType='teams'):

        # since we may be showing our map list momentarily,
        # lets go ahead and preload all map preview textures
        bsMap.preloadPreviewMedia()
        self._playlistType = playlistType
        _setupPlaylistTypeVars(self,playlistType)
        self._existingPlaylist = existingPlaylist

        # make sure config exists
        try: self._playlists = bs.config[self._configName+' Playlists']
        except Exception: self._playlists = bs.config[self._configName+' Playlists'] = {}
        self._selectedIndex = 0
        if existingPlaylist:
            self._name = existingPlaylist
            # filter out invalid games
            self._playlist = bs.filterPlaylist(copy.deepcopy(bs.config[self._configName+' Playlists'][existingPlaylist]),sessionType=playlistType)
        else:
            if playlist is not None: self._playlist = playlist
            else: self._playlist = []
            if playlistName is not None: self._name = playlistName
            else:
                # find a good unused name
                i = 1
                while True:
                    self._name = self._defaultNewListName+((' '+str(i)) if i > 1 else '')
                    if self._name not in bs.config[self._configName+' Playlists']: break
                    i += 1
            # also we want it to come up with 'add' highlighted since its empty and thats all they can do
            self._editUISelection = 'addButton'

        uiGlobals['mainMenuWindow'] = EditPlaylistWindow(editSession=self,transition=transition).getRootWidget()

    def getSessionType(self):
        return self._playlistType

    def _addGamePressed(self):
        bs.containerWidget(edit=uiGlobals['mainMenuWindow'],transition='outLeft')
        uiGlobals['mainMenuWindow'] = AddGameWindow(editSession=self).getRootWidget()

    def _editGamePressed(self):
        if len(self._playlist) == 0: return
        self._showEditUI(gameType=bsUtils.resolveTypeName(self._playlist[self._selectedIndex]['type']),
                         config=self._playlist[self._selectedIndex])

    def _addGameCancelled(self):
        bs.containerWidget(edit=uiGlobals['mainMenuWindow'],transition='outRight')
        uiGlobals['mainMenuWindow'] = EditPlaylistWindow(editSession=self,transition='inLeft').getRootWidget()

    def _showEditUI(self,gameType,config):
        self._editingGame = (config is not None)
        self._editingGameType = gameType
        gameType.configUI(self._sessionType,copy.deepcopy(config),{'editType':'edit' if self._editingGame else 'add'},self._editGameDone)

    def _addGameTypeSelected(self,gameType):
        self._showEditUI(gameType=gameType,config=None)

    def _editGameDone(self,config):
        if config is None:
            # if we were editing, go back to our list
            if self._editingGame:
                bs.containerWidget(edit=uiGlobals['mainMenuWindow'],transition='outRight')
                uiGlobals['mainMenuWindow'] = EditPlaylistWindow(editSession=self,transition='inLeft').getRootWidget()
            # otherwise we were adding; go back to the add type choice list
            else:
                bs.containerWidget(edit=uiGlobals['mainMenuWindow'],transition='outRight')
                uiGlobals['mainMenuWindow'] = AddGameWindow(editSession=self,transition='inLeft').getRootWidget()
        else:
            # make sure type is in there..
            config['type'] = bsUtils.getTypeName(self._editingGameType)

            if self._editingGame:
                self._playlist[self._selectedIndex] = copy.deepcopy(config)
            else:
                # add a new entry to the playlist..
                insertIndex = min(len(self._playlist),self._selectedIndex+1)
                self._playlist.insert(insertIndex,copy.deepcopy(config))
                self._selectedIndex = insertIndex

            if not self._editingGame:
                bs.playSound(bs.getSound('gunCocking'))
            bs.containerWidget(edit=uiGlobals['mainMenuWindow'],transition='outRight')
            uiGlobals['mainMenuWindow'] = EditPlaylistWindow(editSession=self,transition='inLeft').getRootWidget()


def standardGameConfigUI(gameClass,sessionType,config,editInfo,completionCall):
    # replace the main window once we come up successfully
    prevWindow = uiGlobals['mainMenuWindow']
    uiGlobals['mainMenuWindow'] = GameSettingsSelectWindow(gameClass,sessionType,config,editInfo,completionCall=completionCall).getRootWidget()
    bs.containerWidget(edit=prevWindow,transition='outLeft')
        

class PlaylistWindow(Window):

    def __init__(self,transition='inRight',playlistType='teams',selectPlaylist=None):
        global gMainWindow
        if playlistType == 'teams': gMainWindow = "Team Game Select"
        elif playlistType == 'freeForAll': gMainWindow = "Free-for-All Game Select"
        else: raise Exception('invalid playlist type: '+playlistType)

        self._playlistType = playlistType
        _setupPlaylistTypeVars(self,playlistType)

        self._R = R = bs.getResource('gameListWindow')

        self._width = 500
        self._height = 360 if gSmallUI else 440
        spacing = 40
        buttonWidth = 350
        self._rootWidget = bs.containerWidget(size=(self._width,self._height),transition=transition,
                                              scale = 2.0 if gSmallUI else 1.45 if gMedUI else 1.0,
                                              stackOffset=(0,-5) if gSmallUI else (0,0))

        backButton = b = bs.buttonWidget(parent=self._rootWidget,position=(25,self._height-60),size=(160,68),scale=0.77,textScale=1.3,label=bs.getResource('backText'),buttonType='back')
        self._playButton = b = bs.buttonWidget(parent=self._rootWidget,position=(self._width-145,self._height-60),size=(160,64),scale=0.77,textScale=1.3,label=bs.getResource('playText'))
        t = bs.textWidget(parent=self._rootWidget,position=(0,self._height-47),
                          size=(self._width,25),
                          text=self._windowTitleName,color=gHeadingColor,
                          maxWidth=180,
                          hAlign="center",vAlign="center")
        v = self._height - 100
        bs.textWidget(parent=self._rootWidget,position=(20,v+7),size=(100,30),
                      scale=0.7,
                      text=R.gameListText,
                      color=(0.6,0.6,0.66,1),
                      hAlign="left",
                      vAlign="bottom")

        self._scrollHeight = 230
        if gSmallUI: self._scrollHeight -= 65
        scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(40,v-self._scrollHeight),size=(self._width-80,self._scrollHeight+10))
        c = self._columnWidget = bs.columnWidget(parent=scrollWidget,printListExitInstructions=True)
        v -= (self._scrollHeight+35)
        h = 36
        hspacing = 15
        bColor = (0.6,0.53,0.63)
        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(87,31),
                            call=self._newPlaylist,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=R.newText)
        h += 84+hspacing
        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(95,31),
                            call=self._editPlaylist,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=R.editText)
        h += 90+hspacing
        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(105,31),
                            call=self._duplicatePlaylist,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=R.duplicateText)
        h += 100+hspacing
        b = bs.buttonWidget(parent=self._rootWidget,position=(h,v),size=(105,31),
                            call=self._deletePlaylist,
                            color=bColor,
                            textColor=(0.75,0.7,0.8),
                            textScale=0.7,
                            label=R.deleteText)
        h += 100+hspacing
        v -= 47

        try: self._doRandomizeVal = bs.config[self._configName+' Playlist Randomize']
        except Exception: self._doRandomizeVal = 0

        def _cbCallback(val):
            self._doRandomizeVal = val

        bs.checkBoxWidget(parent=self._rootWidget,position=(120,v+5),scale=0.9,size=(self._width-90,30),
                          text=R.shuffleGameOrderText,maxWidth=300,
                          textColor=(0.8,0.8,0.8),value=self._doRandomizeVal,valueCall=_cbCallback)

        # make sure config exists
        try: self._playlists = bs.config[self._configName+' Playlists']
        except Exception: bs.config[self._configName+' Playlists'] = self._playlists = {}

        self._selectedPlaylist = None
        self._selectedPlaylistIndex = None
        self._playlistWidgets = []

        self._refresh(selectPlaylist=selectPlaylist)

        bs.buttonWidget(edit=backButton,call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=backButton)

        bs.buttonWidget(edit=self._playButton,call=self._choosePlaylist)
        bs.containerWidget(edit=self._rootWidget,startButton=self._playButton,selectedChild=scrollWidget)

    def _back(self):
        if self._selectedPlaylist != None:
            bs.config[self._configName+' Playlist Selection'] = self._selectedPlaylist
            bs.config[self._configName+' Playlist Randomize'] = self._doRandomizeVal
            bs.writeConfig()

        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = StartGameWindow(transition='inLeft').getRootWidget()

    def _select(self,name,index):
        self._selectedPlaylist = name
        self._selectedPlaylistIndex = index

    def _choosePlaylist(self):
        import _bs

        # if we're an expired demo, bring up the purchase ui for convenience
        if _bs._getDemoTimeRemaining() == 0:
            showDemoExpiredUI()
            return

        if self._selectedPlaylist is None: return
        self._savePlaylistSelection()
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        def _run():
            #bs.fadeLocalScreen(True)
            bs.unlockAllInput()
            bs.hostSession(self._sessionTypeName)

        bs.fadeLocalScreen(False,time=250,endCall=_run)
        bs.lockAllInput()

    def _refresh(self,selectPlaylist=None):
        oldSelection = self._selectedPlaylist

        # if there was no prev selection, look in prefs
        if oldSelection is None:
            try: oldSelection = bs.config[self._configName+' Playlist Selection']
            except Exception: pass

        oldSelectionIndex = self._selectedPlaylistIndex

        # delete old
        while len(self._playlistWidgets) > 0: bs.deleteWidget(self._playlistWidgets.pop())

        items = self._playlists.items()
        items.sort(key=lambda x:x[0].lower())
        #items = [[self._defaultListName,None]] + items # default is always first
        items = [['__default__',None]] + items # default is always first
        index = 0
        for pName,p in items:
            w = bs.textWidget(parent=self._columnWidget,size=(self._width-40,24),
                              text=self._getPlaylistDisplayName(pName),
                              hAlign='left',vAlign='center',
                              color=(0.8,0.8,0.8,1.0),
                              alwaysHighlight=True,
                              selectCall=bs.Call(self._select,pName,index),
                              activateAction=bs.WeakCall(bs.activateWidget,self._playButton),
                              selectable=True)
            self._playlistWidgets.append(w)
            # select this one if the user requested itbs
            if selectPlaylist != None:
                if pName == selectPlaylist:
                    bs.columnWidget(edit=self._columnWidget,selectedChild=w,visibleChild=w)
            else:
                # select this one if it was previously selected
                if oldSelectionIndex != None: # go by index if there's one
                    if index == oldSelectionIndex:
                        bs.columnWidget(edit=self._columnWidget,selectedChild=w,visibleChild=w)
                else: # otherwise look by name
                    if pName == oldSelection:
                        bs.columnWidget(edit=self._columnWidget,selectedChild=w,visibleChild=w)

            index += 1

    def _savePlaylistSelection(self):
        # store the selected playlist in prefs..
        # this serves dual purposes of letting us re-select it next time
        # if we want and also lets us pass it to the game (since we reset the whole python environment thats not actually easy)
        bs.config[self._configName+' Playlist Selection'] = self._selectedPlaylist
        bs.config[self._configName+' Playlist Randomize'] = self._doRandomizeVal
        bs.writeConfig()

    def _newPlaylist(self):
        self._savePlaylistSelection() # in case they cancel so we can return to this state..
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        GameListEditSession(playlistType=self._playlistType)

    def _editPlaylist(self):
        if self._selectedPlaylist is None: return
        #if self._selectedPlaylist == self._defaultListName:
        if self._selectedPlaylist == '__default__':
            bs.playSound(bs.getSound('error'))
            bs.screenMessage(self._R.cantEditDefaultText)
            return
        self._savePlaylistSelection()
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        GameListEditSession(self._selectedPlaylist,playlistType=self._playlistType)

    def _doDeletePlaylist(self):
        try: del bs.config[self._configName+' Playlists'][self._selectedPlaylist]
        except Exception: pass
        bs.writeConfig()
        bs.playSound(bs.getSound('shieldDown'))

        # (we don't use len()-1 here because the default list adds one)
        if self._selectedPlaylistIndex > len(bs.config[self._configName+' Playlists']):
            self._selectedPlaylistIndex = len(bs.config[self._configName+' Playlists'])
        self._refresh()

    def _deletePlaylist(self):
        if self._selectedPlaylist is None: return
        #if self._selectedPlaylist == self._defaultListName:
        if self._selectedPlaylist == '__default__':
            bs.playSound(bs.getSound('error'))
            bs.screenMessage(self._R.cantDeleteDefaultText)
        else:
            ConfirmWindow(self._R.deleteConfirmText.replace('${LIST}',self._selectedPlaylist),self._doDeletePlaylist,450,150)
    
    def _getPlaylistDisplayName(self,playlist):
        if playlist == '__default__':
            return self._defaultListName
        else: return playlist

    def _duplicatePlaylist(self):

        if self._selectedPlaylist is None: return
        #if self._selectedPlaylist == self._defaultListName:
        if self._selectedPlaylist == '__default__':
            pl = self._getDefaultListCall()
        else:
            pl = self._playlists[self._selectedPlaylist]

        copyWord = bs.getResource('copyText')

        # find a valid dup name that doesn't exist
        testIndex = 1
        copyStr = '' if copyWord in self._selectedPlaylist else ' '+copyWord
        baseName = self._getPlaylistDisplayName(self._selectedPlaylist)

        # if its a copy, strip digits and spaces off the end
        if copyWord in baseName:
            while baseName[-1].isdigit() or baseName[-1] == ' ': baseName = baseName[:-1]
        while True:
            testName = baseName + (copyStr if testIndex < 2 else copyStr+' '+str(testIndex))
            if not testName in self._playlists:
                break
            testIndex += 1

        self._playlists[testName] = copy.deepcopy(pl)
        bs.writeConfig()
        bs.playSound(bs.getSound('gunCocking'))
        self._refresh(selectPlaylist=testName)


class StartGameWindow(Window):

    def __init__(self,transition='inRight'):
        width = 800
        height = 400
        spacing = 90
        buttonWidth = 400

        R = bs.getResource('playWindow')

        self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,scale=1.3 if gSmallUI else 1.2 if gMedUI else 0.9)
        backButton = b = bs.buttonWidget(parent=self._rootWidget,position=(55,height-92),size=(120,60),scale=1.1,
                                         textScale=1.2,
                                         label=bs.getResource('backText'),buttonType='back')

        t = bs.textWidget(parent=self._rootWidget,position=(0,height-74),
                          size=(width,25),
                          text=R.titleText,
                          scale=1.7,
                          color=gHeadingColor,
                          hAlign="center",vAlign="top")
        v = height - 60
        v -= 100
        c = (0.6,0.7,0.6,1.0)
        v -= 180
        hoffs = 0
        sc = 0.68

        self._lineupTex = bs.getTexture('playerLineup')
        angryComputerTransparentModel = bs.getModel('angryComputerTransparent')
        self._lineup1TransparentModel = bs.getModel('playerLineup1Transparent')
        self._lineup2TransparentModel = bs.getModel('playerLineup2Transparent')
        self._lineup3TransparentModel = bs.getModel('playerLineup3Transparent')
        self._lineup4TransparentModel = bs.getModel('playerLineup4Transparent')
        self._eyesModel = bs.getModel('plasticEyesTransparent')

        self._coopButton = b = bs.buttonWidget(parent=self._rootWidget,position=(hoffs,v),size=(sc*buttonWidth,sc*360),
                                         label="", buttonType='square', textScale=1.13, call=self._coop)

        self._drawDude(0,b,hoffs,v,sc,position=(140,30),color=(0.72,0.4,1.0))
        self._drawDude(1,b,hoffs,v,sc,position=(185,53),color=(0.71,0.5,1.0))
        self._drawDude(2,b,hoffs,v,sc,position=(220,27),color=(0.67,0.44,1.0))
        self._drawDude(3,b,hoffs,v,sc,position=(255,57),color=(0.7,0.3,1.0))
        bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(230),v+sc*(153)),size=(sc*115,sc*115),texture=self._lineupTex,modelTransparent=angryComputerTransparentModel)

        bs.textWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(-10),v+sc*(95)),size=(sc*buttonWidth,sc*50),
                      text=R.coopText,
                      maxWidth=sc*buttonWidth*0.8,
                      hAlign='center',vAlign='center',
                      color=(0.7,0.9,0.7,1.0),scale=sc*2.3)
        bs.textWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(-10),v+(sc*54)),size=(sc*buttonWidth,sc*30),
                      text=R.oneToFourPlayersText,hAlign='center',vAlign='center',
                      scale=0.78,
                      maxWidth=sc*buttonWidth*0.8,
                      color=c)

        hoffs += 260
        self._teamsButton = b = bs.buttonWidget(parent=self._rootWidget,position=(hoffs,v),size=(sc*buttonWidth,sc*360),
                                         label="", buttonType='square', textScale=1.13, call=self._teamTourney)

        xx = -14
        self._drawDude(2,b,hoffs,v,sc,position=(xx+148,30),color=(0.2,0.4,1.0))
        self._drawDude(3,b,hoffs,v,sc,position=(xx+181,53),color=(0.3,0.4,1.0))
        self._drawDude(1,b,hoffs,v,sc,position=(xx+216,33),color=(0.3,0.5,1.0))
        self._drawDude(0,b,hoffs,v,sc,position=(xx+245,57),color=(0.3,0.5,1.0))

        xx = 155
        self._drawDude(0,b,hoffs,v,sc,position=(xx+151,30),color=(1.0,0.5,0.4))
        self._drawDude(1,b,hoffs,v,sc,position=(xx+189,53),color=(1.0,0.58,0.58))
        self._drawDude(3,b,hoffs,v,sc,position=(xx+223,27),color=(1.0,0.5,0.5))
        self._drawDude(2,b,hoffs,v,sc,position=(xx+257,57),color=(1.0,0.5,0.5))

        bs.textWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(-10),v+sc*(95)),size=(sc*buttonWidth,sc*50),
                      text=R.teamsText,
                      maxWidth=sc*buttonWidth*0.8,
                      hAlign='center',vAlign='center',
                      color=(0.7,0.9,0.7,1.0),scale=sc*2.3)
        bs.textWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(-10),v+(sc*54)),size=(sc*buttonWidth,sc*30),
                      text=R.twoToEightPlayersText,hAlign='center',vAlign='center',
                      scale=0.78,
                      maxWidth=sc*buttonWidth*0.8,
                      color=c)

        hoffs += 260
        self._freeForAllButton = b = bs.buttonWidget(parent=self._rootWidget,position=(hoffs,v),size=(sc*buttonWidth,sc*360),
                                         label="", buttonType='square', textScale=1.13, call=self._freeForAll)

        xx = -5
        self._drawDude(0,b,hoffs,v,sc,position=(xx+140,30),color=(0.4,1.0,0.4))
        self._drawDude(3,b,hoffs,v,sc,position=(xx+185,53),color=(1.0,0.4,0.5))
        self._drawDude(1,b,hoffs,v,sc,position=(xx+220,27),color=(0.4,0.5,1.0))
        self._drawDude(2,b,hoffs,v,sc,position=(xx+255,57),color=(0.5,1.0,0.4))
        xx = 140
        self._drawDude(2,b,hoffs,v,sc,position=(xx+148,30),color=(1.0,0.9,0.4))
        self._drawDude(0,b,hoffs,v,sc,position=(xx+182,53),color=(0.7,1.0,0.5))
        self._drawDude(3,b,hoffs,v,sc,position=(xx+233,27),color=(0.7,0.5,0.9))
        self._drawDude(1,b,hoffs,v,sc,position=(xx+266,53),color=(1.0,0.6,0.4))
        bs.textWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(-10),v+sc*(95)),size=(sc*buttonWidth,sc*50),
                      text=R.freeForAllText,
                      maxWidth=sc*buttonWidth*0.8,
                      hAlign='center',vAlign='center',
                      color=(0.7,0.9,0.7,1.0),scale=sc*1.9)
        bs.textWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(-10),v+(sc*54)),size=(sc*buttonWidth,sc*30),
                      text=R.twoToEightPlayersText,hAlign='center',vAlign='center',
                      scale=0.78,
                      maxWidth=sc*buttonWidth*0.8,
                      color=c)

        bs.buttonWidget(edit=backButton,call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=backButton,selectedChild=self._coopButton)

        selected = None
        try:
            global gGameTypeSelection
            if gGameTypeSelection == 'Team Games': selected = self._teamsButton
            elif gGameTypeSelection == 'Free-for-All Games': selected = self._freeForAllButton
            elif gGameTypeSelection == 'Co-op Games': selected = self._coopButton
            else: raise Exception("unknown selection type: "+gGameTypeSelection)
        except Exception,e:
            print 'Err in StartGameWindow select restore:',e

        if selected != None:
            bs.containerWidget(edit=self._rootWidget,selectedChild=selected)

        # bs.textWidget(parent=self._rootWidget,position=(width*0.5,40),size=(200,50),editable=True,
        #               text='foobar',hAlign='right',vAlign='center',maxWidth=50)
        # bs.textWidget(parent=self._rootWidget,position=(width*0.5,40),size=(200,50),editable=True,
        #               text='foobarifficreallylong',hAlign='right',vAlign='center')
        # bs.textWidget(parent=self._rootWidget,position=(width*0.5,40),size=(200,50),editable=True,scale=0.7,
        #               text='foobar',hAlign='left',vAlign='bottom',color=(1,0,0,1))

    def _back(self):
        self._storeSelection()
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = MainMenuWindow(transition='inLeft').getRootWidget()

    def _coop(self):
        self._storeSelection('Co-op Games')
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = CooperativeSelectWindow().getRootWidget()

    def _teamTourney(self):
        self._storeSelection('Team Games')
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = PlaylistWindow(transition='inRight',playlistType='teams').getRootWidget()

    def _freeForAll(self):
        self._storeSelection('Free-for-All Games')
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = PlaylistWindow(transition='inRight',playlistType='freeForAll').getRootWidget()

    def _drawDude(self,i,b,hoffs,v,sc,position,color):
        hExtra = -100
        vExtra = 130
        eyeColor = (0.7*1.0+0.3*color[0],
                    0.7*1.0+0.3*color[1],
                    0.7*1.0+0.3*color[2])
        if i == 0:
            t = bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(hExtra+position[0]),v+sc*(vExtra+position[1])),size=(sc*60,sc*80),color=color,
                               texture=self._lineupTex,modelTransparent=self._lineup1TransparentModel)
            bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(hExtra+position[0]+12),v+sc*(vExtra+position[1]+53)),size=(sc*36,sc*18),texture=self._lineupTex,color=eyeColor,modelTransparent=self._eyesModel)
        elif i == 1:
            t = bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(hExtra+position[0]),v+sc*(vExtra+position[1])),size=(sc*45,sc*90),color=color,
                               texture=self._lineupTex,modelTransparent=self._lineup2TransparentModel)
            bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(hExtra+position[0]+5),v+sc*(vExtra+position[1]+67)),size=(sc*32,sc*16),texture=self._lineupTex,color=eyeColor,modelTransparent=self._eyesModel)
        elif i == 2:
            t = bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(hExtra+position[0]),v+sc*(vExtra+position[1])),size=(sc*45,sc*90),color=color,
                               texture=self._lineupTex,modelTransparent=self._lineup3TransparentModel)
            bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(hExtra+position[0]+5),v+sc*(vExtra+position[1]+59)),size=(sc*34,sc*17),texture=self._lineupTex,color=eyeColor,modelTransparent=self._eyesModel)
        elif i == 3:
            t = bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(hExtra+position[0]),v+sc*(vExtra+position[1])),size=(sc*48,sc*96),color=color,
                               texture=self._lineupTex,modelTransparent=self._lineup4TransparentModel)
            bs.imageWidget(parent=self._rootWidget,drawController=b,position=(hoffs+sc*(hExtra+position[0]+2),v+sc*(vExtra+position[1]+62)),size=(sc*38,sc*19),texture=self._lineupTex,color=eyeColor,modelTransparent=self._eyesModel)

    def _storeSelection(self,value=None):
        # None implies to grab highlighted choice
        if value is None:
            s = bs.getWidgetSelectedChild(self._rootWidget)
            if s == self._teamsButton: value = 'Team Games'
            elif s == self._coopButton: value = 'Co-op Games'
            elif s == self._freeForAllButton: value = 'Free-for-All Games'
            else: s = 'Team Games'
        if value != None:
            global gGameTypeSelection
            gGameTypeSelection = value

class LockedErrorWindow(Window):

    def __init__(self,name,depName):
        width = 600
        height = 200
        lockTex = bs.getTexture('lock')
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight',scale=1.5 if gSmallUI else 1.25 if gMedUI else 1.0)
        t = bs.textWidget(parent=self._rootWidget,position=(150,height*0.65),size=(0,0),
                          hAlign="left",
                          vAlign='center',
                          text=bs.getResource('levelIsLockedText').replace('${LEVEL}',name),
                          maxWidth=400,
                          color=(1,0.8,0.3,1),scale=1.1)
        t = bs.textWidget(parent=self._rootWidget,position=(160,height*0.46),size=(0,0),
                          hAlign="left",
                          vAlign='center',
                          text=bs.getResource('levelMustBeCompletedFirstText').replace('${LEVEL}',depName),
                          maxWidth=400,
                          color=gInfoTextColor,scale=0.8)
        bs.imageWidget(parent=self._rootWidget,position=(56,height*0.39),size=(80,80),texture=lockTex,opacity=1.0)
        b = bs.buttonWidget(parent=self._rootWidget,position=((width-140)/2,10),
                            size=(140,50),label=bs.getResource('okText'),call=self._ok)
        bs.containerWidget(edit=self._rootWidget,selectedChild=b,startButton=b)
        bs.playSound(bs.getSound('error'))
    def _ok(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')

class CooperativeSelectWindow(Window):

    def __init__(self,transition='inRight'):
        width = 1120
        height = 735
        global gMainWindow
        gMainWindow = "Coop Select"
        self._R = R = bs.getResource('coopSelectWindow')

        bs.gCreateGameType = None
        self._rootWidget = bs.containerWidget(size=(width,height),stackOffset=(0,-21),transition=transition, scale = 1.02 if gSmallUI else 0.9 if gMedUI else 0.8)
        backButton = b = bs.buttonWidget(parent=self._rootWidget,position=(75,height-87),size=(120,60),scale=1.2,label=bs.getResource('backText'),buttonType='back');
        try: self._lastSelected = bs.config['Selected Coop Game']
        except Exception: self._lastSelected = 'Default:Onslaught Training'

        self._doSetLastSelected = False # dont want initial construction affecting our last-selected
        v = height - 95
        campaignText = bs.textWidget(parent=self._rootWidget,position=(width/2-50,v+34),size=(100,30),
                      text=R.titleText,hAlign="center",color=gTitleColor,scale=1.7,
                      vAlign="bottom")
        campaignTextV = v

        self._starTex = bs.getTexture('star')
        self._lsbt = bs.getModel('levelSelectButtonTransparent')
        self._lsbo = bs.getModel('levelSelectButtonOpaque')
        self._aOutlineTex = bs.getTexture('achievementOutline')
        self._aOutlineModel = bs.getModel('achievementOutline')

        scrollWidth = width-130
        scrollHeight = height-160

        subContainerWidth = 800
        subContainerHeight = 714

        self._scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=(65,70),size=(scrollWidth,scrollHeight))
        self._subContainer = c = bs.containerWidget(parent=self._scrollWidget,scale=1.21,size=(subContainerWidth,subContainerHeight),singleDepth=True,background=False)

        # so we can still select root level widgets with controllers
        bs.containerWidget(edit=self._scrollWidget,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)
        bs.containerWidget(edit=self._subContainer,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)

        bs.containerWidget(edit=self._rootWidget,cancelButton=backButton,selectedChild=self._scrollWidget)

        wParent = self._subContainer

        hBase = 6

        h = hBase
        hSpacing = 150
        items = [['Onslaught\nTraining','Default:Onslaught Training','doomShroomPreview'],
                 ['Rookie\nOnslaught','Default:Rookie Onslaught','courtyardPreview'],
                 ['Rookie\nFootball','Default:Rookie Football','footballStadiumPreview'],
                 ['Pro\nOnslaught','Default:Pro Onslaught','doomShroomPreview'],
                 ['Pro\nFootball','Default:Pro Football','footballStadiumPreview']]
        row1 = []
        bs.textWidget(parent=wParent,position=(hBase+27,543),size=(0,30),
                                     text=R.campaignText,hAlign="left",color=gTitleColor,scale=1.1)
        for i in items:
            row1.append(self._gameButton(wParent,i[0],i[1],i[2],h,374,(i[1]==self._lastSelected)))
            h += 150

        items = [['Pro\nRunaround','Default:Pro Runaround','towerDPreview'],
                 ['Uber\nOnslaught','Default:Uber Onslaught','courtyardPreview'],
                 ['Uber\nFootball','Default:Uber Football','footballStadiumPreview'],
                 ['Uber\nRunaround','Default:Uber Runaround','towerDPreview'],
                 ['The\nLast Stand','Default:The Last Stand','rampagePreview']]
        row2 = []
        h = hBase
        for i in items:
            row2.append(self._gameButton(wParent,i[0],i[1],i[2],h,210,(i[1]==self._lastSelected)))
            h += 150

        items = [['Infinite\nOnslaught','Challenges:Onslaught','doomShroomPreview'],
                 ['Infinite\nRunaround','Challenges:Runaround','towerDPreview']]
        row3 = []
        #h = hBase-22+240
        h = hBase
        bs.textWidget(parent=wParent,position=(hBase+27,171),size=(0,30),
                                     text=R.customText,hAlign="left",color=gTitleColor,scale=1.1)
        for i in items:
            row3.append(self._gameButton(wParent,i[0],i[1],i[2],h,5,(i[1]==self._lastSelected)))
            h += 150

        for i in range(5):
            #bs.buttonWidget(edit=row1[i],downWidget=row2[i],upWidget=backButton)
            bs.buttonWidget(edit=row1[i],downWidget=row2[i])
            bs.buttonWidget(edit=row2[i],upWidget=row1[i])

        bs.buttonWidget(edit=row2[0],downWidget=row3[0])
        bs.buttonWidget(edit=row2[1],downWidget=row3[1])
        bs.buttonWidget(edit=row2[2],downWidget=row3[1])
        bs.buttonWidget(edit=row3[0],upWidget=row2[0])
        bs.buttonWidget(edit=row2[3],downWidget=row3[1])
        bs.buttonWidget(edit=row2[4],downWidget=row3[1])
        bs.buttonWidget(edit=row3[1],upWidget=row2[1])
        bs.buttonWidget(edit=backButton,call=self._back)
        self._doSetLastSelected = True

    def _gameButton(self,parent,name,game,image,x,y,select):

        x += (random.random()-0.5)*2
        y += (random.random()-0.5)*2
        sxRand = (random.random()-0.5)*14
        syRand = (random.random()-0.5)*14
        sx = (145+sxRand)
        sy = (170+syRand)
        x += (sxRand*-0.5)
        y += (syRand*-0.5)

        campaignName,levelName = game.split(':')
        campaign = bsCoopGame.getCampaign(campaignName)

        levels = campaign.getLevels()
        rating = campaign.getLevelRating(levelName)
        if rating is None: stars = 0
        elif rating >= 9.5: stars = 3
        elif rating >= 7.5: stars = 2
        else: stars = 1
        # if this campaign is sequential, make sure we've completed the one before this
        complete = True
        if campaign.isSequential():
            for i in range(len(levels)):
                if levels[i]['name'] == levelName:
                    break
                if not levels[i]['Complete']:
                    complete = False
                    break

        b = bs.buttonWidget(parent=parent,position=(x+25,y+4),size=(sx,sy),label='',call=bs.Call(self._run,game),
                            color=None if complete else (0.5,0.5,0.5),
                            showBufferTop=30.0,
                            buttonType='square',
                            selectCall=bs.Call(self._selChange,game))
        if select:
            bs.containerWidget(edit=parent,selectedChild=b,visibleChild=b)
        bs.imageWidget(parent=parent,drawController=b,position=(x+30+random.random()*3.0,y+80+random.random()*3.0),size=(120,30),
                       modelTransparent=self._lsbt,modelOpaque=self._lsbo,
                       texture=bs.getTexture(image),opacity=1.0 if complete else 0.3)

        try: translated = bs.getResource('translations.coopIconNames')[name]
        except Exception: translated = None
        if translated is None: translated = name

        bs.textWidget(parent=parent,drawController=b,position=(x+93,y+134),
                      size=(0,0),hAlign='center',text=translated,
                      vAlign='center',
                      maxWidth=sy*0.7,
                      scale=0.75,color=(0.8,1.0,0.8,1.0) if complete else (0.7,0.7,0.7,0.7))
        xs = x+48
        ys = y+54

        for i in range(stars):
            bs.imageWidget(parent=parent,drawController=b,position=(xs,ys),size=(31,31),
                           color=(2.2,1.2,0.3) if complete else (1,1,1),
                           texture=self._starTex,opacity=1.0 if complete else 0.3)
            xs += 30
        for i in range(3-stars):
            bs.imageWidget(parent=parent,drawController=b,position=(xs,ys),size=(31,31),
                           color=(0,0,0),
                           texture=self._starTex,opacity=0.2 if complete else 0.2)
            xs += 30

        achievements = bsAchievement.getAchievementsForCoopLevel(game)
        xa = x+50
        ya = y+28
        for a in achievements:
            aComplete = a.isComplete()
            bs.imageWidget(parent=parent,drawController=b,position=(xa,ya),size=(27,27),
                           color=tuple(a.getIconColor(aComplete)[:3]) if aComplete else (2,2,2),
                           texture=a.getIconTexture(aComplete),opacity=1.0 if (aComplete and complete) else 0.3)
            if aComplete:
                bs.imageWidget(parent=parent,drawController=b,position=(xa,ya),size=(27,27),
                               color=(2,1.4,0.4),
                               opacity=1.0 if complete else 0.2,
                               texture=self._aOutlineTex,
                               modelTransparent=self._aOutlineModel)
            xa += 30

        if not complete:
            bs.imageWidget(parent=parent,drawController=b,position=(x+66,y+58),size=(60,60),
                           texture=bs.getTexture('lock'))
        return b

    def _run(self,game):
        bs.config['Selected Coop Game'] = game
        bs.writeConfig()
        if bs._handleRunChallengeGame(game):
            bs.containerWidget(edit=self._rootWidget,transition='outLeft')

    def _back(self):
        # if something is selected, store it
        bs.config['Selected Coop Game'] = self._lastSelected
        bs.writeConfig()

        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = StartGameWindow(transition='inLeft').getRootWidget()

    def _selChange(self,game):
        if self._doSetLastSelected:
            self._lastSelected = game

def pause():
    global gCurrentGameSpeed
    # if we're in a game and 'playing' (ie: not at a score screen) then we set speed to 0..
    if bs.isPlaying():
        if bs.getGameSpeed() > 0.0:
            gCurrentGameSpeed = bs.getGameSpeed()
        bs.setGameSpeed(0)

def resume():
    global gCurrentGameSpeed
    # resume to prev saved value *unless* someone else has switched us off of zero..
    if bs.getGameSpeed() == 0.0:
        bs.setGameSpeed(gCurrentGameSpeed)

gMainMenuResumeCallbacks = []

def addMainMenuCloseCallback(call):

    # if there's no main menu up, just call immediately
    if uiGlobals['mainMenuWindow'] is None or not bs.widgetExists(uiGlobals['mainMenuWindow']):
        call()
    else:
        gMainMenuResumeCallbacks.append(call)

class MainMenuWindow(Window):

    def __init__(self,transition='inRight'):

        inGame = (bs.getGameName() != 'Main Menu')
        buttonHeight = 45
        buttonWidth = 200
        padding = 10

        tDelay = 0
        tDelayInc = 0
        tDelayPlay = 0

        self._R = R = bs.getResource('mainMenu')

        haveQuitButton = bs.getInterfaceType() == 'desktop' or bs.isOuyaVersion() or bs.isRunningOnFantasia()

        haveStoreButton = True if not inGame else False
        if bs.isTestBuild():
            haveStoreButton = False

        #haveStoreButton = False # not ready yet..

        positions = []
        pIndex = 0

        inputDevice = bs.getUIInputDevice()
        self._inputPlayer = bs.getInputDevicePlayer(inputDevice) if inputDevice != None else None

        if inGame:
            width = 250
            height = 250 if self._inputPlayer != None else 180
            height += 50*(len(customMenuUI)+len(customMenuUI2))
            self._rootWidget = bs.containerWidget(size=(width,height),transition=transition,scale=2.15 if gSmallUI else 1.6 if gMedUI else 1.0)
            h = 125
            v = height-80 if self._inputPlayer != None else height-60
            hOffset = 0
            dhOffset = 0
            vOffset = -50
            for i in range(6+len(customMenuUI)+len(customMenuUI2)):
                positions.append((h,v,1.0))
                v += vOffset; h += hOffset; hOffset += dhOffset

        # not in game
        else:
            global gDidMenuIntro
            if gDidMenuIntro == False:
                tDelay = 2000
                tDelayInc = 20
                tDelayPlay = 1700
                gDidMenuIntro = True

            width = 400
            height = 200

            gameService = bs.getGameService()
            #gameService = 'Game Center'
            if gameService == 'Game Center':
                gameServiceName = bs.getResource('gameCenterText')
                gameServiceIcon=bs.getTexture("gameCenterIcon")
            elif gameService == 'Game Circle':
                gameServiceName = bs.getResource('gameCircleText')
                gameServiceIcon=bs.getTexture("gameCircleIcon")


            bCount = 4 # play, help, credits, settings
            if gameService is not None: bCount += 1
            if haveQuitButton: bCount += 1
            if haveStoreButton: bCount += 1

            if gSmallUI:
                rootWidgetScale = 1.6
                playButtonWidth = buttonWidth*0.8
                playButtonHeight = buttonHeight*1.0
                smallButtonScale = 0.51 if bCount > 6 else 0.63
                buttonYOffs = -20
                buttonYOffs2 = -60
                buttonHeight *= 1.3
                buttonSpacing = 1.04
            elif gMedUI:
                rootWidgetScale = 1.3
                playButtonWidth = buttonWidth
                playButtonHeight = buttonHeight
                smallButtonScale = 0.6
                buttonYOffs = -55
                buttonYOffs2 = -75
                buttonHeight *= 1.25
                buttonSpacing = 1.1
            else:
                rootWidgetScale = 1.0
                playButtonWidth = buttonWidth
                playButtonHeight = buttonHeight
                smallButtonScale = 0.75
                buttonYOffs = -80
                buttonYOffs2 = -100
                buttonHeight *= 1.2
                buttonSpacing = 1.1

            spc = buttonWidth*smallButtonScale*buttonSpacing

            self._rootWidget = bs.containerWidget(size=(width,height),background=False,transition=transition,scale=rootWidgetScale)

            positions = [[width*0.5,buttonYOffs,1.7]]
            xOffs = width*0.5-(spc*(bCount-1)*0.5)+(spc*0.5)
            for i in range(bCount-1):
                positions.append([xOffs+spc*i,buttonYOffs+buttonYOffs2,smallButtonScale])

        if not inGame:

            if bs.isOuyaVersion():
                self._demoVersionUpdateTimer = bs.Timer(1000,bs.WeakCall(self._updateDemoButton),repeat=True,timeType='real')



            # play button
            h,v,scale = positions[pIndex]
            pIndex += 1
            startButton = b = bs.buttonWidget(parent=self._rootWidget, position=(h - playButtonWidth*0.5*scale,v),
                                              size=(playButtonWidth,playButtonHeight),
                                              scale=scale,
                                              label = bs.getResource('playText'),
                                              transitionDelay=tDelayPlay,
                                              call=self._playPress)

            bs.containerWidget(edit=self._rootWidget,startButton=startButton,selectedChild=startButton)

            # how-to-play button
            h,v,scale = positions[pIndex]
            pIndex += 1
            howToPlayButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h - buttonWidth*0.5*scale,v),
                                                  scale=scale,
                                                  size=(buttonWidth,buttonHeight),
                                                  label = R.howToPlayText,
                                                  transitionDelay=tDelay,
                                                  call=self._howToPlay)

            # credits button
            tDelay += tDelayInc
            h,v,scale = positions[pIndex]
            pIndex += 1
            creditsText = R.creditsText
            creditsButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h - buttonWidth*0.5*scale,v),
                                                size=(buttonWidth,buttonHeight),
                                                label = creditsText,
                                                scale=scale,
                                                transitionDelay=tDelay,
                                                call=self._credits)
            tDelay += tDelayInc

        # in-game
        else:
            pause()
            if bs.isPlaying() and transition == 'inRight': bs.playSound(bs.getSound('refWhistle'))

            # (player name if applicable)
            if self._inputPlayer != None:
                h,v,scale = positions[pIndex]
                v += 35
                b = bs.textWidget(parent=self._rootWidget, position=(h - buttonWidth/2,v), size=(buttonWidth,buttonHeight),
                                  color=(1,1,1,0.5),
                                  scale=0.7,
                                  hAlign='center',text = bs.getPlayerName(self._inputPlayer))

            h,v,scale = positions[pIndex]
            pIndex += 1

            b = bs.buttonWidget(parent=self._rootWidget, position=(h - buttonWidth/2,v), size=(buttonWidth,buttonHeight),scale=scale,
                                label = R.resumeText,
                                call=self._resume)
            bs.containerWidget(edit=self._rootWidget,cancelButton=b)

            # add any custom options defined by the current game
            for ui in customMenuUI:
                h,v,scale = positions[pIndex]
                pIndex += 1
                b = bs.buttonWidget(parent=self._rootWidget, position=(h - buttonWidth/2,v), size=(buttonWidth,buttonHeight),scale=scale,
                                    call=bs.Call(self._resumeAndCall,ui[1]),label = ui[0])

            # add a 'leave' button if the menu-owner has a player
            if self._inputPlayer != None:
                h,v,scale = positions[pIndex]
                pIndex += 1
                b = bs.buttonWidget(parent=self._rootWidget, position=(h - buttonWidth/2,v), size=(buttonWidth,buttonHeight),scale=scale,
                                    call=self._leave,label = R.leaveText)
                                    #call=self._leave,label = "Leave ("+bs.getPlayerName(self._inputPlayer)+")")

            for ui in customMenuUI2:
                h,v,scale = positions[pIndex]
                pIndex += 1
                b = bs.buttonWidget(parent=self._rootWidget, position=(h - buttonWidth/2,v), size=(buttonWidth,buttonHeight),scale=scale,
                                    call=bs.Call(self._resumeAndCall,ui[1]),label = ui[0])

        h,v,scale = positions[pIndex]
        pIndex += 1

        settingsButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h - buttonWidth*0.5*scale,v),size=(buttonWidth,buttonHeight),scale=scale,
                                             label = R.settingsText,transitionDelay=tDelay,call=self._settings)
        tDelay += tDelayInc

        if inGame:
            h,v,scale = positions[pIndex]
            pIndex += 1

            b = bs.buttonWidget(parent=self._rootWidget,position=(h - buttonWidth*0.5*scale,v),scale=scale,
                                size=(buttonWidth,buttonHeight),
                                label = R.endGameText,
                                call=self._confirmEndGame)


        if not inGame and gameService is not None:
            thisBWidth = buttonWidth
            h,v,scale = positions[pIndex]
            pIndex += 1
            gcButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h - thisBWidth*0.5*scale,v),size=(thisBWidth,buttonHeight),scale=scale,
                                           label=gameServiceName,
                                           call=bs.showOnlineScoreUI,
                                           icon=gameServiceIcon,
                                           transitionDelay=tDelay,
                                           enableSound=False,
                                           selectable=False if gameService == 'Game Center' else True)
            tDelay += tDelayInc

        if haveStoreButton:
            thisBWidth = buttonWidth
            h,v,scale = positions[pIndex]
            pIndex += 1


            storeButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h - thisBWidth*0.5*scale,v),
                                              size=(thisBWidth,buttonHeight),scale=scale,
                                              label=bs.getResource('storeText'),
                                              call=self._onStorePressed,
                                              transitionDelay=tDelay)
            iconSize = 55 if gSmallUI else 55 if gMedUI else 70
            iw = bs.imageWidget(parent=self._rootWidget,position=(h - iconSize*0.5,v+buttonHeight*scale-iconSize*0.23),transitionDelay=tDelay,size=(iconSize,iconSize),
                           texture=bs.getTexture('santaNub'))
            bs.imageWidget(edit=iw,drawController=storeButton)
            tDelay += tDelayInc
            # h,v,scale = positions[pIndex]
            # pIndex += 1

            # elif gameService == 'Game Circle':
            #     thisBWidth = buttonWidth
            #     leaderboardsButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h - thisBWidth/2,v),size=(thisBWidth,buttonHeight),scale=scale,
            #                                              label = bs.getResource('gameCircleText'),
            #                                              call=bs.Call(bs.showOnlineScoreUI,'general'),
            #                                              icon=bs.getTexture("gameCircleIcon"),
            #                                              transitionDelay=tDelay,enableSound=False,upWidget=startButton)
            #     tDelay += tDelayInc
            #     h,v,scale = positions[pIndex]
            #     pIndex += 1
                


        if not inGame and haveQuitButton:
            h,v,scale = positions[pIndex]
            pIndex += 1
            quitButton = b = bs.buttonWidget(parent=self._rootWidget,position=(h - buttonWidth*0.5*scale,v),size=(buttonWidth,buttonHeight),scale=scale,
                                label = R.quitText if 'Mac' in bs.getUserAgentString() else R.exitGameText,call = QuitWindow,transitionDelay=tDelay)
            #if bs.isOuyaVersion() or bs.isRunningOnFantasia():
            bs.containerWidget(edit=self._rootWidget,cancelButton=quitButton)
            tDelay += tDelayInc

        # on the main menu, set up buttons so we can navigate spatially instead of always going
        # to the next widget
        if not inGame:
            #bs.buttonWidget(edit=startButton,downWidget=settingsButton)
            bs.buttonWidget(edit=creditsButton,upWidget=startButton)
            bs.buttonWidget(edit=settingsButton,upWidget=startButton)
            if haveQuitButton:
                bs.buttonWidget(edit=quitButton,upWidget=startButton)
                #bs.buttonWidget(edit=startButton,rightWidget=quitButton)
            #else:
                #bs.buttonWidget(edit=startButton,rightWidget=settingsButton)

    def _updateDemoButton(self):
        import _bs
        timeRemaining = _bs._getDemoTimeRemaining()

        # if we have a time value, display it with the button
        if timeRemaining is not None and timeRemaining < 1000000:
            if not hasattr(self,'_demoWidgets'):
                self._demoWidgets = []
                bx = -130
                by = 25
                bSizeX = 240
                bSizeY = 200
                i = bs.imageWidget(parent=self._rootWidget,position=(bx-50,by+100-48),size=(100,100),
                                   transitionDelay=200,texture=bs.getTexture('cuteSpaz'))
                self._demoWidgets.append(i)
                b = self._demoButton = bs.buttonWidget(parent=self._rootWidget,position=(bx-bSizeX*0.5+5,by-bSizeY*0.5-8),size=(bSizeX,bSizeY),
                                                       transitionDelay=200,
                                                       call=bs.WeakCall(_bs._purchase,'game'),buttonType='square',label='',color=(0.3,0.5,0.3))
                bs.imageWidget(edit=i,drawController=b)
                self._demoWidgets.append(b)

                t = bs.textWidget(parent=self._rootWidget,text=bs.getResource('purchaseGameText'),hAlign='center',
                                  size=(0,0),position=(bx,by+40),color=(0,1,0,1),
                                  transitionDelay=200,
                                  scale=0.95,vAlign='center',drawController=b)
                self._demoWidgets.append(t)

                self._priceTextWidget = t = bs.textWidget(parent=self._rootWidget,text='',hAlign='center',
                                  size=(0,0),position=(bx,by+13),color=(0,1,0,1),
                                  transitionDelay=200,
                                  scale=0.7,vAlign='center',drawController=b)
                self._demoWidgets.append(t)

                t = bs.textWidget(parent=self._rootWidget,text=bs.getResource('remainingInTrialText'),hAlign='center',
                                  size=(0,0),position=(bx,by-60),color=(0.5,1,0.5,0.7),
                                  transitionDelay=200,
                                  scale=0.55,vAlign='center',drawController=b)
                self._demoWidgets.append(t)

                t = self._demoTimeText = bs.textWidget(parent=self._rootWidget,hAlign='center',
                                                       size=(0,0),position=(bx,by-27),color=(1,1,0.3,1),
                                                       transitionDelay=200,
                                                       scale=1.7,vAlign='center',drawController=b)
                self._demoWidgets.append(t)

            # update time-remaining and the cost (we may not be able to fetch cost immediately)
            bs.textWidget(edit=self._demoTimeText,text=bsUtils.getTimeString(timeRemaining*1000,centi=False))
            cost = _bs._getPrice('FullGameBombSquad')
            if cost is None: cost = ''
            if '?' in cost:
                print 'eww; wonky cost:',cost
                cost = cost.replace('?','')
            bs.textWidget(edit=self._priceTextWidget,text=cost)

        # got None or a huge number; means we're registered or have effectively infinite time
        else:
            if hasattr(self,'_demoWidgets'):
                for w in self._demoWidgets:
                    if bs.widgetExists(w):
                        bs.deleteWidget(w)
                del(self._demoWidgets)

    def _onStorePressed(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = StoreWindow().getRootWidget()

    def _confirmEndGame(self):
        ConfirmWindow(self._R.exitToMenuText,self._endGame)

    def _endGame(self):
        resume()
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        bs.getSession().end()

    def _leave(self):
        bs.removePlayer(self._inputPlayer)
        self._resume()

    def _credits(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = CreditsWindow().getRootWidget()

    def _howToPlay(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = HelpWindow(mainMenu=True).getRootWidget()

    def _settings(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        uiGlobals['mainMenuWindow'] = SettingsWindow().getRootWidget()

    def _resumeAndCall(self,c):
        self._resume()
        c()

    def _playPress(self):
        global gShouldAskToMakeProfile
        # if we have no profiles, lets ask if they wanna make one
        if gShouldAskToMakeProfile and ('Player Profiles' not in bs.config or len(bs.config['Player Profiles']) == 0):
            AskToMakeProfileWindow()
        else:
            bs.containerWidget(edit=self._rootWidget,transition='outLeft')
            uiGlobals['mainMenuWindow'] = StartGameWindow().getRootWidget()

    def _resume(self):
        if bs.getGameSpeed() == 0.0:
            bs.playSound(bs.getSound('refWhistle'))

        resume()
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = None

        # if there's callbacks waiting for this window to go away, call them
        for call in gMainMenuResumeCallbacks: call()
        del gMainMenuResumeCallbacks[:]


class ConfigErrorWindow(Window):

    def __init__(self):

        self._configFilePath = bs.getConfigFilePath()
        width = 800
        self._rootWidget = bs.containerWidget(size=(width,300),transition='inRight')
        padding = 20
        t = bs.textWidget(parent=self._rootWidget,position=(padding,220),size=(width-2*padding,100-2*padding),
                          hAlign="center",vAlign="top", scale=0.73,
                          text="Error reading BombSquad config file:\n\n\nCheck the console (press ~ twice) for details.\n\nWould you like to quit and try to fix it by hand\nor overwrite it with defaults?\n\n(high scores, player profiles, etc will be lost if you overwrite)")
        t2 = bs.textWidget(parent=self._rootWidget,position=(padding,198),size=(width-2*padding,100-2*padding),
                          hAlign="center",vAlign="top", scale=0.5,
                          text=self._configFilePath)
        quitButton = bs.buttonWidget(parent=self._rootWidget,position=(35,30),size=(240,54),label="Quit and Edit",call=self._quit)
        b = bs.buttonWidget(parent=self._rootWidget,position=(width-370,30),size=(330,54),label="Overwrite with Defaults",call=self._defaults)
        bs.containerWidget(edit=self._rootWidget,cancelButton=quitButton,selectedChild=quitButton)

    def _quit(self):
        bs.realTimer(1,self._editAndQuit)
        bs.lockAllInput()

    def _editAndQuit(self):
        bs.editTextFile(self._configFilePath)
        bs.realTimer(100,bs.quit)

    def _defaults(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        bs.playSound(bs.getSound('gunCocking'))
        bs.screenMessage("settings reset.",color=(1,1,0))
        # at this point settings are already set; lets just commit them to disk..
        bs.writeConfig(force=True)


def runPostLaunchStuff():

    # def doPrint():
    #     import bsLanguageGibberish
    #     print 'HELLO WORLD '+bsLanguageGibberish.values['achievementText']
    # bs.realTimer(1000,doPrint)

    # def foo():
    #     DebugWindow()
    # bs.realTimer(1000,foo)

    # first off, if there's a leftover log file, attempt to upload it to the server
    # and/or get rid of it..
    try:
        if os.path.exists(bs.getLogFilePath()):
            f = open(bs.getLogFilePath())
            info = eval(f.read())
            f.close()
            # try: doSend = bs.config['Submit Debug Info']
            # except Exception: doSend = True
            doSend = bs._shouldSubmitDebugInfo()

            if doSend:
                def response(data):
                    # non-None response means we were succesful; lets kill it.
                    if data is not None: os.remove(bs.getLogFilePath())
                bsUtils.serverPut('bsLog',info,response)
            else:
                # if they dont want logs uploaded just kill it
                os.remove(bs.getLogFilePath())
    except Exception,e:
        print 'Error handling leftover log file:',e


    # notify the user if we're using custom system scripts:
    if bs.getSystemScriptsDirectory() != 'data/scripts':
        bs.screenMessage("Using custom system scripts...",color=(0,1,0))

    # (only do this stuff if our config file is healthy so we dont
    # overwrite a broken one or whatnot and wipe out data)
    if not bs.gConfigFileIsHealthy:
    # if True:
        ua = bs.getUserAgentString()
        if 'Mac' in ua or 'linux' in ua or 'windows' in ua:
            ConfigErrorWindow()
            return
        else:
            # for now on other systems we just overwrite the bum config
            # at this point settings are already set; lets just commit them to disk..
            bs.writeConfig(force=True)

    # if we're using a non-default playlist lets go ahead and get
    # our itunes controller thread going since it may hitch
    # (better while we're faded out than later)
    try:
        if bs.config['Soundtrack'] not in['__default__','Default Soundtrack']:
            bsUtils.getITunesThread()
    except Exception: pass

    # if this is our first launch ever, show the help screen
    try: launchCount = bs.config['launchCount']
    except Exception: launchCount = 0
    launchCount += 1

    # debugging - make note if we're using the local test server so we dont accidentally
    # leave it on in a release
    if 'localhost' in bs.getServerAddress():
        bs.realTimer(2000,bs.WeakCall(bs.screenMessage,"Note: using local server",(1,1,0)))

    gamePadConfigQueryDelay = 1000
    
    if launchCount == 1:
        global gShouldAskToMakeProfile
        gShouldAskToMakeProfile = True
        def _showHelp():
            # only show if our launch count is *still* 1..
            # we may have synced from the cloud in which case our 'real' launch
            # count may be bigger
            if bs.config['launchCount'] == 1:
                bs.playSound(bs.getSound('swish'))
                HelpWindow(mainMenu=False)
            bs.unlockAllInput()
                
        bs.lockAllInput()
        bs.realTimer(2500,_showHelp)

        # make sure we dont ask to configure gamepads before this shows up..
        gamePadConfigQueryDelay = 3000

    # inform folks on 10.8.0 or 10.8.1 that they get game-center if the update to 10.8.2
    ua = bs.getUserAgentString()
    if 'Mac OS X 10.8.1' in ua or 'Mac OS X 10.8.0' in ua:
        bs.screenMessage("Update to OSX 10.8.2 or higher to use Game Center.",color=(1,1,0))
        bs.screenMessage("You are running "+('10.8.1' if '10.8.1' in ua else '10.8.0'),color=(1,1,0))

    # on the mac version, beg for ratings after they've been playing a few times
    if launchCount == 5 and 'Mac' in bs.getUserAgentString():

        def askForRating():

            global gCanAskToConfigGamePads
            gCanAskToConfigGamePads = False
            bs.playSound(bs.getSound('swish'))
            bs.unlockAllInput()
            width = 700
            height = 400
            spacing = 40
            d = bs.containerWidget(size=(width,height),transition='inRight')
            v = height - 50
            v -= spacing
            v -= 140
            w = bs.imageWidget(parent=d,position=(width/2-100,v+10),size=(200,200),
                               texture=bs.getTexture("cuteSpaz"))
            t = bs.textWidget(parent=d,position=(15,v),size=(width-30,30),
                              color=gInfoTextColor,
                              text=bs.getResource('pleaseRateText'),
                              maxWidth=width*0.95,
                              scale=0.85,
                              hAlign="center",vAlign="top")

            def doRating():
                # doing this in a timer seems to keep bombsquad from maintaining focus
                bs.openURL('macappstore://itunes.apple.com/app/id416482767?ls=1&mt=12')
                # hmm this one is cleaner but goes through safari which is ugly
                #bs.openURL('http://appstore.com/mac/bombsquad')
                bs.containerWidget(edit=d,transition='outLeft')
                global gCanAskToConfigGamePads
                gCanAskToConfigGamePads = True

            b = bs.buttonWidget(parent=d,position=(60,30),size=(200,40),
                                label=bs.getResource('wellSureText'),
                                call=doRating)

            def close():
                bs.containerWidget(edit=d,transition='outLeft')
                global gCanAskToConfigGamePads
                gCanAskToConfigGamePads = True
                
            b = bs.buttonWidget(parent=d,position=(width - 270,30),size=(200,40),
                                label=bs.getResource('noThanksText'),
                                call=close)
            bs.containerWidget(edit=d,cancelButton=b,selectedChild=b)
            return d

        bs.lockAllInput()
        bs.realTimer(2500,askForRating)

        # make sure we dont ask to configure gamepads before this shows up..
        gamePadConfigQueryDelay = 3000

    # ask if we can submit debug info
    try: canDebug = bs.config['Submit Debug Info']
    except Exception: canDebug = None

    if launchCount > 5 and canDebug is None:

        # ask about submitting debug info on the desktop since
        # the user can tweak stuff there
        if bs.getInterfaceType() == 'desktop':
            def askIfCanDebug():

                bs.playSound(bs.getSound('swish'))
                bs.unlockAllInput()
                width = 700
                height = 400
                spacing = 40
                d = bs.containerWidget(size=(width,height),transition='inRight')
                v = height - 50
                v -= spacing
                v -= 140
                w = bs.imageWidget(parent=d,position=(width/2-100,v+10),size=(200,200),
                                   texture=bs.getTexture("cuteSpaz"))
                t = bs.textWidget(parent=d,position=(15,v),size=(width-30,30),
                                  color=gInfoTextColor,
                                  scale=0.8,
                                  text=bs.getResource('canWeDebugText'),maxWidth=650,
                                  hAlign="center",vAlign="top")

                def canDebug():
                    bs.config['Submit Debug Info'] = True
                    bs.writeConfig()
                    bs.containerWidget(edit=d,transition='outLeft')

                b1 = bs.buttonWidget(parent=d,position=(60,30),size=(200,45),
                                    label=bs.getResource('yesAllowText'),
                                    call=canDebug)

                def canNotDebug():
                    bs.config['Submit Debug Info'] = False
                    bs.writeConfig()
                    bs.containerWidget(edit=d,transition='outLeft')

                b2 = bs.buttonWidget(parent=d,position=(width - 270,30),size=(200,45),
                                    label=bs.getResource('noThanksText'),
                                    call=canNotDebug)

                bs.containerWidget(edit=d,cancelButton=b2,selectedChild=b1)

                return d
            bs.lockAllInput()
            bs.realTimer(2500,askIfCanDebug)

        # on non-desktop, just set yes for now.
        else:
            bs.config['Submit Debug Info'] = True
            bs.writeConfig()

    # lets make a list of unconfigured joysticks so if we're sitting idle
    # we can ask the user to config them
    global gUntestedGamePads
    gUntestedGamePads += bs.getConfigurableGamePads()

    # if there's any gamepads in there, set up our timer to go through them..
    if len(gUntestedGamePads) > 0:
        bs.realTimer(gamePadConfigQueryDelay,_setupGamePadConfigTimer)

    bs.config['launchCount'] = launchCount

    # write out our config in a few seconds.. this delay gives a little time
    # for cloud settings to come in so we're less likely to inadvertantly overwrite cloud
    # values with stale old values of ours and stuff..
    bs.realTimer(5000,bs.writeConfig)

    # lets go ahead and scan our games to avoid hitches later
    bs.getGameTypes()
    
def checkGamePadConfigs():
    " can be called during idle time to ask the user to configure any unconfigured joysticks "

    global gCanAskToConfigGamePads

    # go through the list until we find one we ask about or we have to stop asking
    while len(gUntestedGamePads) > 0 and gCanAskToConfigGamePads:
        gamepad = gUntestedGamePads.pop()

        # ok lets just ask for a button assignment for this joystick.. if we dont get anything
        # back lets ask the user if they want to configure it..
        try:
            getControllerValue(gamepad,"buttonJump",exceptionOnUnknown=True)
            known = True
        except Exception:
            known = False

        # in some cases we dont bother asking (if it looks like its a keyboard or a mouse that just happens
        # to be showing up as a joystick)
        ask = True
        try:
            name = bs.getInputDeviceName(gamepad)
            if 'mouse' in name.lower() or 'keyboard' in name.lower(): ask = False
        except Exception: ask = False

        if not known and ask:
            def askToConfig():
                try: name = bs.getInputDeviceName(gamepad)
                except Exception,e: name = None

                # only config if we can get its name
                if name is not None:
                    bs.playSound(bs.getSound('swish'))
                    # dont ask about more while we're configing this one..
                    global gCanAskToConfigGamePads
                    gCanAskToConfigGamePads = False

                    width = 600
                    height = 200
                    spacing = 40
                    d = bs.containerWidget(scale=1.8 if gSmallUI else 1.4 if gMedUI else 1.0,
                                           size=(width,height),transition='inRight')

                    v = height - 60
                    t = bs.textWidget(parent=d,position=(15,v),size=(width-30,30),
                                      color=gInfoTextColor,
                                      text=bs.getResource('unconfiguredControllerDetectedText'),
                                      scale=0.8,
                                      hAlign="center",vAlign="top")
                    v -= 30
                    t = bs.textWidget(parent=d,position=(15,v),size=(width-30,30),
                                      color=(1,1,1,1.0),
                                      text=name,
                                      hAlign="center",vAlign="top")
                    v -= 30
                    t = bs.textWidget(parent=d,position=(15,v),size=(width-30,30),
                                      color=gInfoTextColor,
                                      text=bs.getResource('configureItNowText'),
                                      scale=0.8,
                                      hAlign="center",vAlign="top")

                    def doConfigure(gamepad):
                        bs.containerWidget(edit=d,transition='outLeft')
                        GamePadConfigWindow(gamepad,isMainMenu=False)

                    b = bs.buttonWidget(parent=d,position=(20,20),size=(200,60),
                                        label=bs.getResource('configureText'),
                                        call=bs.WeakCall(doConfigure,gamepad))

                    def close():
                        bs.containerWidget(edit=d,transition='outLeft')

                        # if they dont wanna configure now, lets cancel all..
                        global gUntestedGamePads
                        gUntestedGamePads = []

                        # ok can go back to asking now..
                        global gCanAskToConfigGamePads
                        gCanAskToConfigGamePads = True

                    b = bs.buttonWidget(parent=d,position=(width - 230,20),size=(200,60),
                                        label=bs.getResource('notNowText'),
                                        call=close)

            def cb(data):
                global gCanAskToConfigGamePads
                gCanAskToConfigGamePads = True

                try: name = bs.getInputDeviceName(gamepad)
                except Exception,e: name = None

                haveConfig = False
                if data and 'config' in data and data['config'] is not None and name is not None:

                    # if there's currently no default config entry, write this one
                    configLoc = getConfigLocationForInputDevice(gamepad,default=True)
                    configDict = configLoc[0][configLoc[1]]
                    if len(configDict) == 0:
                        configLoc[0][configLoc[1]] = copy.deepcopy(data['config'])
                        haveConfig = True
                        #print '!!!GOT CONFIG FOR',name,':',data['config']
                        bs.applySettings()
                        bs.writeConfig()
                # finally if we still have no config, ask the user..
                if not haveConfig:
                    askToConfig()

            # if we can get its name, ask the server about it
            # we include our input-map-hash so that differently mapped systems
            # will return different sets of results
            try: name = bs.getInputDeviceName(gamepad)
            except Exception,e: name = None
            if name is not None:
                gCanAskToConfigGamePads = False # lock while waiting for response
                bsUtils.serverGet('controllerConfig',{'ua':bs.getUserAgentString(),'name':name,'inputMapHash':getInputMapHash(gamepad)},callback=cb)


    # ok, lastly, if there's no more gamepads to ask about, lets kill our timer
    if len(gUntestedGamePads) == 0:
        global gGamePadConfigQueryTimer
        gGamePadConfigQueryTimer = None

def _setupGamePadConfigTimer():
    # make/rename a timer to periodically ask about gamepads until none are left
    global gGamePadConfigQueryTimer
    gGamePadConfigQueryTimer = bs.Timer(1000,checkGamePadConfigs,repeat=True,timeType='real')


def gamePadConnected(index):
    # put this on the list so we can offer to config it if its not..
    gUntestedGamePads.append(index)
    _setupGamePadConfigTimer()


class AskToMakeProfileWindow(Window):

    def __init__(self):

        global gShouldAskToMakeProfile

        # only ask once
        gShouldAskToMakeProfile = False

        width = 650
        height = 265
        spacing = 40
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight',scale=1.4 if gSmallUI else 1.2 if gMedUI else 1.0)
        v = height - 43
        v -= spacing
        t = bs.textWidget(parent=self._rootWidget,position=(15,v),size=(width-30,30),
                          color=gInfoTextColor,
                          scale=1.2,
                          text=bs.getResource('createAPlayerProfileText'),
                          maxWidth=width*0.9,
                          hAlign="center",vAlign="top")

        t = bs.textWidget(parent=self._rootWidget,position=(15,v-35),size=(width-30,30),
                          color=(1,1,1,0.7),
                          scale=0.8,
                          text=bs.getResource('profileInfoText'),
                          maxWidth=width*0.9,
                          hAlign="center",vAlign="top")

        t = bs.textWidget(parent=self._rootWidget,position=(15,v-84),size=(width-30,30),
                          color=(1,1,1,0.6),
                          scale=0.6,
                          text=(bs.getResource('editProfilesAnyTimeText')),
                          maxWidth = width*0.9,
                          hAlign="center",vAlign="top")

        editButton = b = bs.buttonWidget(parent=self._rootWidget,position=(50,30),size=(200,50),
                                         label=bs.getResource('createText'),
                                         call=self._configProfiles)
        def close():
            bs.containerWidget(edit=self._rootWidget,transition='outLeft')

        b = bs.buttonWidget(parent=self._rootWidget,position=(width - 250,30),size=(200,50),
                            label=bs.getResource('notNowText'),
                            call=close)

        bs.containerWidget(edit=self._rootWidget,cancelButton=b,selectedChild=editButton)

    def _configProfiles(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')
        ConfigPlayerProfilesWindow(inMainMenu=False)

def runMainMenu():
    uiGlobals['mainMenuWindow'] = MainMenuWindow().getRootWidget()

class LocalConnectFailMessageWindow(Window):
    def __init__(self):
        width = 670
        height = 270
        bs.playSound(bs.getSound('error'))
        self._rootWidget = bs.containerWidget(size=(width,height),transition='inRight')
        t = bs.textWidget(parent=self._rootWidget,position=(15,height-55),size=(width-30,30),
                          text=("Fatal Error:\n\n"
                                +"BombSquad is unable to connect to itself on socket port "+str(bs.getGamePort())+".\n"
                                +"It may be getting blocked by security software such as 'Hands Off!'.\n\n"
                                +"Change your your settings to allow BombSquad to\n"
                                +"host connections on port "+str(bs.getGamePort())+" and and try again.\n\n"
                                +"If this does not work, email support@froemling.net for help."),
                          color=(1,0.6,0.2,1),
                          scale=0.8,
                          hAlign="center",vAlign="top")
        b = bs.buttonWidget(parent=self._rootWidget,position=((width-160)/2,10),size=(160,50),label="Quit",call=bs.quit)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

def handleDeviceMenuPress(device):
    inMainMenu = (uiGlobals['mainMenuWindow'] != None and bs.widgetExists(uiGlobals['mainMenuWindow']))
    if not inMainMenu:
        bs.setUIInputDevice(device)
        bs.playSound(bs.getSound('swish'))
        uiGlobals['mainMenuWindow'] = MainMenuWindow().getRootWidget()

class DemoExpiredWindow(Window):
    def __init__(self):
        import _bs
        width = 600
        height = 460
        spacing = 30

        bs.playSound(bs.getSound('error'))
        self._rootWidget = bs.containerWidget(size=(width,height),scale=1.3,transition='inRight')
        b = bs.buttonWidget(parent=self._rootWidget,position=(55,height-70),size=(180,60),label=bs.getResource('cancelText'),scale=0.85,call=self._dismiss)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        # t = bs.textWidget(parent=self._rootWidget,position=(width*0.5,height-50),size=(0,0),text="Keep Playing?",color=gTitleColor,
        #                   hAlign="center",vAlign="center")

        v = height - 50
        v -= spacing
        v -= 200
        w = bs.imageWidget(parent=self._rootWidget,position=(width/2-100,v+10),size=(200,200),
                           texture=bs.getTexture("cuteSpaz"))
        t = bs.textWidget(parent=self._rootWidget,position=(15,v-30),size=(width-30,30),
                          color=gInfoTextColor,
                          scale=0.82,
                          text=bs.getResource('trialPeriodEndedText'),
                          maxWidth=width*0.9,
                          hAlign="center",vAlign="center")
        cost = _bs._getPrice('FullGameBombSquad')
        if cost is None: cost = '<unavailable>'
        if '?' in cost:
            print 'eww; wonky cost:',cost
            cost = cost.replace('?','')
        b = bs.buttonWidget(parent=self._rootWidget,position=(width*0.5-180,v-120),size=(360,60),label=bs.getResource('purchaseForText').replace('${PRICE}',cost),scale=1.0,call=self._purchase)

    def _purchase(self):
        import _bs
        self._dismiss()
        _bs._purchase('game')

    def _dismiss(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')


def showDemoExpiredUI():
    DemoExpiredWindow()

def handleTelnetAccessRequest():

    class _TelnetAccessWindow(Window):

        def __init__(self):
            width = 400
            height = 100
            text = bs.getResource('telnetAccessText')

            self._rootWidget = bs.containerWidget(size=(width,height+40),transition='inRight',scale=1.7 if gSmallUI else 1.3 if gMedUI else 1.0)
            padding = 20
            t = bs.textWidget(parent=self._rootWidget,position=(padding,padding+33),size=(width-2*padding,height-2*padding),
                              hAlign="center",vAlign="top",text=text)
            b = bs.buttonWidget(parent=self._rootWidget,position=(20,20),size=(140,50),label=bs.getResource('denyText'),call=self._cancel)
            bs.containerWidget(edit=self._rootWidget,cancelButton=b)
            bs.containerWidget(edit=self._rootWidget,selectedChild=b)

            b = bs.buttonWidget(parent=self._rootWidget,position=(width-155,20),size=(140,50),label=bs.getResource('allowText'),call=self._ok)

        def _cancel(self):
            bs.containerWidget(edit=self._rootWidget,transition='outRight')
            bs.setTelnetAccessEnabled(False)

        def _ok(self):
            bs.containerWidget(edit=self._rootWidget,transition='outLeft')
            bs.setTelnetAccessEnabled(True)
            bs.screenMessage(bs.getResource('telnetAccessGrantedText'))

    _TelnetAccessWindow()

class ShowURLWindow(Window):

    def __init__(self,address):
        self._width = 800
        self._height = 200
        self._rootWidget = bs.containerWidget(size=(self._width,self._height+40),transition='inRight',scale=1.25 if gSmallUI else 1.25 if gMedUI else 1.25)
        t = bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,self._height-10),size=(0,0),
                          color=gTitleColor,hAlign="center",vAlign="center",text=bs.getResource('directBrowserToURLText'),maxWidth=self._width*0.95)
        t = bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,self._height*0.5+29),size=(0,0),scale=1.3,
                          color=gInfoTextColor,hAlign="center",vAlign="center",text=address,maxWidth=self._width*0.95)
        buttonWidth = 200
        b = bs.buttonWidget(parent=self._rootWidget,position=(self._width*0.5-buttonWidth*0.5,20),size=(buttonWidth,65),label=bs.getResource('doneText'),call=self._done)
        bs.containerWidget(edit=self._rootWidget,selectedChild=b,startButton=b)

    def _done(self):
        bs.containerWidget(edit=self._rootWidget,transition='outLeft')

class StoreWindow(Window):


    def __init__(self):

        self._width = 940
        self._height = 595 if gSmallUI else 675 if gMedUI else 800
        self._currentTabIndex = -1
        self._request = None
        self._R = bs.getResource('store')
        self._lastBuyTime = 0

        self._rootWidget = bs.containerWidget(size=(self._width,self._height+40),transition='inRight',
                                              scale=1.22 if gSmallUI else 1.05 if gMedUI else 0.8,
                                              stackOffset=(0,-47) if gSmallUI else (0,-40) if gMedUI else (0,0))

        b = bs.buttonWidget(parent=self._rootWidget,position=(80,self._height-24),size=(120,60),scale=1.1,
                            label=bs.getResource('backText'),buttonType='back', call=self._back)
        bs.containerWidget(edit=self._rootWidget,cancelButton=b)

        t = bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,self._height),size=(0,0),
                          color=gTitleColor,scale=1.5,hAlign="center",vAlign="center",text=bs.getResource('storeText'),maxWidth=400)

        self._tabButtons = []

        self._tabButtonWidth = (self._width-130)*0.27

        b = bs.buttonWidget(parent=self._rootWidget,position=(self._width*0.5-self._tabButtonWidth*1.52,self._height - 80),
                            buttonType='tab',size=(self._tabButtonWidth,50),label=self._R.charactersText,call=bs.Call(self._setTab,0))
        self._tabButtons.append(b)
        bs.containerWidget(edit=self._rootWidget,selectedChild=b)

        b = bs.buttonWidget(parent=self._rootWidget,position=(self._width*0.5-self._tabButtonWidth*0.5,self._height - 80),
                            buttonType='tab',size=(self._tabButtonWidth,50),label=self._R.mapsText,call=bs.Call(self._setTab,1))
        self._tabButtons.append(b)

        b = bs.buttonWidget(parent=self._rootWidget,position=(self._width*0.5+self._tabButtonWidth*0.52,self._height - 80),
                            buttonType='tab',size=(self._tabButtonWidth,50),label=self._R.miniGamesText,call=bs.Call(self._setTab,2))
        self._tabButtons.append(b)

        self._scrollWidth = self._width-130
        self._scrollHeight = self._height-160

        self._scrollWidget = None
        self._statusTextWidget = None
        self._setTab(0)


    def _setTab(self,index):

        if index == self._currentTabIndex: return
        self._currentTabIndex = index

        # update buttons
        for i,b in enumerate(self._tabButtons):
            if index == i: bs.buttonWidget(edit=b,color=(0.5,0.4,0.93),textColor=(0.8,0.7,0.9)) # lit
            else: bs.buttonWidget(edit=b,color=(0.52,0.48,0.63),textColor=(0.55,0.5,0.6)) # unlit

        # recreate scroll widget
        if self._scrollWidget is not None and bs.widgetExists(self._scrollWidget):
            bs.deleteWidget(self._scrollWidget)

        self._scrollWidget = bs.scrollWidget(parent=self._rootWidget,position=((self._width-self._scrollWidth)*0.5,self._height-self._scrollHeight-79),size=(self._scrollWidth,self._scrollHeight))

        # stop updating anything that was there
        self._buttonInfos = {}
        self._updateButtonsTimer = None

        # hitting down from a tab should go to the content
        for tb in self._tabButtons:
            bs.widget(edit=tb,downWidget=self._scrollWidget)

        # so we can still select root level widgets with controllers
        bs.containerWidget(edit=self._scrollWidget,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)

        # show status over top..
        if self._statusTextWidget is not None and bs.widgetExists(self._statusTextWidget):
            bs.deleteWidget(self._statusTextWidget)
        self._statusTextWidget = bs.textWidget(parent=self._rootWidget,
                                               position=(self._width*0.5-bs.getStringWidth(self._R.loadingText)*0.5,self._height*0.5),size=(0,0),
                                               color=(1,0.7,1,0.5),hAlign="left",vAlign="center",
                                               text=self._R.loadingText,maxWidth=self._scrollWidth*0.9)
        self._statusTextWidgetUpdateTimer = bs.Timer(500,bs.WeakCall(self._updateStatusText),repeat=True,timeType='real')
        self._statusTextDots = 1


        class _Request(object):
            def __init__(self,window):
                self._window = weakref.ref(window)
                data = {'foo':'bar'}
                bs.realTimer(random.randrange(250),bs.WeakCall(self._onResponse,data))
            def _onResponse(self,data):
                window = self._window()
                if window is not None and (window._request is self):
                    window._request = None
                    window._onResponse(data)

        # kick off a server request
        self._request = _Request(self)
        
    def _updateStatusText(self):
        bs.textWidget(edit=self._statusTextWidget,
                      text=self._R.loadingText+'.'*self._statusTextDots)
        
        self._statusTextDots += 1
        if self._statusTextDots > 3:
            self._statusTextDots = 0

    def _buy(self,item):
        import _bs
        # prevent pressing buy within a few seconds of the last press
        # (gives the buttons time to disable themselves and whatnot)
        t = bs.getRealTime()
        if t - self._lastBuyTime < 1000:
            bs.playSound(bs.getSound('error'))
        else:
            self._lastBuyTime = t
            _bs._purchase(item)

    def _printAlreadyOwn(self,charName):
        bs.screenMessage(self._R.alreadyOwnText.replace('${NAME}',charName))
        bs.playSound(bs.getSound('error'))

    def _updateButtons(self):
        #bs.screenMessage("upd but "+str(bs.getRealTime()))
        import _bs
        for bType,bInfo in self._buttonInfos.items():
            if _bs._getPurchased(bType):
                color = (0.6,0.6,0.6)
                frameColor = (0.42,0.42,0.42)
                call = bs.WeakCall(self._printAlreadyOwn,bInfo['name'])
                priceText = ''
            else:
                color = (0.5,0.7,0.2)
                frameColor = (0.35,0.5,0.15)
                call = bInfo['call']
                priceText = bInfo['price']
            bs.textWidget(edit=bInfo['priceWidget'],text=priceText)
            bs.buttonWidget(edit=bInfo['button'],color=color,call=call)
            bs.imageWidget(edit=bInfo['frameWidget'],color=frameColor)

    def _onResponse(self,data):

        # clear status text..
        if self._statusTextWidget is not None and bs.widgetExists(self._statusTextWidget):
            bs.deleteWidget(self._statusTextWidget)
            self._statusTextWidgetUpdateTimer = None
        
        if data is None:
            self._statusTextWidget = bs.textWidget(parent=self._rootWidget,position=(self._width*0.5,self._height*0.5),size=(0,0),
                                                   scale=1.3,
                                                   transitionDelay=100,
                                                   color=(1,0.3,0.3,1.0),hAlign="center",vAlign="center",
                                                   text=self._R.loadErrorText,maxWidth=self._scrollWidth*0.9)
        else:
            class _Store(object):

                def __init__(self,storeWindow,data,width):

                    self._storeWindow = storeWindow

                    self._width = width
                    self._rows = []

                    self.addButtonRow(buttonSize=(355*0.6,430*0.6),title=bs.getResource('store.holidaySpecialText'),buttons=['frosty','santa'])
                    # self.addButtonRow(buttonSize=(355*0.6,430*0.6),title=bs.getResource('store.holidaySpecialText'),buttons=['frosty','santa','santa'])
                    # self.addButtonRow(buttonSize=(140,120),title='ButtonRow2',buttons=['foo2','bar2','eep2','gpoing2','floof2','flofff'])
                    # self.addButtonRow(buttonSize=(100,100),title='ButtonRow3',buttons=['foo3'])
                    # self.addButtonRow(buttonSize=(100,100),title='ButtonRow4',buttons=['foo4','bar','eep','gpoing','floof','bop','gra','rarhr','arf','wheep','spllz','flof','grz'])
                    # self.addButtonRow(buttonSize=(100,100),title='ButtonRow4',buttons=['foo4','bar','eep','gpoing','floof','bop','rah','pow','grow','zoo'])
                    # self.addButtonRow(buttonSize=(100,100),title='ButtonRow4',buttons=['foo4','bar','eep','gpoing','floof','bop','pop','dingle'])

                def addButtonRow(self,buttonSize,title,buttons):
                    self._rows.append({'buttonSize':buttonSize,
                                       'title':title,
                                       'buttons':copy.deepcopy(buttons)})

                def instantiate(self,scrollWidget,tabButton):
                    import _bs

                    titleSpacing = 40
                    buttonBorder = 20
                    buttonSpacing = 4
                    buttonOffsetH = 40
                    
                    self._height = 20
                    for i,row in enumerate(self._rows):
                        self._height += titleSpacing
                        bWidth,bHeight = row['buttonSize']
                        bColumnCount = int(math.floor((self._width-buttonOffsetH-20)/(bWidth+buttonSpacing)))
                        bRowCount = int(math.ceil(float(len(row['buttons']))/bColumnCount))
                        bHeightTotal = 2*buttonBorder+bRowCount*bHeight+(bRowCount-1)*buttonSpacing
                        self._height += bHeightTotal

                    c = bs.containerWidget(parent=scrollWidget,scale=1.0,size=(self._width,self._height),
                                                                singleDepth=True,background=False)
                    bs.containerWidget(edit=c,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)
                    v = self._height - 20
                    
                    prevRowButtons = None
                    thisRowButtons = []

                    
                    delay = 300
                    for row in self._rows:
                        bs.textWidget(parent=c,position=(45,v-titleSpacing*0.7),size=(0,0),
                                      scale=1.0,
                                      transitionDelay=delay,
                                      color=(0.7,0.9,0.7,1),hAlign="left",vAlign="center",
                                      text=row['title'],maxWidth=self._width*0.7)
                        delay = max(100,delay-100)
                        v -= titleSpacing
                        v -= buttonBorder
                        bWidth,bHeight = row['buttonSize']
                        bCount = len(row['buttons'])
                        bColumnCount = int(math.floor((self._width-buttonOffsetH-20)/(bWidth+buttonSpacing)))
                        col = 0
                        for i,bType in enumerate(row['buttons']):

                            bInfo = self._storeWindow._buttonInfos[bType] = {}
                            
                            # temp - hard-coded
                            if bType == 'frosty':
                                bInfo['name'] = title = bs.translate('characterNames','Frosty')
                                tintColor=(0.5,0.5,1.0)
                                tint2Color=(1,0.5,0)
                                iconTex='frostyIcon'
                                tintTex='frostyIconColorMask'
                                #action = bs.WeakCall(self._storeWindow._buy,'frosty')
                                bInfo['call'] = bs.WeakCall(self._storeWindow._buy,'frosty')
                            elif bType == 'santa':
                                bInfo['name'] = title = bs.translate('characterNames','Santa Claus')
                                tintColor=(1,0,0)
                                tint2Color=(1,1,1)
                                iconTex='santaIcon'
                                tintTex='santaIconColorMask'
                                #action = bs.WeakCall(self._storeWindow._buy,'santa')
                                bInfo['call'] = bs.WeakCall(self._storeWindow._buy,'santa')

                            bPos = (buttonOffsetH+(bWidth+buttonSpacing)*col,v-bHeight)
                            bInfo['button'] = b = bs.buttonWidget(parent=c,position=bPos,
                                                                  transitionDelay=delay,
                                                                  showBufferTop=76.0,
                                                                  enableSound=False,
                                                                  buttonType='square',
                                                                  size=(bWidth,bHeight),
                                                                  #call=action,
                                                                  label='')
                            bOffsX = -0.015*bWidth

                            frameSize = bWidth*0.7
                            insetTex = bs.getTexture('frameInset')
                            insetModel = bs.getModel('frameInset')

                            imDim = frameSize*(100.0/113.0)
                            imPos = (bPos[0]+bWidth*0.5-imDim*0.5+bOffsX,bPos[1]+bHeight*0.57-imDim*0.5)
                            framePos = (bPos[0]+bWidth*0.5-frameSize*0.5+bOffsX,bPos[1]+bHeight*0.57-frameSize*0.5)
                            bs.imageWidget(parent=c,position=imPos,size=(imDim,imDim),color=(1,1,1),
                                           hasAlphaChannel=False,
                                           transitionDelay=delay,
                                           drawController=b,
                                           texture=bs.getTexture(iconTex,minQuality='High'),
                                           tintTexture=bs.getTexture(tintTex),
                                           tintColor=tintColor,tint2Color=tint2Color)
                            bInfo['frameWidget'] = bs.imageWidget(parent=c,position=framePos,
                                                                  size=(frameSize,frameSize),
                                                                  #color=(0.35,0.5,0.15),
                                                                  transitionDelay=delay,
                                                                  drawController=b,
                                                                  texture=insetTex,modelTransparent=insetModel)


                            bs.textWidget(parent=c,text=title,position=(bPos[0]+bWidth*0.5+bOffsX,bPos[1]+bHeight*0.24),
                                                               transitionDelay=delay,scale=bWidth*(1.0/230.0),
                                                               maxWidth=bWidth*0.9,size=(0,0),hAlign='center',vAlign='center',
                                                               drawController=b,
                                                               color=(0.7,0.9,0.7,1.0))
                            price = _bs._getPrice(bType)
                            if price is None: price = ''
                            bInfo['price'] = price
                            bInfo['priceWidget'] = bs.textWidget(parent=c,text='',position=(bPos[0]+bWidth*0.5+bOffsX,bPos[1]+bHeight*0.14),
                                                               transitionDelay=delay,scale=bWidth*(1.0/300.0),
                                                               maxWidth=bWidth*0.9,size=(0,0),hAlign='center',vAlign='center',
                                                               drawController=b,
                                                               color=(0.1,1.0,0.1,1.0))
                            delay = max(100,delay-100)
                            thisRowButtons.append(b)



                            # wire this button to the equivalent in the previous row
                            if prevRowButtons is not None:
                                if len(prevRowButtons) > col:
                                    bs.widget(edit=b,upWidget=prevRowButtons[col])
                                    bs.widget(edit=prevRowButtons[col],downWidget=b)
                                    # if we're the last button in our row, wire any in the previous row past
                                    # our position to go to us if down is pressed
                                    if col+1 == bColumnCount or i == bCount-1:
                                        for bPrev in prevRowButtons[col+1:]:
                                            bs.widget(edit=bPrev,downWidget=b)
                                else:
                                    bs.widget(edit=b,upWidget=prevRowButtons[-1])
                            else:
                                bs.widget(edit=b,upWidget=tabButton)
                                
                            col += 1
                            if col == bColumnCount or i == bCount-1:
                                prevRowButtons = thisRowButtons
                                thisRowButtons = []
                                col = 0
                                v -= bHeight
                                if i < bCount-1: v -= buttonSpacing

                        v -= buttonBorder

                    # set a timer to update these buttons periodically as long as we're alive
                    # (so if we buy one it will grey out, etc)
                    self._storeWindow._updateButtonsTimer = bs.Timer(500,bs.WeakCall(self._storeWindow._updateButtons),repeat=True,timeType='real')
                    # also update them immediately
                    self._storeWindow._updateButtons()
                    
            if self._currentTabIndex in [0]:
                store = _Store(self,data,self._scrollWidth)
                store.instantiate(scrollWidget=self._scrollWidget,tabButton=self._tabButtons[self._currentTabIndex])

            else:
                c = bs.containerWidget(parent=self._scrollWidget,scale=1.0,size=(self._scrollWidth,self._scrollHeight*0.95),
                                       singleDepth=True,background=False)
                bs.containerWidget(edit=c,claimsLeftRight=True,claimsTab=True,selectionLoopToParent=True)
                self._statusTextWidget = bs.textWidget(parent=c,position=(self._scrollWidth*0.5,self._scrollHeight*0.5),size=(0,0),
                                                       scale=1.3,
                                                       transitionDelay=100,
                                                       color=(1,1,0.3,1.0),hAlign="center",vAlign="center",
                                                       text=self._R.comingSoonText,maxWidth=self._scrollWidth*0.9)
                

    def _back(self):
        bs.containerWidget(edit=self._rootWidget,transition='outRight')
        uiGlobals['mainMenuWindow'] = MainMenuWindow(transition='inLeft').getRootWidget()
